".Last.value" <- 
function(obj, sim.number = 1, ...)
NULL

"coords.aniso" <- 
function(coords, aniso.pars)
{
	n <- nrow(coords)
	rotation <- aniso.pars[1]
	lambda <- aniso.pars[2]
	if(lambda < 1)
		stop("anisotropy ratio must be geater then 1")
	rm <- matrix(c(cos(rotation),  - sin(rotation), sin(rotation), cos(
		rotation)), ncol = 2)
	tm <- matrix(c(1, 0, 0, 1/lambda), ncol = 2)
	coords.cor <- coords %*% rm %*% tm
	return(coords.cor)
}

"cov.spatial" <- 
function(obj, cov.model = c("exponential", "matern", "gaussian", "spherical",
	"circular", "cubic", "wave", "powered.exponential", "cauchy", 
	"gneiting", "gneiting.matern", "pure.nugget"), kappa = NULL, cov.pars
	 = stop("no cov.pars argument provived"))
{
	# Computes the covariance values from a distances matrix
	# obj : object (vector, matrix, array) with distances
	# cov.model : 'matern', 'wave', 'sph', 'gaussian', "cauchy","gneiting","gneiting.matern"
	## see TECH. REPORT ST-99-10, BY SCHLATHER, M. DEPT MATHS & STATS, LANCASTER UN. for a description os the covariance functions and associated random fields.
	# kappa : extra parameter for matern , powered-exponential, gneiting and gneiting.matern covariance functions
	# cov.pars: (sigmasq,phi) If cov.pars is a matrix, each line identify 1 structure
	cov.model <- match.arg(cov.model)
	if(cov.model == "matern" | cov.model == "powered.exponential" | 
		cov.model == "cauchy" | cov.model == "gneiting.matern")
		if(is.null(kappa))
			stop("for matern, powered.exponential, cauchy, gneiting and gneiting.matern covariance functions the parameter kappa must be provided"
				)
	if(cov.model == "gneiting.matern" & length(kappa) != 2)
		stop("gneiting.matern correlation function model requires a vector with 2 parameters in the argument kappa"
			)
	if((cov.model == "matern" | cov.model == "powered.exponential" | 
		cov.model == "cauchy") & length(kappa) != 1)
		stop("for this choice of  correlation function model kappa should be a scalar parameter"
			)
	if(is.vector(cov.pars))
		sigmasq <- cov.pars[1]
	else sigmasq <- cov.pars[, 1]
	if(is.vector(cov.pars))
		phi <- cov.pars[2]
	else phi <- cov.pars[, 2]
	if(is.vector(cov.pars))
		ns <- 1
	else ns <- nrow(cov.pars)
	covs <- array(0, dim = dim(obj))
	for(i in 1:ns) {
		if(phi[i] < 9.9999999999999998e-13)
			cov.model <- "pure.nugget"
		cov <- switch(cov.model,
			pure.nugget = rep(0, length(obj)),
			wave = (1/obj) * (phi[i] * sin(obj/phi[i])),
			exponential = exp( - (obj/phi[i])),
			matern = matern(u = obj, phi = phi[i], kappa = kappa),

				gaussian = exp( - ((obj/phi[i])^2)),
			spherical = ifelse(obj < phi[i], (1 - 1.5 * (obj/phi[
				i]) + 0.5 * (obj/phi[i])^3), 0),
			circular = {
				obj.sc <- obj/phi[i]
				obj.sc[obj.sc > 1] <- 1
				ifelse(obj < phi[i], (1 - (2 * ((obj.sc) * sqrt(
					1 - ((obj.sc)^2)) + asin(obj.sc)))/
					pi), 0)
			}
			,
			cubic = {
				obj.sc <- obj/phi[i]
				ifelse(obj < phi[i], (1 - (7 * (obj.sc^2) -
					8.75 * (obj.sc^3) + 3.5 * (obj.sc^
					5) - 0.75 * (obj.sc^7))), 0)
			}
			,
			powered.exponential = {
				if(kappa > 2 | kappa <= 0)
					stop("for power model parameter phi must be in (0,2]"
						)
				exp( - ((obj/phi[i])^kappa))
			}
			,
			cauchy = (1 + (obj/phi[i])^2)^(-1 * kappa),
			gneiting = {
				obj.sc <- obj/phi[i]
				t2 <- (1 - obj.sc)
				t2 <- ifelse(t2 > 0, (t2^8), 0)
				(1 + 8 * obj.sc + 25 * (obj.sc^2) + 32 * (
					obj.sc^3)) * t2
			}
			,
			gneiting.matern = {
				obj.sc <- obj/(phi[i] * kappa[2])
				t2 <- (1 - obj.sc)
				t2 <- ifelse(t2 > 0, (t2^8), 0)
				cov <- (1 + 8 * obj.sc + 25 * (obj.sc^2) + 32 *
					(obj.sc^3)) * t2
				cov * matern(u = obj, phi = phi[i], kappa = 
					kappa[1])
			}
			,
			stop("wrong or no specification of cov.model"))
		cov <- sigmasq[i] * cov
		covs <- covs + cov
	}
	return(covs)
}

"dist0.krige" <- 
function(x0, coords)
{
	if(length(x0) != 2)
		stop(paste("length of x0 is", length(x0), "(it must be 2)"))
	coords[, 1] <- coords[, 1] - x0[1]
	coords[, 2] <- coords[, 2] - x0[2]
	return(sqrt(coords[, 1]^2 + coords[, 2]^2))
}

"dist2full" <- 
function(dis)
{
	n <- attr(dis, "Size")
	full <- matrix(0, n, n)
	full[lower.tri(full)] <- dis
	full + t(full)
}

"grf" <- 
function(n, grid = "irreg", nx = round(sqrt(n)), ny = round(sqrt(n)), nsim = 1,
	xlims = c(0, 1), ylims = c(0, 1), cov.model = c("exponential", "matern",
	"gaussian", "spherical", "cubic", "wave", "powered.exponential", 
	"cauchy", "gneiting", "gneiting.matern", "pure.nugget"), cov.pars = 
	stop("covariance parameters (sigmasq and phi) needed"), nugget = 0,
	lambda = 1, kappa = NULL, method = c("cholesky", "svd", "eigen", 
	"circular.embedding"), plotgrid = F, plotvario = F, messages.screen = T,
	...)
{
	# This function generates realizations of a gaussian random field with zero mean. "Nested" variograms models (more than 1 covariance stru(varcov.eig$vec %*% sqrt(varcov.eig$val) %*% t(varcov.eig$vec))cture) are also allowed.
	#
	# n : number of points in the random field
	# grid : 'reg' generates a regular grid, 'irreg' generates a irregular grid, or if a matrix (nx2) is entered is used as a grid
	# nx and ny (for simulations on a regular grid only): number of points in X and Y directions
	# nsim : number of simulations (realizations of the GRF)
	# xlims: min and max X coordinates
	# ylims: min and max Y coordinates
	# cov.model: model for covariance (abbreviations allowed) see function cov.spatial
	# cov.pars: variogram parameters, a vector  c(sigmasq, phi) or, for models with more than 1 structure, a matrix with 2 columns for sigma and phi 
	# kappa : required if the correlation function specified in "cov.model" is one of the following: "matern", "powered exponential", "cauchy" and "gneiting.matern"
	## for matern: kappa=0.5 => exponential model
	## for powered exponential: kappa=1 => exponential model, kappa=2 => gaussian model
	# method : simulation algorithm. Option for 'cholesky', 'svd' (singular value decomposition) or "circular.embedding"  (abbreviations allowed)
	# option "circular.embedding": only for regular grids and recommended for simulation of fields with very large number of points
	#  "circular.embedding": algorithm provided by Martin Schlather
	# See Tech. Report ST-99-10, Depto Maths & Stats, Lancaster University
	# plotgrid : plot the locations on the current graphics device
	# plotvario : plot the given variogram model and the estimated variogram(s) for the simulation(s) on a opened graphics devide
	call <- match.call()
	method <- match.arg(method)
	cov.model <- match.arg(cov.model)
	rseed <- .Random.seed
	# Preparing the locations to be simulated
	xy <- grid.grf(grid = grid, n, xlims, ylims, nx, ny, messages.screen = 
		messages.screen)
	n <- dim(xy)[1]
	if(plotgrid == T) {
		par.ori <- par()
		if((xlims[2] - xlims[1]) == (ylims[2] - ylims[1]))
			par(pty = "s")
		plot(xy[, 1], xy[, 2], xlab = "Coord X", ylab = "Coord Y",
			xlim = xlims, ylim = ylims)
		par(pty = "m")
	}
	# Reading the parameters
	tausq <- nugget
	if(is.vector(cov.pars)) {
		sigmasq <- cov.pars[1]
		phi <- cov.pars[2]
		nst <- 1
	}
	else {
		sigmasq <- cov.pars[, 1]
		phi <- cov.pars[, 2]
		nst <- nrow(cov.pars)
	}
	sill.tot <- tausq + sum(sigmasq)
	# Printing some messages 
	messa <- grf.aux1(nst, nugget, sigmasq, phi, kappa, cov.model)
	if(messages.screen == T) {
		print(messa$nst)
		print(messa$nugget)
		print(messa$cov.structures)
		print(paste("Algorithm used is: ", method))
	}
	# Performing simulation: 
	# check == here
	if(all(phi) == 0) zsim <- matrix(rnorm((n * nsim), mean = 0, sd = sqrt(
			sill.tot)), nrow = n, ncol = nsim) else {
		if(method == "cholesky" | method == "svd" | method == "eigen") 
			{
			zsim <- matrix(rnorm((n * nsim)), nrow = n, ncol = nsim
				)
			if(method == "cholesky") {
				cov.decomp <- varcov.spatial(coords = xy, 
					cov.model = cov.model, kappa = kappa,
					nugget = 0, cov.pars = cov.pars, 
					only.decomp = T, func.inv = "chol")
			}
			if(method == "svd") {
				cov.decomp <- varcov.spatial(coords = xy, 
					cov.model = cov.model, kappa = kappa,
					nugget = 0, cov.pars = cov.pars, 
					func.inv = "svd", only.decomp = T)
			}
			if(method == "eigen") {
				cov.decomp <- varcov.spatial(coords = xy, 
					cov.model = cov.model, kappa = kappa,
					nugget = 0, cov.pars = cov.pars, 
					func.inv = "eigen", only.decomp = T)
			}
			zsim <- cov.decomp %*% zsim
		}
		if(method == "circular.embedding") {
			if(grid == "irreg")
				stop("circular.embedding algorithm only allowed for regular grids"
					)
			stepx <- (xlims[2] - xlims[1])/(nx - 1)
			stepy <- (ylims[2] - ylims[1])/(ny - 1)
			if(round(100000000. * stepx) != round(100000000. * 
				stepy))
				stop("distance between points must be the same in X and Y directions"
					)
			temp <- list(n = n, nst = nst, sigmasq = sigmasq, xlims
				 = xlims, ylims = ylims, stepx = stepx, 
				cov.model = cov.model, phi = phi, kappa = kappa
				)
			zsim <- apply(as.matrix(1:nsim), 1, grf.aux3, temp = 
				temp)
		}
		if(nugget != 0) {
			zsim <- zsim + matrix(rnorm((n * nsim), sd = sqrt(
				nugget)), ncol = nsim)
		}
		if(nsim == 1)
			zsim <- as.vector(zsim)
	}
	if(messages.screen == T)
		print(paste(
			"End of simulation procedure. Number of realizations:",
			nsim))
	# Plotting the simulations
	if(plotvario == T) {
		print("preparing the variogram for plotting")
		if(lambda != 1) {
			if(messages.screen == T)
				print("Notice that the plotted variograms are for the UN-transformed (Gaussian) data"
					)
		}
		sim.bin <- variog(coords = xy, data = zsim)
		ymax <- max(1.1000000000000001 * sill.tot, sim.bin$v)
		if(ymax > 2 * sill.tot)
			ymax <- 2 * sill.tot
		plot(sim.bin, ylim = c(0, ymax))
		dmax <- max(sim.bin$u)
		var.model <- list(nugget = nugget, cov.pars = cov.pars, kappa
			 = kappa, dmax = dmax, cov.model = cov.model)
		lines.variomodel(var.model, max.dist = dmax, lwd = 3)
	}
	# Preparing output
	if(lambda != 1) {
		if(lambda != 0)
			zsim <- (zsim * lambda + 1)^(1/lambda)
		else zsim <- exp(zsim)
		if(messages.screen == T)
			print(paste("Transforming data (Box-Cox), lambda =",
				lambda))
		messa$transformation <- paste(
			"Data transformed (Box-Cox), for lambda =", lambda)
	}
	results <- list(coords = xy, data = zsim, cov.model = cov.model, nugget
		 = nugget, cov.pars = cov.pars, kappa = kappa, messages = messa,
		method = method)
	results$.Random.seed <- rseed
	results$call <- call
	if(version$major <= 4)
		class(results) <- setOldClass(c("grf", "geodata"))
	else class(results) <- c("grf", "geodata")
	return(results)
}

"grf.aux1" <- 
function(nst, nugget, sigmasq, phi, kappa, cov.model)
{
	cov.nst <- paste("Process with ", nst, " covariance structure(s)")
	cov.nugget <- paste("The nugget effect is: ", nugget)
	cov.message <- NULL
	for(i in 1:nst) {
		# check == here
		if(phi[i] == 0) cov.message[i] <- paste("covariance model",
				i, "is a pure nugget effect") else {
			if(cov.model == "matern" | cov.model == 
				"powered.exponential" | cov.model == "cauchy" |
				cov.model == "gneiting-matern")
				cov.message[i] <- paste("covariance model ",
					i, " is: ", cov.model, "(sigmasq=",
					sigmasq[i], ", phi=", phi[i], 
					", kappa=", kappa, ")", sep = "")
			else cov.message[i] <- paste("covariance model ", i,
					" is: ", cov.model, "(sigmasq=", 
					sigmasq[i], ", phi=", phi[i], ")",
					sep = "")
		}
	}
	return(list(nst = cov.nst, nugget = cov.nugget, cov.structures = 
		cov.message))
}

"grf.aux2" <- 
function(xlim, ylim, step, cov.model, phi, kappa = NULL)
{
	# This function was provided by Martin Schlather.
	## the C-algorithm plots some numbers when running: if the numbers do 
	## increase, the algorithm does not work well (and an approximating
	## algorithm should be used which is not implemented yet.)
	##
	## xlim=c(*,*), ylim=c(*,*) gives the size of the random field
	## step : gridlength
	## parameter: 1.scale, 2.shape 3.other, see also C code
	##
	if(!is.null(kappa)) if(cov.model == "matern" & kappa == 0.5) {
			cov.model <- "exponential"
		}
	if(cov.model == "spherical") {
		covfct <- 1
		shape <- 1
	}
	if(cov.model == "gneiting.matern") {
		covfct <- 2
		shape <- kappa
	}
	if(cov.model == "powered.exponential") {
		covfct <- 3
		shape <- kappa
	}
	if(cov.model == "exponential") {
		covfct <- 3
		shape <- 1
	}
	if(cov.model == "gaussian") {
		covfct <- 3
		shape <- 2
	}
	if(cov.model == "matern") {
		covfct <- 4
		shape <- kappa
	}
	if(cov.model == "gneiting") {
		covfct <- 5
		shape <- 1
	}
	if(cov.model == "cauchy") {
		covfct <- 6
		shape <- kappa
	}
	if(cov.model == "wave") {
		covfct <- 7
		shape <- 1
	}
	parameter <- c(phi, shape, 1)
	nx <- c(diff(xlim)/step + 1, diff(ylim)/step + 1)
	# how many steps ??
	storage.mode(nx) <- "integer"
	ln <- 2
	storage.mode(ln) <- "integer"
	covnr <- covfct
	storage.mode(covnr) <- "integer"
	simustep <- step
	storage.mode(simustep) <- "double"
	param <- parameter
	storage.mode(param) <- "double"
	lparam <- length(parameter)
	storage.mode(lparam) <- "integer"
	res <- double(nx[1] * nx[2])
	mm <- integer(ln)
	#
	if(!is.loaded("woodandchan")) dyn.load(
			"/home/ribeiro/geoS/woodandchan.o")
	x <- .C("woodandchan",
		covnr,
		nx,
		ln,
		simustep,
		param,
		lparam,
		res = res,
		m = mm)$res
	return(x)
}

"grf.aux3" <- 
function(nsim, temp)
{
	realiz <- rep(0, temp$n)
	for(i in 1:temp$nst) {
		realiz <- realiz + sqrt(temp$sigmasq[i]) * grf.aux2(xlim = temp$
			xlims, ylim = temp$ylims, step = temp$stepx, cov.model
			 = temp$cov.model, phi = temp$phi[i], kappa = temp$
			kappa)
		NULL
	}
	return(realiz)
}

"grfclass" <- 
function(xy, sim, theta, kappa = NULL, estpars = F)
{
	# This function prepares a set of simulations to be used by plot.grf
	xyzsim <- cbind(xy, sim)
	nsim <- dim(xyzsim)[2] - 2
	if(estpars == T) {
		simpars <- matrix(0, nsim, 4)
		for(i in 1:nsim) {
			k <- i + 2
			pars <- trends.lik(xy, xyzsim[, k], ini = theta, kappa
				 = kappa, residuals = F)
			pars <- c(pars$theta, kappa = kappa)
			simpars[i,  ] <- pars
		}
		results <- list(xyzsim = xyzsim, simpars = simpars, theta = 
			theta, kappa = kappa)
	}
	else results <- list(xyzsim = xyzsim, theta = theta, kappa = kappa)
	class(results) <- "grf"
	return(results)
}

"grid.grf" <- 
function(grid, n, xgrid, ygrid, nx, ny, messages.screen = T)
{
	call <- match.call()
	if(is.matrix(grid)) {
		xy <- grid
		n <- dim(xy)[1]
		if(messages.screen == T)
			print("simulations on a grid provided by the user")
	}
	else {
		if(grid == "irreg") {
			xy <- cbind(x = runif(n, xgrid[1], xgrid[2]), y = runif(
				n, ygrid[1], ygrid[2]))
			if(messages.screen == T)
				print(paste("generating irregular grid with ",
					n, " points"))
		}
		else {
			xy <- as.matrix(expand.grid(x = seq(xgrid[1], xgrid[
				2], l = nx), y = seq(ygrid[1], ygrid[2], l = ny
				)))
			n <- nx * ny
			if(messages.screen == T)
				print(paste("generating grid ", nx, " * ",
					ny, " with ", n, " points"))
		}
	}
	return(xy)
}

"hessb" <- 
function(f, x, temp.list)
{
	ep <- 0.0001
	eps <- ep * x
	n <- length(x)
	m <- matrix(0, ncol = n, nrow = n)
	for(i in 1:n) {
		for(j in 1:n) {
			x1 <- x
			x1[i] <- x1[i] + eps[i]
			x1[j] <- x1[j] + eps[j]
			x2 <- x
			x2[i] <- x2[i] + eps[i]
			x2[j] <- x2[j] - eps[j]
			x3 <- x
			x3[i] <- x3[i] - eps[i]
			x3[j] <- x3[j] + eps[j]
			x4 <- x
			x4[i] <- x4[i] - eps[i]
			x4[j] <- x4[j] - eps[j]
			m[i, j] <- (f(x1, temp.list) - f(x2, temp.list) - f(
				x3, temp.list) + f(x4, temp.list))/(4 * eps[
				i] * eps[j])
		}
	}
	solve(m)
}

"image.grf" <- 
function(obj, sim.number = 1, ...)
{
	# plots an 'image plot' from a (selected) simulation of a random field generated by the function 'grf'
	# The simulated random filed must be regularly spaced
	x <- as.numeric(levels(as.factor(obj$coords[, 1])))
	nx <- length(x)
	y <- as.numeric(levels(as.factor(obj$coords[, 2])))
	ny <- length(y)
	obj$data <- as.matrix(obj$data)
	n <- nrow(obj$data)
	if(nx * ny != n)
		stop("Probably irregular grid")
	m <- matrix(obj$data[, sim.number], ncol = ny)
	image(x, y, m, ...)
	return(invisible())
}

"image.grfsim" <- 
function(obj, sim.number = 1, ...)
{
	# plots an 'image plot' from a (selected) simulation of a random field generated by the function 'grf'
	# The simulated random filed must be regularly spaced
	x <- as.numeric(levels(as.factor(obj$coords[, 1])))
	nx <- length(x)
	y <- as.numeric(levels(as.factor(obj$coords[, 2])))
	ny <- length(y)
	obj$data <- as.matrix(obj$data)
	n <- nrow(obj$data)
	if(nx * ny != n)
		stop("Probably irregular grid")
	m <- matrix(obj$data[, sim.number], ncol = ny)
	image(x, y, m, ...)
	return(invisible())
}

"image.kriging" <- 
function(kriging.obj, locations = kriging.obj$locations, values = kriging.obj$
	predict, ...)
{
	locations <- locations[order(locations[, 2], locations[, 1]),  ]
	x <- as.numeric(levels(as.factor(locations[, 1])))
	nx <- length(x)
	y <- as.numeric(levels(as.factor(locations[, 2])))
	ny <- length(y)
	if(nx == ny)
		par(pty = "s")
	values <- matrix(values, ncol = ny)
	image(x, y, values, ...)
	return(invisible())
}

"image.ksline" <- 
function(obj.ksline, grid = obj.ksline$grid, values = c("est", "kvar", 
	"oktrend", "ktrend", "ktetrend"), ...)
{
	values <- match.arg(values)
	grid <- obj.ksline$grid[order(grid[, 2], grid[, 1]),  ]
	x <- as.numeric(levels(as.factor(grid[, 1])))
	nx <- length(x)
	y <- as.numeric(levels(as.factor(grid[, 2])))
	ny <- length(y)
	if(nx == ny)
		par(pty = "s")
	if(values == "est")
		m <- matrix(obj.ksline$pred, ncol = ny)
	if(values == "kvar")
		m <- matrix(obj.ksline$krige.var, ncol = ny)
	image(x, y, m, ...)
	return(invisible())
}

"krige.bayes" <- 
function(geodata, coords = geodata$coords, data = geodata$data, locations = 
	"no", model = model.control(trend.d = "cte", trend.l = "cte", cov.model
	 = "exponential", kappa = 0.5, aniso.pars = NULL, lambda = 1), prior = 
	prior.control(beta.prior = c("flat", "normal", "fixed"), beta = NULL,
	beta.var = NULL, sill.prior = c("reciprocal", "fixed"), sill = NULL,
	range.prior = c("uniform", "exponential", "fixed", "squared.reciprocal",
	"reciprocal"), exponential.prior.par = 1, range = NULL, range.discrete
	 = NULL, nugget.prior = c("fixed", "uniform"), nugget = 0, 
	nugget.discrete = NULL), output = output.control(n.posterior = 1000,
	n.predictive = NULL, moments = TRUE, simulations.predictive = TRUE,
	keep.simulations = TRUE, mean.estimator = TRUE, quantile.estimator = 
	NULL, probability.estimator = NULL, signal = FALSE, messages.screen = 
	TRUE))
{
	call <- match.call()
	seed <- .Random.seed
	kb.results <- list(posterior = list(), predictive = list())
	##
	## reading input
	##
	cov.model <- model$cov.model
	kappa <- model$kappa
	lambda <- model$lambda
	##
	beta <- prior$beta
	beta.var <- prior$beta.var
	sill <- prior$sill
	range <- prior$range
	nugget <- prior$nugget
	exponential.prior.par <- prior$exponential.prior.par
	##
	n.posterior <- output$n.posterior
	n.predictive <- output$n.predictive
	moments <- output$moments
	messages.screen <- output$messages.screen
	##
	## checking data configuration
	##
	if(is.vector(coords)) {
		coords <- cbind(coords, 0)
		warning("vector of coordinates: one spatial dimention assumed")
	}
	coords <- as.matrix(coords)
	data.dist <- as.vector(dist(coords))
	## check == here
	if(round(1000000000000. * min(data.dist)) == 0) stop(
			"This function does not allow two data at same location"
			)
	if(all(locations == "no")) {
		if(prior$beta.prior != "fixed" & prior$sill.prior != "fixed" &
			prior$range.prior != "fixed")
			cat("krige.bayes: no prediction locations provided. Only Bayesian estimates of model parameters will be returned\n"
				)
		else stop("Locations to be predicted not provided. Bayesian parameter estimation allowed only if priors are provided for beta, sill and range"
				)
	}
	##
	## Cheking priors input
	##
	if(prior$beta.prior == "fixed" & is.null(beta)) stop(
			"if beta is fixed, its value must be provided in the argument beta"
			)
	if(prior$beta.prior == "normal" & (is.null(beta) | is.null(beta.var)))
		stop("if the prior for beta is normal, the prior mean(s) and prior (co)variance must be provided using the argument beta and priorr$bveta.var"
			)
	if(prior$sill.prior == "fixed" & is.null(sill))
		stop("if sill is fixed, its value must be provided in the argument sill"
			)
	if(prior$range.prior == "fixed" & is.null(range))
		stop("if range is fixed, its value must be provided in the argument range"
			)
	if(prior$beta.prior != "flat" & prior$sill.prior != "fixed")
		stop("selected prior for beta not allowed for selected prior for sill"
			)
	if(prior$nugget.prior != "fixed" & is.null(prior$nugget.discrete))
		stop("to include the nugget in the Bayesian analysis the argument nugget.discrete must be provided"
			)
	if(prior$nugget.prior != "fixed")
		if(any(prior$nugget.discrete > 1))
			warning("when nugget is considered as random in the Bayesian analysis the values in the argument nugget.discrete must be relative nugget: tausq/sigmasq"
				)
	##
	## checking output options
	##
	if(!is.null(output$quantile.estimator)) {
		if(is.numeric(output$quantile.estimator))
			if(any(output$quantile.estimator) < 0 | any(output$
				quantile.estimator) > 1)
				stop("quantiles indicators must be numbers in the interval [0,1]"
					)
		if(output$quantile.estimator == TRUE)
			output$quantile.estimator <- c(0.025000000000000001,
				0.5, 0.97499999999999998)
		if(!is.numeric(output$probability.estimator))
			stop("probability.estimator must be a numeric value (or vector) of cutoff value(s)"
				)
	}
	if(lambda != 1 & lambda != 0 & moments) {
		moments <- FALSE
		cat(paste("WARNING: moments cannot be computed for lambda =",
			lambda, ". Argument moments was set to FALSE\n"))
	}
	if(lambda < 0 & output$mean.estimator) {
		output$mean.estimator <- FALSE
		cat("krige.bayes: mean.predictor set to F. The resulting distribution does not has expectation for $lambda < 0$\n"
			)
	}
	##
	## Box-Cox transformation of the data
	##
	if(lambda != 1) {
		if(prior$range.prior == "fixed")
			stop("transformation option available only for the full Bayesian approach"
				)
		if(messages.screen)
			print(paste(
				"Data transformation (Box-Cox) for $lambda=",
				lambda), seq = "")
		if(lambda == 0)
			data <- log(data)
		else data <- ((data^lambda) - 1)/lambda
	}
	##
	## Checking the dimension of points to be predict (1 or 2)	
	##
	if(all(locations != "no")) {
		if(is.vector(locations)) {
			if(length(locations) == 2) {
				locations <- t(as.matrix(locations))
				warning("THE FUNCTION IS CONSIDERING THAT YOU HAVE ENTERED WITH 1 POINT TO BE PREDICTED IN A TWO DIMENSION REGION\n"
					)
			}
			else locations <- as.matrix(cbind(locations, 0))
		}
		else locations <- as.matrix(locations)
	}
	##
	## Building trend matrices:	
	##
	if(inherits(model$trend.d, "formula") | inherits(model$trend.l, 
		"formula")) {
		if(any(locations != "no"))
			if((inherits(model$trend.d, "formula") == FALSE) | (
				inherits(model$trend.l, "formula") == FALSE))
				stop("model$trend.d and model$trend.l must have similar specification"
					)
		if(messages.screen)
			cat("krige.bayes: Kriging with external trend to be performed using covariates provided by the user\n"
				)
	}
	else {
		if(any(locations != "no") & (model$trend.d != model$trend.l)) 
			{
			stop("model$trend.l is different from model$trend.d")
		}
		if(messages.screen) {
			if(model$trend.d == "cte")
				cat("krige.bayes: analysis assuming constant mean\n"
					)
			if(model$trend.d == "1st")
				cat("krige.bayes: analysis assuming a 1st degree polinomial trend\n"
					)
			if(model$trend.d == "2nd")
				cat("krige.bayes: analysis assuming a 2nd degree polinomial trend\n"
					)
		}
	}
	trend.d <- trend.spatial(trend = model$trend.d, coords = coords)
	if(any(locations != "no"))
		trend.l <- trend.spatial(trend = model$trend.l, coords = 
			locations)
	dimnames(coords) <- list(NULL, NULL)
	dimnames(trend.d) <- list(NULL, NULL)
	if(all(locations != "no")) {
		dimnames(locations) <- list(NULL, NULL)
		dimnames(trend.l) <- list(NULL, NULL)
	}
	beta.size <- ncol(trend.d)
	##
	## Checking dimensions
	##
	if(nrow(coords) != length(data)) stop(
			"number of data is different of number of data locations (coordinates)"
			)
	if(all(locations != "no")) {
		if(nrow(locations) != nrow(trend.l))
			stop("number of points to be estimated is different of the number of trend points"
				)
		dimnames(locations) <- list(NULL, NULL)
	}
	dimnames(coords) <- list(NULL, NULL)
	##
	## Anisotropy correction (should be placed after trend matrices be defined)
	##
	if(!is.null(model$aniso.pars)) {
		if(length(model$aniso.pars) != 2 | !is.numeric(model$aniso.pars
			))
			stop("anisotropy parameters must be provided as a numeric vector with two elements: the rotation angle and the anisotropy ratio"
				)
		cat("krige.bayes: anisotropy parameters provided and assumed to be constants\n"
			)
		coords <- coords.aniso(coords = coords, aniso.pars = model$
			aniso.pars)
		if(all(locations != "no"))
			locations <- coords.aniso(coords = locations, 
				aniso.pars = model$aniso.pars)
	}
	##
	n <- length(data)
	if(is.matrix(trend.l))
		ni <- dim(trend.l)[1]
	else ni <- length(trend.l)
	tausq <- nugget
	if(prior$sill.prior != "fixed")
		sigmasq <- 1
	else sigmasq <- sill
	##
	######################### PART 2 ##############################
	## Prediction with fixed phi
	##
	if(prior$range.prior == "fixed") {
		phi <- range
		## covariance matrix for data points
		invcov <- varcov.spatial(dists.lowertri = data.dist, cov.model
			 = cov.model, kappa = kappa, nugget = nugget, cov.pars
			 = c(sigmasq, phi), inv = TRUE, scaled = TRUE, 
			only.inv.lower.diag = TRUE)
		## Some matrices computations
		###    ttiv <- crossprod(trend.d, invcov)
		###    ttivtt <- ttiv %*% trend.d
		###    ittivtt <- solve(ttivtt)
		ttivtt <- as.double(rep(0, beta.size * beta.size))
		ttivtt <- .C("quadraticform_XAY",
			as.double(invcov$lower.inverse),
			as.double(invcov$diag.inverse),
			as.double(as.vector(trend.d)),
			as.double(as.vector(trend.d)),
			as.integer(beta.size),
			as.integer(beta.size),
			as.integer(n),
			res = ttivtt)$res
		attr(ttivtt, "dim") <- c(beta.size, beta.size)
		ittivtt <- solve(ttivtt)
		###    ttivz <- ttiv %*% data
		ttivz <- as.double(rep(0, beta.size))
		ttivz <- .C("quadraticform_XAY",
			as.double(invcov$lower.inverse),
			as.double(invcov$diag.inverse),
			as.double(as.vector(trend.d)),
			as.double(as.vector(data)),
			as.integer(beta.size),
			as.integer(1),
			as.integer(n),
			res = ttivz)$res
		## covariance vector between data points and prediction locations
		v0mat <- as.double(rep(0, n * ni))
		.C("loccoords",
			as.double(as.vector(locations[, 1])),
			as.double(as.vector(locations[, 2])),
			as.double(as.vector(coords[, 1])),
			as.double(as.vector(coords[, 2])),
			as.integer(ni),
			as.integer(n),
			v0mat,
			DUP = FALSE)
		attr(v0mat, "dim") <- c(n, ni)
		v0mat <- cov.spatial(obj = v0mat, cov.model = cov.model, kappa
			 = kappa, cov.pars = c(1, phi))
		###    tv0iv <- t(apply(v0mat, 2, crossprod, y = invcov))
		###    tv0ivz <- tv0iv %*% data
		tv0ivz <- as.double(rep(0, ni))
		tv0ivz <- .C("quadraticform_XAY",
			as.double(invcov$lower.inverse),
			as.double(invcov$diag.inverse),
			as.double(as.vector(v0mat)),
			as.double(data),
			as.integer(ni),
			as.integer(1),
			as.integer(n),
			res = tv0ivz)$res
		###    tv0ivx <- apply(tv0iv, 1, crossprod, y = trend.d)
		tv0ivx <- as.double(rep(0, ni * beta.size))
		tv0ivx <- .C("quadraticform_XAY",
			as.double(invcov$lower.inverse),
			as.double(invcov$diag.inverse),
			as.double(as.vector(v0mat)),
			as.double(as.vector(trend.d)),
			as.integer(ni),
			as.integer(beta.size),
			as.integer(n),
			res = tv0ivx)$res
		attr(tv0ivx, "dim") <- c(ni, beta.size)
		tb <- trend.l - tv0ivx
		##    b <- t(tb)
		###    tv0ivv0 <- diag(tv0iv %*% v0mat)
		tv0ivv0 <- as.double(rep(0, ni))
		tv0ivv0 <- .C("quadraticform_XAX",
			as.double(invcov$lower.inverse),
			as.double(invcov$diag.inverse),
			as.double(as.vector(v0mat)),
			as.integer(ni),
			as.integer(n),
			res = tv0ivv0)$res
		v0mat <- NULL
		##
		########################## PART 2.1 ##############################
		## Simple kriging (fixed beta and sigmasq)
		##
		if(prior$beta.prior == "fixed") {
			beta.mean <- paste("SK : beta provided by user: ",
				beta)
			kb.results$predictive$mean <- as.vector(tv0ivz + 
				as.vector(tb %*% beta))
			if(output$signal)
				kb.results$predictive$variance <- as.vector(
					sigmasq * (1 - tv0ivv0))
			else kb.results$predictive$variance <- as.vector(nugget +
					sigmasq * (1 - tv0ivv0))
			priors.messa <- c("no priors for parameters", 
				"results corresponds to simple kriging")
			sill.mean <- paste("sill provided by user: ", sill)
		}
		##
		######################### PART 2.2 ##################################
		## uncertainty in beta and sigmasq
		##
		if(prior$beta.prior == "flat") {
			beta.mean <- ittivtt %*% ttivz
			###      predict <- tv0ivz + b %*% beta.mean
			kb.results$predictive$mean <- as.vector(tv0ivz + 
				as.vector(tb %*% beta.mean))
			##      bi <- apply(tb, 2, crossprod, y = ittivtt)
			##      if(ncol(trend.d) > 1)
			##        bi <- t(bi)
			##      bitb <- diag(bi %*% tb)
			if(beta.size == 1) bitb <- as.vector(tb^2) * as.vector(
					ittivtt) else {
				bitb <- as.double(rep(0, ni))
				bitb <- .C("quadraticform_XAX",
					as.double(ittivtt[lower.tri(ittivtt)]),

						as.double(diag(ittivtt)),
					as.double(as.vector(t(tb))),
					as.integer(ni),
					as.integer(beta.size),
					res = bitb)$res
			}
			if(prior$sill.prior == "fixed") {
				##
				## 2.2.1 Ordinary kriging: uncertainty only in beta, flat prior
				##
				if(output$signal) kb.results$predictive$
						variance <- as.vector(sigmasq *
						(1 - tv0ivv0 + bitb)) else 
						kb.results$predictive$variance <-
						as.vector(nugget + sigmasq *
						(1 - tv0ivv0 + bitb))
				priors.messa <- c(
					"uninformative prior for beta", 
					"no prior for sill", 
					"no prior for range", 
					"kriging estimates and variances equivalent to ordinary kriging"
					)
				sill.mean <- paste("sill provided by user: ",
					sill)
			}
			else {
				##
				## 2.2.2 using a conjugate normal prior
				##
				resid <- as.vector(data - trend.d %*% beta.mean
					)
				###        sill.mean <- (resid %*% invcov %*% resid)/
				###          (n - length(beta.mean) - 2)
				resivres <- as.double(0)
				resivres <- .C("quadraticform_XAX",
					as.double(invcov$lower.inverse),
					as.double(invcov$diag.inverse),
					as.double(resid),
					as.integer(1),
					as.integer(n),
					res = resivres)$res
				sill.mean <- resivres/(n - length(beta.mean) -
					2)
				if(output$signal)
					kb.results$predictive$variance <- 
						as.vector(sill.mean * (1 - 
						tv0ivv0 + bitb))
				else kb.results$predictive$variance <- 
						as.vector(nugget + sill.mean *
						(1 - tv0ivv0 + bitb))
				priors.messa <- c(
					"uninformative prior for beta", 
					"uninformative prior for sill", 
					"kriging estimates (but not variances) equivalent to ordinary kriging"
					)
			}
		}
		##
		## 2.2.3 Uncertainty only in the mean parameter (with normal prior)
		##
		if(prior$beta.prior == "normal") {
			ibeta.var <- solve(beta.var)
			varbgz <- solve(ibeta.var + ttivtt/sigmasq)
			beta.mean <- varbgz %*% (beta %*% ibeta.var + ttivz/
				sigmasq)
			kb.results$predictive$mean <- as.vector(tv0ivz + 
				as.vector(tb %*% beta.mean))
			###      bi <- apply(tb, 2, crossprod, y = varbgz)
			###      bitb <- diag(bi %*% tb)
			if(beta.size == 1) bitb <- as.vector(tb^2) * as.vector(
					ittivtt) else {
				bitb <- as.double(rep(0, ni))
				bitb <- .C("quadraticform_XAX",
					as.double(ittivtt[lower.tri(ittivtt)]),

						as.double(diag(ittivtt)),
					as.double(as.vector(t(tb))),
					as.integer(ni),
					as.integer(beta.size),
					res = bitb)$res
			}
			if(output$signal)
				kb.results$predictive$variance <- as.vector(
					sigmasq * (1 - tv0ivv0) + bitb)
			else kb.results$predictive$variance <- as.vector(nugget +
					sigmasq * (1 - tv0ivv0) + bitb)
			priors.messa <- c("Normal prior for beta", 
				"no prior for sill", "no prior for range")
			sill.mean <- paste("provided by user: ", sill)
		}
		##
		######################### PART 2.3 ##################################
		## Preparing output
		##
		###    kb.results$predictive$mean <- as.vector(predict)
		###    kb.results$predictive$variance <- as.vector(krige.var)
		kb.results$posterior$beta.mean <- beta.mean
		kb.results$posterior$sigmasq.mean <- sill.mean
		kb.results$posterior$phi.mean <- paste("provided by user: ",
			range)
		kb.results$type.prediction <- priors.messa
	}
	else {
		##	
		######################### PART 3 ##############################
		## considering the uncertainty in the parameter phi
		##
		if(messages.screen) cat(
				"krige.bayes: computing the discrete approximation for the posterior distribution of phi\n"
				)
		##	
		######################### PART 3.1 ##############################
		## Computing the discrete posterior distribution for phi
		##
		gl <- as.vector(n - beta.size)
		phidist <- list()
		if(is.null(prior$range.discrete)) {
			phidist$phi <- seq(0, max(data.dist), l = 51)
			warning("argument range.discrete not provided. Default values assumed"
				)
		}
		else phidist$phi <- prior$range.discrete
		if(is.null(prior$nugget.discrete)) {
			nugget.discrete <- nugget
			cat(paste(
				"krige.bayes: nugget assumed to be fixed and equals to",
				nugget, "\n"))
		}
		else nugget.discrete <- prior$nugget.discrete
		n.range <- length(phidist$phi)
		n.nugget <- length(nugget.discrete)
		phi.val <- phidist$phi
		phidist$phi <- as.matrix(expand.grid(phidist$phi, 
			nugget.discrete))
		dimnames(phidist$phi) <- list(NULL, NULL)
		temp.res <- apply(phidist$phi, 1, krige.bayes.aux1, coords = 
			coords, trend.d = trend.d, data = data, cov.model = 
			cov.model, kappa = kappa, range.prior = prior$
			range.prior, exponential.prior.par = 
			exponential.prior.par, n = n)
		extract.f <- function(obj)
		{
			return(obj$s2phi)
		}
		phidist$s2 <- sapply(temp.res, extract.f)
		attr(phidist$s2, "dim") <- c(n.range, n.nugget)
		extract.f <- function(obj)
		{
			return(obj$logprobphi)
		}
		phidist$logprobphi <- sapply(temp.res, extract.f)
		attr(phidist$logprobphi, "dim") <- c(n.range, n.nugget)
		extract.f <- function(obj)
		{
			return(obj$inv.lower)
		}
		inv.lower <- array(t(sapply(temp.res, extract.f)), dim = c(
			n.range, n.nugget, (n * (n - 1))/2), dimnames = list(
			NULL, NULL, NULL))
		extract.f <- function(obj)
		{
			return(obj$inv.diag)
		}
		inv.diag <- array(t(sapply(temp.res, extract.f)), dim = c(
			n.range, n.nugget, n), dimnames = list(NULL, NULL,
			NULL))
		extract.f <- function(obj)
		{
			return(obj$vbetaphi)
		}
		if(beta.size == 1) {
			phidist$vbeta.vector <- sapply(temp.res, extract.f)
			attr(phidist$vbeta.vector, "dim") <- c(n.range, 
				n.nugget, 1)
		}
		else {
			phidist$vbeta.vector <- t(sapply(temp.res, extract.f))
			attr(phidist$vbeta.vector, "dim") <- c(n.range, 
				n.nugget, (beta.size^2))
		}
		extract.f <- function(obj)
		{
			return(obj$betaphi)
		}
		if(beta.size == 1) {
			phidist$beta <- sapply(temp.res, extract.f)
			attr(phidist$beta, "dim") <- c(n.range, n.nugget, 1)
		}
		else {
			phidist$beta <- t(sapply(temp.res, extract.f))
			attr(phidist$beta, "dim") <- c(n.range, n.nugget, 
				beta.size)
		}
		temp.res <- NULL
		phidist$logprobphi <- phidist$logprobphi + abs(min(phidist$
			logprobphi))
		phidist$probphi <- exp(phidist$logprobphi)
		sumprob <- sum(phidist$probphi)
		phidist$probphi <- phidist$probphi/sumprob
		##
		######################### PART 3.2 ##############################
		## Sampling from the (parameters) posterior distribution
		##
		if(messages.screen) cat(
				"krige.bayes: sampling from multivariate distribution of the parameters\n"
				)
		ind <- sample((1:(n.range * n.nugget)), n.posterior, replace = 
			T, prob = as.vector(phidist$probphi))
		ind.unique <- sort(unique(ind))
		ind.length <- length(ind.unique)
		ind.table <- table(ind)
		phi.unique <- phidist$phi[ind.unique,  ]
		if(messages.screen) {
			cat("krige.bayes: samples and their frequencies from the distribution of  phi and tau.rel\n"
				)
			print(rbind(phi = phi.unique[, 1], nugget.rel = 
				phi.unique[, 2], frequency = ind.table))
		}
		phi.sam <- phidist$phi[ind,  ]
		vecpars.back.order <- order(ind)
		vec.s2 <- rep(as.vector(phidist$s2)[ind.unique], ind.table)
		samples.chisq <- rchisq(n.posterior, df = gl)
		sigmasq <- (gl * vec.s2)/samples.chisq
		if(beta.size == 1) {
			samples.beta <- rnorm(n.posterior, mean = 0, sd = 1)
			vec.beta <- rep(as.vector(phidist$beta)[ind.unique],
				ind.table)
			vec.vbeta <- rep(as.vector(phidist$vbeta.vector)[
				ind.unique], ind.table)
			beta <- vec.beta + sqrt(sigmasq * vec.vbeta) * 
				samples.beta
		}
		else {
			ind.beta <- matrix(phidist$beta, ncol = beta.size)[
				ind.unique,  ]
			ind.beta <- ind.beta[rep(1:ind.length, ind.table),
				]
			ind.vbeta <- matrix(phidist$vbeta.vector, ncol = 
				beta.size^2)[ind.unique,  ]
			ind.vbeta <- ind.vbeta[rep(1:ind.length, ind.table),
				] * sigmasq
			##      print("2.4: try to speed up this bit!")
			temp.res <- apply(ind.vbeta, 1, krige.bayes.aux3, 
				beta.size = beta.size)
			beta <- ind.beta + t(temp.res)
			temp.res <- NULL
		}
		if(beta.size == 1) {
			trend.mean <- mean(beta)
			trend.median <- median(beta)
		}
		else {
			trend.mean <- apply(beta, 2, mean)
			trend.median <- apply(beta, 2, median)
		}
		sill.mean <- mean(sigmasq)
		sill.median <- median(sigmasq)
		range.marg <- apply(phidist$probphi, 1, sum)
		range.marg <- range.marg/(sum(range.marg))
		range.mean <- phi.val %*% range.marg
		range.median <- median(phi.sam[, 1])
		## check == here
		range.mode <- phi.val[range.marg == max(range.marg)]
		nugget.marg <- apply(phidist$probphi, 2, sum)
		nugget.marg <- nugget.marg/(sum(nugget.marg))
		nugget.mean <- nugget.discrete %*% nugget.marg
		nugget.median <- median(phi.sam[, 2])
		## check == here
		nugget.mode <- nugget[nugget.marg == max(nugget.marg)]
		##
		## Computing the conditional (on phi and tausq.rel, 
		## the later if the case) modes for beta and sigmasq
		##
		invcov.mode <- varcov.spatial(dists.lowertri = data.dist, 
			cov.model = cov.model, kappa = kappa, nugget = 
			nugget.mode, cov.pars = c(1, range.mode), inv = TRUE)
		ivt <- crossprod(invcov.mode, trend.d)
		beta.mode.cond <- solve(crossprod(ivt, trend.d), crossprod(
			ivt, data))
		resid.mode <- as.vector(data - trend.d %*% beta.mode.cond)
		sill.mode.cond <- (resid.mode %*% invcov.mode %*% resid.mode)/
			(n - length(beta.mode.cond) + 2)
		if(beta.size == 1)
			kb.results$posterior$beta.summary <- c(mean = 
				trend.mean, median = trend.median, mode.cond = 
				beta.mode.cond)
		else kb.results$posterior$beta.summary <- cbind(mean = 
				trend.mean, median = trend.median, mode.cond = 
				beta.mode.cond)
		kb.results$posterior$sigmasq.summary <- c(mean = sill.mean,
			median = sill.median, mode.cond = sill.mode.cond)
		kb.results$posterior$phi.summary <- c(mean = range.mean, median
			 = range.median, mode = range.mode)
		if(prior$nugget.prior != "fixed")
			kb.results$posterior$tausq.summary <- c(mean = 
				nugget.mean, median = nugget.median, mode = 
				nugget.mode)
		else kb.results$posterior$tausq.summary <- paste(
				"fixed tausq with value =", nugget)
		kb.results$posterior$beta.samples <- as.matrix(beta)[
			vecpars.back.order,  ]
		beta <- NULL
		if(beta.size == 1)
			kb.results$posterior$beta.samples <- as.vector(
				kb.results$posterior$beta.samples)
		kb.results$posterior$sigmasq.samples <- sigmasq[
			vecpars.back.order]
		sigmasq <- NULL
		kb.results$posterior$phi.samples <- phi.sam[vecpars.back.order,
			1]
		if(prior$nugget.prior != "fixed")
			kb.results$posterior$tausq.samples <- phi.sam[
				vecpars.back.order, 2]
		else kb.results$posterior$tausq.samples <- paste(
				"fixed tausq with value =", nugget)
		phi.lev <- unique(phidist$phi[, 1])
		kb.results$posterior$phi.marginal.prediction <- data.frame(
			phi = phi.lev, expected = apply(phidist$probphi, 1,
			sum), sampled = as.vector(table(factor(phi.sam[, 1],
			levels = phi.lev)))/n.posterior)
		nug.lev <- unique(phidist$phi[, 2])
		if(prior$nugget.prior != "fixed")
			kb.results$posterior$nugget.marginal.prediction <- 
				data.frame(nugget = nug.lev, expected = apply(
				phidist$probphi, 2, sum), sampled = as.vector(
				table(factor(phi.sam[, 2], levels = nug.lev)))/
				n.posterior)
		else kb.results$posterior$nugget.marginal.prediction <- paste(
				"fixed tausq with value =", nugget)
		##
		######################### PART 3.3 ##############################
		## Predictive distribution: sampling and computation
		##
		if(all(locations != "no")) {
			if(messages.screen)
				cat("krige.bayes: Prediction at selected locations\n"
					)
			if(output$mean.estimator != FALSE | !is.null(output$
				probability.estimator) | !is.null(output$
				quantile.estimator))
				prediction.samples <- TRUE
			message.prediction <- character()
			ni <- dim(trend.l)[1]
			if(is.null(n.predictive)) {
				include.it <- FALSE
				n.predictive <- n.posterior
				phi.sam <- phidist$phi[ind,  ]
				message.prediction <- c(message.prediction,
					"Samples of parameters the phi/tausq.rel for the predictive are same as for the posterior"
					)
				if(messages.screen)
					cat("krige.bayes:", message.prediction,
						"\n")
			}
			else {
				include.it <- TRUE
				ind <- sample((1:(dim(phidist$phi)[1])), 
					n.predictive, replace = TRUE, prob = 
					as.vector(phidist$probphi))
				ind.unique <- sort(unique(ind))
				ind.length <- length(ind.unique)
				ind.table <- table(ind)
				phi.unique <- phidist$phi[ind.unique,  ]
				message.prediction <- c(message.prediction,
					"Samples of the parameters phi/tausq.rel for the predictive are NOT the same as for the posterior "
					)
				if(messages.screen) {
					cat("krige.bayes:", message.prediction,
						"\n")
					cat("krige.bayes: samples and their frequencies from the distribution of  phi and tau.rel when drawing from the predictive distribution\n"
						)
					print(rbind(phi = phi.unique[, 1],
						nugget.rel = phi.unique[, 2],
						frequency = ind.table))
				}
				phi.sam <- phidist$phi[ind,  ]
				vecpars.back.order <- order(ind)
			}
			##      d0mat <- apply(locations, 1, d0.krige)
			d0mat <- as.double(rep(0, ni * n))
			.C("loccoords",
				as.double(as.vector(locations[, 1])),
				as.double(as.vector(locations[, 2])),
				as.double(as.vector(coords[, 1])),
				as.double(as.vector(coords[, 2])),
				as.integer(ni),
				as.integer(n),
				d0mat,
				DUP = FALSE)
			attr(d0mat, "dim") <- c(n, ni)
			ind.coincide <- (apply(d0mat < 1e-10, 1, sum) == 1)
			ind.coincide <- (1:ni)[ind.coincide]
			data.coincide <- (apply(d0mat < 1e-10, 2, sum) == 1)
			assign(".kb.pl", list(ind.length = ind.length, phi.val
				 = phi.val, nugget.discrete = nugget.discrete,
				cov.model = cov.model, kappa = kappa, n = n,
				ni = ni, beta.size = beta.size, inv.lower = 
				inv.lower, inv.diag = inv.diag, trend.d = 
				trend.d, data = data, gl = gl, d0mat = d0mat,
				ind.coincide = ind.coincide, data.coincide = 
				data.coincide, trend.l = trend.l, signal = 
				output$signal, messages.screen = 
				messages.screen), where = 1)
			##
			## 3.3.1 Estimating the moments
			##
			if(moments) {
				cat("krige.bayes: computing moments of the predictive distributions\n"
					)
				assign(".temp.ap", 1, where = 1)
				assign(".kb.n.disc", dim(phidist$phi)[1], where
					 = 1)
				temp.res <- apply(phidist$phi, 1, 
					krige.bayes.aux10, moments = TRUE,
					phidist = phidist)
				remove(".temp.ap", where = 1)
				remove(".kb.n.disc", where = 1)
				temp.res <- temp.res %*% as.vector(phidist$
					probphi)
				kb.results$predictive$moments <- data.frame(
					expect.y0 = as.vector(temp.res[1:ni]),
					expect.y0.2 = as.vector(temp.res[(
					ni + 1):(2 * ni)]))
				temp.res <- NULL
				kb.results$predictive$moments$var.y0 <- 
					kb.results$predictive$moments[, 2] -
					(kb.results$predictive$moments[, 1]^
					2)
				if(lambda == 0) {
					temp <- kb.results$predictive$moments$
						expect.y0
					kb.results$predictive$moments$expect.y0 <-
						exp(temp + 0.5 * (kb.results$
						predictive$moments$var.y0))
					kb.results$predictive$moments$var.y0 <-
						(exp(2 * temp + kb.results$
						predictive$moments$var.y0)) *
						(exp(kb.results$predictive$
						moments$var.y0) - 1)
					temp <- NULL
					kb.results$predictive$moments$
						expect.y0.2 <- kb.results$
						predictive$moments$var.y0 +
						(kb.results$predictive$moments$
						expect.y0^2)
				}
			}
			##
			## 3.3.2 Sampling from the predictive
			##
			if(output$simulations.predictive) {
				assign(".temp.ap", 1, where = 1)
				temp.res <- apply(phi.unique, 1, 
					krige.bayes.aux10, moments = FALSE,
					phidist = phidist)
				remove(".temp.ap", where = 1)
				inv.lower <- inv.diag <- NULL
				temp.res <- array(temp.res, dim = c(ni, 2,
					ind.length), dimnames = list(NULL,
					NULL, NULL))
				kb.results$predictive$simulations <- temp.res[
					, 1, rep(1:ind.length, ind.table),
					drop = TRUE] + matrix(rnorm((
					n.predictive * ni)), ncol = 
					n.predictive, nrow = ni, byrow = T) *
					sqrt((temp.res[, 2, rep(1:ind.length,
					ind.table), drop = TRUE] * gl)/matrix(
					rchisq((n.predictive * ni), df = gl),
					ncol = n.predictive, nrow = ni, byrow
					 = T))
				temp.res <- NULL
				if(messages.screen)
					cat("krige.bayes: preparing output of predictive distribution\n"
						)
				##
				## Back transforming
				##
				if(lambda != 1) {
					cat("krige.bayes: Data transformation (Box-Cox) performed. Results returned in the original scale\n"
						)
					if(lambda == 0)
						kb.results$predictive$
							simulations <- exp(
							kb.results$predictive$
							simulations)
					else {
						if(lambda > 0)
							kb.results$predictive$
								simulations[
								kb.results$
								predictive$
								simulations <
								-1/lambda] <-
								-1/lambda
						if(lambda < 0)
							kb.results$predictive$
								simulations[
								kb.results$
								predictive$
								simulations >
								-1/lambda] <-
								-1/lambda
						kb.results$predictive$
							simulations <- ((
							kb.results$predictive$
							simulations * lambda) +
							1)^(1/lambda)
					}
				}
				remove(".kb.pl", where = 1)
				##
				## 3.3.3 mean estimators
				##
				if(output$mean.estimator) {
					kb.results$predictive$mean.simulations <-
						as.vector(apply(kb.results$
						predictive$simulations, 1,
						mean))
					kb.results$predictive$
						variance.simulations <- 
						as.vector(apply(kb.results$
						predictive$simulations, 1,
						var))
					message.prediction <- c(
						message.prediction, 
						"mean at each location in $predictive$mean and predicted variances in $predictive$variances"
						)
				}
				##
				## 3.3.4 quantile estimators
				##
				if(!is.null(output$quantile.estimator)) {
					kb.results$predictive$quantiles <- 
						apply(kb.results$predictive$
						simulations, 1, quantile, probs
						 = output$quantile.estimator)
					if(length(output$quantile.estimator) >
						1) {
						kb.results$predictive$quantiles <-
							as.data.frame(t(
							kb.results$predictive$
							quantiles))
						qname <- rep(0, length(output$
							quantile.estimator))
						for(i in 1:length(output$
							quantile.estimator))
							qname[i] <- paste(
								"q", 100 * 
								output$
								quantile.estimator[
								i], sep = "")
						names(kb.results$predictive$
							quantiles) <- qname
					}
					else {
						kb.results$predictive$quantiles <-
							as.vector(kb.results$
							predictive$quantiles)
					}
					message.prediction <- c(
						message.prediction, 
						"Predicted quantile(s) at each location returned in $predictive$quantiles"
						)
				}
				##
				## 3.3.5 probability estimators
				##
				if(!is.null(output$probability.estimator)) {
					kb.results$predictive$probability <-
						apply(kb.results$predictive$
						simulations, 1, 
						krige.bayes.aux2, cutoff = 
						output$probability.estimator)
					if(length(output$probability.estimator) >
						1) {
						kb.results$predictive$
							probability <- 
							as.data.frame(t(
							kb.results$predictive$
							probability))
						pname <- rep(0, length(output$
							probability.estimator))
						for(i in 1:length(output$
							probability.estimator)
							)
							pname[i] <- paste(
								"cutoff", 
								output$
								probability.estimator[
								i], sep = "")
						names(kb.results$predictive$
							probability) <- pname
					}
					else kb.results$predictive$probability <-
							as.vector(kb.results$
							predictive$probability)
					message.prediction <- c(
						message.prediction, 
						"Estimated probabilities of being less than the provided cutoff(s), at each location, returned in $predictive$probability"
						)
				}
				##
				## samples from  predictive
				##
				if(output$keep.simulations) kb.results$
						predictive$simulations <- 
						kb.results$predictive$
						simulations[, 
						vecpars.back.order] else 
						kb.results$predictive$
						simulations <- NULL
				if(include.it) {
					phi.lev <- unique(phidist$phi[, 1])
					kb.results$predictive$
						phi.marginal.prediction <- 
						data.frame(phi = phi.lev, 
						expected = apply(phidist$
						probphi, 1, sum), sampled = 
						as.vector(table(factor(phi.sam[
						, 1], levels = phi.lev)))/
						n.predictive)
					nug.lev <- unique(phidist$phi[, 2])
					if(prior$nugget.prior != "fixed")
						data.frame(nugget = nug.lev,
							expected = apply(
							phidist$probphi, 2,
							sum), sampled = 
							as.vector(table(factor(
							phi.sam[, 2], levels = 
							nug.lev)))/n.predictive
							)
					else kb.results$predictive$
							nugget.marginal.prediction <-
							paste(
							"fixed tausq with value =",
							nugget)
					kb.results$predictive$
						nugget.marginal.prediction <-
						data.frame(nugget = nug.lev,
						expected = apply(phidist$
						probphi, 2, sum), sampled = 
						as.vector(table(factor(phi.sam[
						, 2], levels = nug.lev)))/
						n.predictive)
				}
			}
			kb.results$predictive$type.prediction <- 
				"Spatial prediction performed taking into account uncertainty in trend (mean), sill and range parameters"
			kb.results$message.prediction <- message.prediction
		}
		else {
			kb.results$predictive <- 
				"no locations to perform prediction were provided"
			kb.results$message.prediction <- 
				"Only Bayesian estimation of model parameters were computed"
			if(messages.screen)
				print(kb.results$message.prediction)
		}
	}
	if(all(nugget != 0) & prior$nugget.prior == "fixed")
		kb.results$nugget.fixed <- nugget
	kb.results$.Random.seed <- seed
	kb.results$call <- call
	if(version$major <= 4)
		class(kb.results) <- setOldClass(c("krige.bayes", "kriging"))
	else class(kb.results) <- c("krige.bayes", "kriging")
	if(messages.screen)
		cat("krige.bayes: done!\n")
	return(kb.results)
}

"krige.bayes.aux1" <- 
function(phinug, coords, trend.d, data, cov.model, kappa, range.prior, 
	exponential.prior.par, n)
{
	# auxiliary function to perform calculation for the function krige.bayes
	phi <- phinug[1]
	nugget <- phinug[2]
	# check == here
	if(phi == 0) covphi <- list(inverse = diag((1/(1 + nugget)), n), 
			log.det.to.half = ((n/2) * log((1 + nugget)))) else 
			covphi <- varcov.spatial(coords = coords, cov.model = 
			cov.model, kappa = kappa, nugget = nugget, cov.pars = c(
			1, phi), inv = T, det = T)
	ttiv <- crossprod(trend.d, covphi$inverse)
	ttivtt <- ttiv %*% trend.d
	vbetaphi <- solve(ttivtt)
	eival.x <- eigen(ttivtt, symmetric = T, only.values = T)
	ttivttdet <- prod(eival.x$values)
	betaphi <- as.vector(vbetaphi %*% ttiv %*% data)
	res <- data - trend.d %*% betaphi
	sqres <- as.vector(crossprod(res, covphi$inverse) %*% res)
	gl <- length(data) - length(betaphi)
	s2phi <- (1/gl) * sqres
	logprobphi <- (-0.5) * log(ttivttdet) - (covphi$log.det.to.half) - (
		gl/2) * log(s2phi)
	if(range.prior == "reciprocal" & round(100000000. * phi) != 0)
		logprobphi <- logprobphi - log(phi)
	if(range.prior == "squared.reciprocal" & round(100000000. * phi) != 0)
		logprobphi <- logprobphi - log(phi^2)
	if(range.prior == "exponential")
		logprobphi <- logprobphi - exponential.prior.par * phi
	#  probphi <- ttivttdet^(-0.5) * (1/exp(covphi$log.det.to.half)) * s2phi^
	#  ( - gl/2)
	#  if(range.prior == "reciprocal")
	#    probphi <- probphi/phi
	#  if(range.prior == "squared.reciprocal")
	#    probphi <- probphi/(phi^2)
	#  if(range.prior == "exponential")
	#    probphi <- probphi * exp( - exponential.prior.par * phi)
	inv.lower <- covphi$inverse[lower.tri(covphi$inverse)]
	inv.diag <- diag(covphi$inverse)
	results <- list(vbetaphi = vbetaphi, betaphi = betaphi, s2phi = s2phi,
		logprobphi = logprobphi, inv.lower = inv.lower, inv.diag = 
		inv.diag)
	return(results)
}

"krige.bayes.aux10" <- 
function(phinug, moments, phidist)
{
	ni <- .kb.pl$ni
	n <- .kb.pl$n
	beta.size <- .kb.pl$beta.size
	if(.kb.pl$messages.screen)
		krige.bayes.messages(moments = moments, n.disc = .kb.n.disc,
			.temp.ap = .temp.ap, ind.length = .kb.pl$ind.length)
	phinug <- as.vector(phinug)
	phi.ind <- order(.kb.pl$phi.val)[round(100000000. * .kb.pl$phi.val) ==
		round(100000000. * phinug[1])]
	nug.ind <- order(.kb.pl$nugget.discrete)[round(100000000. * .kb.pl$
		nugget.discrete) == round(100000000. * phinug[2])]
	v0 <- cov.spatial(obj = .kb.pl$d0mat, cov.model = .kb.pl$cov.model,
		kappa = .kb.pl$kappa, cov.pars = c(1, phinug[1]))
	tb <- as.double(rep(0, ni * beta.size))
	tb <- .C("quadraticform_XAY",
		as.double(as.vector(.kb.pl$inv.lower[phi.ind, nug.ind,  ])),

			as.double(as.vector(.kb.pl$inv.diag[phi.ind, nug.ind,
			])),
		as.double(as.vector(v0)),
		as.double(as.vector(.kb.pl$trend.d)),
		as.integer(ni),
		as.integer(beta.size),
		as.integer(n),
		res = tb)$res
	attr(tb, "dim") <- c(ni, beta.size)
	tb <- .kb.pl$trend.l - tb
	tv0ivdata <- as.double(rep(0, ni))
	tv0ivdata <- .C("quadraticform_XAY",
		as.double(as.vector(.kb.pl$inv.lower[phi.ind, nug.ind,  ])),

			as.double(as.vector(.kb.pl$inv.diag[phi.ind, nug.ind,
			])),
		as.double(as.vector(v0)),
		as.double(.kb.pl$data),
		as.integer(ni),
		as.integer(1),
		as.integer(n),
		res = tv0ivdata)$res
	tmean <- tv0ivdata + tb %*% as.vector(phidist$beta[phi.ind, nug.ind,
		])
	tv0ivdata <- NULL
	if(round(100000000. * phinug[2]) == 0)
		tmean[.kb.pl$ind.coincide] <- .kb.pl$data[.kb.pl$data.coincide]
	tv0ivv0 <- as.double(rep(0, ni))
	tv0ivv0 <- .C("quadraticform_XAX",
		as.double(as.vector(.kb.pl$inv.lower[phi.ind, nug.ind,  ])),

			as.double(as.vector(.kb.pl$inv.diag[phi.ind, nug.ind,
			])),
		as.double(as.vector(v0)),
		as.integer(ni),
		as.integer(n),
		res = tv0ivv0)$res
	v0 <- NULL
	vbetai <- matrix(as.vector(phidist$vbeta.vector[phi.ind, nug.ind,  ]),
		ncol = beta.size, nrow = beta.size)
	if(beta.size == 1)
		tbivbb <- ((as.vector(tb))^2) * vbetai
	else {
		tbivbb <- as.double(rep(0, ni))
		tbivbb <- .C("quadraticform_XAX",
			as.double(vbetai[lower.tri(vbetai)]),
			as.double(diag(vbetai)),
			as.double(as.vector(t(tb))),
			as.integer(ni),
			as.integer(beta.size),
			res = tbivbb)$res
	}
	s2i <- phidist$s2[phi.ind, nug.ind]
	if(.kb.pl$signal)
		tvar <- s2i * (1 - tv0ivv0 + tbivbb)
	else tvar <- s2i * ((1 + phinug[2]) - tv0ivv0 + tbivbb)
	tb <- tbivbb <- tv0ivv0 <- NULL
	if(round(100000000. * phinug[2]) == 0)
		tvar[.kb.pl$ind.coincide] <- 0
	if(moments) {
		## take care here, re-using object!
		tvar <- (.kb.pl$gl/(.kb.pl$gl - 2)) * tvar
		tvar <- tvar + (tmean)^2
	}
	tvar[tvar < 9.9999999999999998e-17] <- 0
	.temp.ap <<- .temp.ap + 1
	## if moments=T tvar is expc.y0.2 !!!
	return(cbind(tmean, tvar))
}

"krige.bayes.aux11" <- 
#
function(i, mat1, mat2, by)
{
	if(by == "row")
		return(mat1[i,  ] %*% mat2[i,  ])
	if(by == "col")
		return(mat1[, i] %*% mat2[, i])
}

"krige.bayes.aux12" <- 
function(phinug, d0mat, ind.coincide, data.coincide, pred.list, ni, trend.l,
	signal, messages.screen, n.disc)
{
	n <- pred.list$n
	phidist <- pred.list$phidist
	phinug <- as.vector(phinug)
	phi.ind <- order(pred.list$phi.val)[round(100000000. * pred.list$
		phi.val) == round(100000000. * phinug[1])]
	nug.ind <- order(pred.list$nugget)[round(100000000. * pred.list$nugget) ==
		round(100000000. * phinug[2])]
	iv <- matrix(0, n, n)
	iv[lower.tri(iv)] <- as.vector(phidist$inv.lower[phi.ind, nug.ind,
		])
	iv <- iv + t(iv)
	diag(iv) <- as.vector(phidist$inv.diag[phi.ind, nug.ind,  ])
	v0 <- cov.spatial(obj = d0mat, cov.model = pred.list$cov.model, kappa
		 = pred.list$kappa, cov.pars = c(1, phinug[1]))
	ivv0 <- iv %*% v0
	iv <- NULL
	tb <- trend.l - crossprod(ivv0, pred.list$trend.d)
	expec.y0 <- crossprod(ivv0, pred.list$data) + tb %*% as.vector(phidist$
		beta[phi.ind, nug.ind,  ])
	if(round(100000000. * phinug[2]) == 0)
		expec.y0[ind.coincide] <- pred.list$data[data.coincide]
	s2i <- phidist$s2[phi.ind, nug.ind]
	vbetai <- matrix(as.vector(phidist$vbeta.vector[phi.ind, nug.ind,  ]),
		ncol = pred.list$beta.size, nrow = pred.list$beta.size)
	tv0ivv0 <- apply(as.matrix(1:ni), 1, krige.bayes.aux11, mat1 = v0,
		mat2 = ivv0, by = "col")
	tbivb <- tb %*% vbetai
	tbivbb <- apply(as.matrix(1:ni), 1, krige.bayes.aux11, mat1 = tb, mat2
		 = tbivb, by = "row")
	if(signal == T)
		tvar <- s2i * (1 - tv0ivv0 + tbivbb)
	else tvar <- s2i * ((1 + phinug[2]) - tv0ivv0 + tbivbb)
	tvar <- (pred.list$gl/(pred.list$gl - 2)) * tvar
	expec.y0.2 <- tvar + (expec.y0)^2
	if(round(100000000. * phinug[2]) == 0)
		tvar[ind.coincide] <- 0
	tv0ivv0 <- NULL
	tbivbb <- NULL
	tbivb <- NULL
	if(messages.screen == T) {
		if(n.disc <= 20)
			print(paste("prediction for the parameter set", 
				.temp.ap1, "out of", n.disc))
		if(n.disc > 20 & n.disc <= 200)
			if(.temp.ap1 %% 10 == 1)
				print(paste("prediction for the parameter set",
					.temp.ap1, "out of", n.disc))
		if(n.disc > 200)
			if(.temp.ap1 %% 100 == 1)
				print(paste("prediction for the parameter set",
					.temp.ap1, "out of", n.disc))
	}
	assign(".temp.ap1", (.temp.ap1 + 1), where = 1)
	return(cbind(expec.y0, expec.y0.2))
}

"krige.bayes.aux2" <- 
function(x, cutoff)
{
	# auxiliary function to perform calculation for the function krige.bayes
	ncut <- length(cutoff)
	lx <- length(x)
	if(ncut > 1) {
		result <- rep(0, ncut)
		for(i in 1:ncut)
			result[i] <- sum(x <= cutoff[i])/lx
	}
	else result <- sum(x <= cutoff)/lx
	return(result)
}

"krige.bayes.aux3" <- 
##
# This function produces a sample from  a multivariate normal distribution 
# mean is 0 and cov is a vector of length beta.size^2
function(cov, beta.size)
{
	#  dm <- length(mean)
	#  dc <- dim(cov)[1]
	#  if(dm != dc)
	#    stop("mean and cov must have compatible dimensions")
	cov <- matrix(cov, ncol = beta.size)
	cov.svd <- svd(cov)
	cov.decomp <- cov.svd$u %*% (t(cov.svd$u) * sqrt(cov.svd$d))
	zsim <- as.vector(cov.decomp %*% rnorm(beta.size))
	return(zsim)
}

"krige.bayes.aux5" <- 
#
function(phi.s, d0, cov.model, kappa, trend.l, trend.d, data, beta.size, n,
	l)
{
	iv <- matrix(0, n, n)
	# check == here
	iv[lower.tri(iv)] <- as.vector(temp.phidist$inv.lower[temp.phidist$
		phi == phi.s,  ])
	iv <- iv + t(iv)
	# check == here
	diag(iv) <- as.vector(temp.phidist$inv.diag[temp.phidist$phi == phi.s,
		])
	v0 <- cov.spatial(obj = d0, cov.model = cov.model, kappa = kappa, 
		cov.pars = c(1, phi.s))
	tv0 <- t(v0)
	tv0iv <- tv0 %*% iv
	term1 <- tv0iv %*% data
	tb <- as.matrix(trend.l[l,  ] - tv0iv %*% trend.d)
	ttb <- t(tb)
	# check == here
	tmean <- term1 + ttb %*% as.vector(temp.phidist$beta[temp.phidist$
		phi == phi.s,  ])
	# check == here
	s2i <- as.vector(temp.phidist$s2[temp.phidist$phi == phi.s])
	# check == here
	vbetai <- matrix(as.vector(temp.phidist$vbeta.vector[temp.phidist$
		phi == phi.s,  ]), ncol = beta.size, nrow = beta.size)
	tvar <- s2i * (1 - tv0iv %*% v0 + ttb %*% vbetai %*% tb)
	return(list(tmean = tmean, tvar = tvar))
}

"krige.bayes.aux6" <- 
#
function(phi.s, d0, cov.model, kappa, trend.l, trend.d, data, beta.size, n,
	l, inv.lower, inv.diag, beta, s2, vbeta.vector)
{
	iv <- matrix(0, n, n)
	iv[lower.tri(iv)] <- inv.lower
	iv <- iv + t(iv)
	diag(iv) <- inv.diag
	v0 <- cov.spatial(obj = d0, cov.model = cov.model, kappa = kappa, 
		cov.pars = c(1, phi.s))
	tv0 <- t(v0)
	tv0iv <- tv0 %*% iv
	term1 <- tv0iv %*% data
	tb <- as.matrix(trend.l[l,  ] - tv0iv %*% trend.d)
	ttb <- t(tb)
	tmean <- term1 + ttb %*% as.vector(beta)
	s2i <- s2
	vbetai <- matrix(vbeta.vector, ncol = beta.size, nrow = beta.size)
	tvar <- s2i * (1 - tv0iv %*% v0 + ttb %*% vbetai %*% tb)
	return(list(tmean = tmean, tvar = tvar))
}

"krige.bayes.messages" <- 
function(moments, n.disc, .temp.ap, ind.length)
{
	if(moments) {
		if(n.disc <= 20)
			cat(paste("computing moments: point", .temp.ap, 
				"out of", n.disc, "\n"))
		if(n.disc > 20 & n.disc <= 200)
			if(.temp.ap %% 10 == 1)
				cat(paste("computing moments: point", .temp.ap,
					"out of", n.disc, "\n"))
		if(n.disc > 200)
			if(.temp.ap %% 100 == 1)
				cat(paste("computing moments: point", .temp.ap,
					"out of", n.disc, "\n"))
	}
	else {
		if(ind.length <= 20)
			cat(paste("prediction for the parameter set", .temp.ap,
				"out of", ind.length, "\n"))
		if(ind.length > 20 & ind.length <= 200)
			if(.temp.ap %% 10 == 1)
				cat(paste("prediction for the parameter set",
					.temp.ap, "out of", ind.length, "\n"))
		if(ind.length > 200)
			if(.temp.ap %% 100 == 1)
				cat(paste("prediction for the parameter set",
					.temp.ap, "out of", ind.length, "\n"))
	}
}

"krige.bayes.predict" <- 
#
function(locations, trend.l, info, signal, n.predictive = "n.posterior", 
	messages.screen = T, save.call = T, moments, distribution.keep, 
	mean.estimator, quantile.estimator, probability.estimator)
{
	if(save.call == T)
		call <- match.call()
	if(missing(distribution.keep))
		distribution <- info$distribution
	if(missing(mean.estimator))
		mean.estimator <- info$mean.estimator
	if(missing(quantile.estimator))
		quantile.estimator <- info$quantile.estimator
	if(missing(probability.estimator))
		probability.estimator <- info$probability.estimator
	n <- info$n
	locations <- as.matrix(locations)
	dimnames(locations) <- list(NULL, NULL)
	trend.l <- trend.spatial(trend = trend.l, coords = locations)
	if(messages.screen == T) {
		if(inherits(trend.l, "formula"))
			print("prediction to be performed including (external) covariate(s) provided by the user"
				)
		else {
			if(length(trend.l) == 1) {
				if(trend.l == "cte")
					print("prediction to be performed with constant mean"
						)
				if(trend.l == "1st")
					print("prediction to be performed with 1st degree polinomial trend"
						)
				if(trend.l == "2nd")
					print("prediction to be performed with 2nd degree polinomial"
						)
			}
		}
	}
	ni <- dim(trend.l)[1]
	if(n.predictive == "n.posterior") {
		n.predictive <- info$n.posterior
		phi.sam <- info$phidist$phi[info$ind,  ]
		if(messages.screen == T)
			print("samples of parameters for the predictive are the same as for the posterior"
				)
	}
	else {
		ind <- sample((1:(dim(info$phidist$phi)[1])), n.predictive,
			replace = T, prob = as.vector(info$phidist$probphi))
		ind.unique <- sort(unique(ind))
		info$ind.length <- length(ind.unique)
		info$ind.table <- table(ind)
		info$phi.unique <- info$phidist$phi[ind.unique,  ]
		if(messages.screen == T) {
			print("the sampled values from the phi/tau.rel distribution and their frequencies are:"
				)
			print(rbind(phi = info$phi.unique[, 1], nugget.rel = 
				info$phi.unique[, 2], frequency = info$
				ind.table))
		}
		phi.sam <- info$phidist$phi[ind,  ]
		info$vecpars.back.order <- order(ind)
	}
	d0mat <- apply(locations, 1, dist0.krige, info$coords)
	ind.coincide <- (apply(d0mat < 1e-10, 2, sum) == 1)
	ind.coincide <- (1:ni)[ind.coincide]
	data.coincide <- (apply(d0mat < 1e-10, 1, sum) == 1)
	assign(".temp.ap", 1, where = 1)
	temp.res <- apply(info$phi.unique, 1, krige.bayes.aux10, d0mat = d0mat,
		ind.coincide = ind.coincide, data.coincide = data.coincide,
		ni = ni, trend.l = trend.l, signal = signal, pred.list = info,
		messages.screen = messages.screen)
	remove(".temp.ap", where = 1)
	tmean <- t(temp.res[1:ni,  ])
	tsdv <- sqrt(t(temp.res[ - (1:ni),  ]))
	tmean <- tmean[rep(1:info$ind.length, info$ind.table),  ]
	tsdv <- tsdv[rep(1:info$ind.length, info$ind.table),  ]
	results.pred <- list()
	results.pred$dist.pred <- tmean + (tsdv * matrix(rnorm((n.predictive *
		ni)), nrow = n.predictive, ncol = ni) * sqrt(info$gl))/sqrt(
		matrix(rchisq((n.predictive * ni), df = info$gl), nrow = 
		n.predictive, ncol = ni))
	#
	# Preparing output for posterior distributions
	#
	if(messages.screen == T) print(
			"preparing output of predictive distribution")
	#
	# Back transforming
	#
	if(info$lambda != 1) {
		print("Data transformation (Box-Cox) performed. Results returned in the original scale"
			)
		if(info$lambda == 0)
			results.pred$dist.pred <- exp(results.pred$dist.pred)
		else {
			if(info$lambda > 0)
				results.pred$dist.pred[results.pred$dist.pred <
					-1/info$lambda] <- (-1/info$lambda)
			if(info$lambda < 0)
				results.pred$dist.pred[results.pred$dist.pred >
					-1/info$lambda] <- (-1/info$lambda)
			results.pred$dist.pred <- ((results.pred$dist.pred *
				info$lambda) + 1)^(1/info$lambda)
		}
	}
	#
	if(mean.estimator == T) {
		results.pred$predict.mean <- as.vector(apply(results.pred$
			dist.pred, 2, mean))
		results.pred$krige.var <- as.vector(apply(results.pred$
			dist.pred, 2, var))
		message.prediction <- 
			"predicted mean at each location in $predict and predicted variances in $krige.var"
	}
	if(all(quantile.estimator != F)) {
		results.pred$quantiles <- apply(results.pred$dist.pred, 2,
			quantile, probs = quantile.estimator)
		if(length(quantile.estimator) > 1) {
			results.pred$quantiles <- t(results.pred$quantiles)
			qname <- rep(0, length(quantile.estimator))
			for(i in 1:length(quantile.estimator))
				qname[i] <- paste("q", 100 * quantile.estimator[
					i], sep = "")
			dimnames(results.pred$quantiles) <- list(NULL, qname)
		}
		else {
			results.pred$quantiles <- as.vector(results.pred$
				quantiles)
		}
		message.prediction <- 
			"Predicted quantile(s) at each location returned in $quantiles"
	}
	if(probability.estimator != F) {
		results.pred$probability <- apply(results.pred$dist.pred, 2,
			krige.bayes.aux2, cutoff = probability.estimator)
		if(length(probability.estimator) > 1)
			results.pred$probability <- t(results.pred$probability)
		message.prediction <- 
			"Estimated probabilities of being less than the provided cutoff(s), at each location, returned in $probability"
	}
	if(distribution == T)
		results.pred$dist.pred <- results.pred$dist.pred[info$
			vecpars.back.order,  ]
	else results.pred$dist.pred <- NULL
	results.pred$type.kriging <- c(
		"Spatial prediction performed taking into account uncertainty in trend (mean), sill and range parameters"
		)
	results.pred$message.prediction <- message.prediction
	phi.lev <- unique(info$phidist$phi[, 1])
	results.pred$phi.marginal.prediction <- data.frame(phi = phi.lev, 
		expected = apply(info$phidist$probphi, 1, sum), sampled = 
		as.vector(table(category(phi.sam[, 1], levels = phi.lev)))/
		n.predictive)
	nug.lev <- unique(info$phidist$phi[, 2])
	results.pred$nugget.marginal.prediction <- data.frame(nugget = nug.lev,
		expected = apply(info$phidist$probphi, 2, sum), sampled = 
		as.vector(table(category(phi.sam[, 2], levels = nug.lev)))/
		n.predictive)
	results.pred$call <- call
	#
	# Estimating the moments
	#
	if(missing(moments)) moments <- info$moments
	if(moments == T) {
		print("computing moments of the predictive distributions")
		assign(".temp.ap1", 1, where = 1)
		n.disc <- dim(info$phidist$phi)[1]
		temp.res <- apply(info$phidist$phi, 1, krige.bayes.aux12, d0mat
			 = d0mat, ind.coincide = ind.coincide, data.coincide = 
			data.coincide, ni = ni, trend.l = trend.l, signal = 
			signal, pred.list = info, messages.screen = 
			messages.screen, n.disc = n.disc)
		remove(".temp.ap1", where = 1)
		n.disc <- length(info$phidist$probphi)
		temp.res <- temp.res %*% info$phidist$probphi
		#    print(dim(temp.res))
		expec.y0 <- as.vector(temp.res[1:ni])
		expec.y0.2 <- as.vector(temp.res[(ni + 1):(2 * ni)])
		var.y0 <- expec.y0.2 - (expec.y0)^2
		if(info$lambda == 0) {
			temp <- expec.y0
			expec.y0 <- exp(temp + 0.5 * (var.y0))
			var.y0 <- (exp(2 * temp + var.y0)) * (exp(var.y0) -
				1)
			expec.y0.2 <- var.y0 + (expec.y0)^2
		}
		results.pred$moments <- cbind(expect.y0 = expec.y0, expect.y0.2
			 = expec.y0.2, var.y0 = var.y0)
	}
	return(results.pred)
}

"krige.control" <- 
function(type.krige = "ok", beta = NULL, trend.d = "cte", trend.l = "cte",
	cov.model = c("exponential", "matern", "gaussian", "spherical", 
	"circular", "cubic", "wave", "powered.exponential", "cauchy", 
	"gneiting", "gneiting.matern", "pure.nugget"), cov.pars = stop(
	"covariance parameters (sigmasq and phi) should be provided"), kappa = 
	0.5, nugget = 0, micro.scale = 0, dist.epsilon = 1e-10, aniso.pars = 
	NULL, lambda = 1, signal = FALSE, n.samples.backtransform = 500, n.sim
	 = 0)
{
	cov.model <- match.arg(cov.model)
	return(list(type.krige = type.krige, beta = beta, trend.d = trend.d,
		trend.l = trend.l, cov.model = cov.model, cov.pars = cov.pars,
		kappa = kappa, nugget = nugget, micro.scale = micro.scale,
		dist.epsilon = dist.epsilon, aniso.pars = aniso.pars, lambda = 
		lambda, signal = signal, n.samples.backtransform = 
		n.samples.backtransform, n.sim = n.sim))
}

"krige.conv" <- 
function(geodata, coords = geodata$coords, data = geodata$data, locations,
	krige = krige.control(type.krige, beta = NULL, trend.d, trend.l, 
	cov.model, cov.pars, kappa = 0.5, nugget = 0, micro.scale = 0, 
	dist.epsilon = 1e-10, aniso.pars = NULL, lambda = 1, signal = FALSE,
	n.samples.backtransform = 500, n.sim = 0), messages.screen = TRUE,
	...)
{
	call <- match.call()
	##
	## reading input
	##
	cov.model <- krige$cov.model
	kappa <- krige$kappa
	lambda <- krige$lambda
	##
	beta <- krige$beta
	cov.pars <- krige$cov.pars
	nugget <- krige$nugget
	##
	signal <- krige$signal
	n.sim <- krige$n.sim
	n.samples.backtransform <- krige$n.samples.backtransform
	##
	## checking input
	##
	if(krige$micro.scale > nugget) stop(
			"krige$micro.scale must be in the interval [0, nugget]"
			)
	if(krige$type.krige == "ok" | krige$type.krige == "OK")
		beta.prior <- "flat"
	if(krige$type.krige == "sk" | krige$type.krige == "SK") {
		if(missing(beta))
			stop("argument beta must be provided in order to perform simple kriging"
				)
		beta.prior <- "deg"
	}
	##
	if(is.vector(coords)) {
		coords <- cbind(coords, 0)
		warning("vector of coordinates: one spatial dimention assumed")
	}
	coords <- as.matrix(coords)
	if(is.vector(locations)) {
		if(length(locations) == 2) {
			locations <- t(as.matrix(locations))
			if(messages.screen)
				warning("it was assumed that locations has coordinates of 1 prediction point"
					)
		}
		else {
			warning("vector of locations: one spatial dimention assumed"
				)
			locations <- as.matrix(cbind(locations, 0))
		}
	}
	else locations <- as.matrix(locations)
	dimnames(coords) <- list(NULL, NULL)
	dimnames(locations) <- list(NULL, NULL)
	##
	if(inherits(krige$trend.d, "formula") | inherits(krige$trend.l, 
		"formula")) {
		if(any(locations != "no"))
			if((inherits(krige$trend.d, "formula") == FALSE) | (
				inherits(krige$trend.l, "formula") == FALSE))
				stop("krige$trend.d and krige$trend.l must have similar specification"
					)
		if(messages.screen)
			print("Kriging with external trend to be performed using covariates provided by the user"
				)
	}
	else {
		if(any(locations != "no") & (krige$trend.d != krige$trend.l)) 
			{
			stop("krige$trend.l is different from krige$trend.d")
		}
		if(messages.screen) {
			if(krige$trend.d == "cte" & krige$type.krige == "sk")
				cat("krige.conv: Simple kriging to be performed with constant mean provided by the user\n"
					)
			if(krige$trend.d == "cte" & krige$type.krige == "ok")
				cat("krige.conv: Ordinary kriging to be performed filtering a constant mean\n"
					)
			if(krige$trend.d == "1st")
				cat("krige.conv: Trend (or universal) kriging to be performed filtering a 1st degree polinomial trend\n"
					)
			if(krige$trend.d == "2nd")
				cat("krige.conv: Trend (or universal) kriging to be performed filtering a 2nd degree polinomial trend\n"
					)
		}
	}
	trend.d <- trend.spatial(trend = krige$trend.d, coords = coords)
	beta.size <- ncol(trend.d)
	trend.l <- trend.spatial(trend = krige$trend.l, coords = locations)
	##
	## Anisotropy correction (should be placed AFTER trend.d/trend.l
	##
	if(!is.null(krige$aniso.pars)) {
		if(length(krige$aniso.pars) != 2 | !is.numeric(krige$aniso.pars
			))
			stop("anisotropy parameters must be provided as a numeric vector with two elements: the rotation angle (in radians) and the anisotropy ratio (a number greater than 1)"
				)
		cat("krige.conv: anisotropy correction performed\n")
		coords <- coords.aniso(coords = coords, aniso.pars = krige$
			aniso.pars)
		locations <- coords.aniso(coords = locations, aniso.pars = 
			krige$aniso.pars)
	}
	##
	## Box-Cox transformation
	##
	if(lambda != 1) {
		if(messages.screen)
			print("Data transformation (Box-Cox) performed.")
		if(lambda == 0)
			data <- log(data)
		else data <- ((data^lambda) - 1)/lambda
	}
	## 
	## setting covariance parameters
	##
	tausq <- nugget
	if(is.vector(cov.pars)) {
		sigmasq <- cov.pars[1]
		phi <- cov.pars[2]
	}
	else {
		sigmasq <- cov.pars[, 1]
		phi <- cov.pars[, 2]
	}
	sill.partial <- sum(sigmasq)
	sill.tot <- tausq + sill.partial
	n <- length(data)
	ni <- nrow(trend.l)
	##
	## starting kriging calculations
	##
	kc.result <- list()
	invcov <- varcov.spatial(coords = coords, cov.model = cov.model, kappa
		 = kappa, nugget = nugget, cov.pars = cov.pars, inv = TRUE,
		only.inv.lower.diag = TRUE)
	temp <- as.double(rep(0, beta.size * beta.size))
	temp <- .C("quadraticform_XAY",
		as.double(invcov$lower.inverse),
		as.double(invcov$diag.inverse),
		as.double(as.vector(trend.d)),
		as.double(as.vector(trend.d)),
		as.integer(beta.size),
		as.integer(beta.size),
		as.integer(n),
		res = temp)$res
	attr(temp, "dim") <- c(beta.size, beta.size)
	ittivtt <- solve(temp)
	temp <- NULL
	if(beta.prior == "flat") {
		temp <- as.double(rep(0, beta.size))
		temp <- .C("quadraticform_XAY",
			as.double(invcov$lower.inverse),
			as.double(invcov$diag.inverse),
			as.double(as.vector(trend.d)),
			as.double(as.vector(data)),
			as.integer(beta.size),
			as.integer(1),
			as.integer(n),
			res = temp)$res
		beta.flat <- ittivtt %*% temp
		temp <- NULL
	}
	v0mat <- as.double(rep(0, n * ni))
	.C("loccoords",
		as.double(as.vector(locations[, 1])),
		as.double(as.vector(locations[, 2])),
		as.double(as.vector(coords[, 1])),
		as.double(as.vector(coords[, 2])),
		as.integer(ni),
		as.integer(n),
		v0mat,
		DUP = FALSE)
	attr(v0mat, "dim") <- c(n, ni)
	if(n.sim == 0)
		locations <- NULL
	if(krige$micro.scale != 0) {
		v0mat <- ifelse(v0mat < krige$dist.epsilon, krige$micro.scale +
			sill.partial, cov.spatial(obj = v0mat, cov.model = 
			cov.model, kappa = kappa, cov.pars = cov.pars))
	}
	else {
		v0mat <- cov.spatial(obj = v0mat, cov.model = cov.model, kappa
			 = kappa, cov.pars = cov.pars)
	}
	if(n.sim > 0) {
		reduce.var <- as.double(rep(0, ni * ni))
		.C("quadraticform_XAY",
			as.double(invcov$lower.inverse),
			as.double(invcov$diag.inverse),
			as.double(as.vector(v0mat)),
			as.double(as.vector(v0mat)),
			as.integer(ni),
			as.integer(ni),
			as.integer(n),
			reduce.var,
			DUP = FALSE)
		attr(reduce.var, "dim") <- c(ni, ni)
		tv0ivv0 <- diag(reduce.var)
	}
	else {
		tv0ivv0 <- as.double(rep(0, ni))
		tv0ivv0 <- .C("quadraticform_XAX",
			as.double(invcov$lower.inverse),
			as.double(invcov$diag.inverse),
			as.double(as.vector(v0mat)),
			as.integer(ni),
			as.integer(n),
			res = tv0ivv0)$res
	}
	tb <- as.double(rep(0, ni * beta.size))
	tb <- .C("quadraticform_XAY",
		as.double(invcov$lower.inverse),
		as.double(invcov$diag.inverse),
		as.double(as.vector(v0mat)),
		as.double(as.vector(trend.d)),
		as.integer(ni),
		as.integer(beta.size),
		as.integer(n),
		res = tb)$res
	attr(tb, "dim") <- c(ni, beta.size)
	tb <- trend.l - tb
	if(beta.prior == "deg") {
		tv0ivdata <- as.double(rep(0, ni))
		tv0ivdata <- .C("quadraticform_XAY",
			as.double(invcov$lower.inverse),
			as.double(invcov$diag.inverse),
			as.double(as.vector(v0mat)),
			as.double(data),
			as.integer(ni),
			as.integer(1),
			as.integer(n),
			res = tv0ivdata)$res
		remove("v0mat", "invcov")
		kc.result$predict <- tv0ivdata + as.vector(tb %*% beta)
		remove("tb", "tv0ivdata")
		if(krige.signal)
			kc.result$krige.var <- as.vector(sill.partial - tv0ivv0
				)
		else kc.result$krige.var <- as.vector(sill.tot - tv0ivv0)
		beta.est <- cat(
			"krige.conv: Simple kriging (beta provided by user)")
	}
	if(beta.prior == "flat") {
		tv0ivdata <- as.double(rep(0, ni))
		tv0ivdata <- .C("quadraticform_XAY",
			as.double(invcov$lower.inverse),
			as.double(invcov$diag.inverse),
			as.double(as.vector(v0mat)),
			as.double(data),
			as.integer(ni),
			as.integer(1),
			as.integer(n),
			res = tv0ivdata)$res
		v0mat <- invcov <- NULL
		kc.result$predict <- tv0ivdata + as.vector(tb %*% beta.flat)
		tv0ivdata <- NULL
		if(n.sim > 0) {
			if(beta.size == 1)
				bitb <- outer(tb, tb) * as.vector(ittivtt)
			else {
				b <- t(tb)
				tb <- NULL
				ok.add.var <- as.double(rep(0, ni * ni))
				.C("quadraticform_XAY",
					as.double(ittivtt[lower.tri(ittivtt)]),

						as.double(diag(ittivtt)),
					as.double(as.vector(b)),
					as.double(as.vector(b)),
					as.integer(ni),
					as.integer(ni),
					as.integer(beta.size),
					ok.add.var,
					DUP = FALSE)
			}
			b <- NULL
			reduce.var <- reduce.var + ok.add.var
			bitb <- diag(ok.add.var)
			ok.add.var <- NULL
		}
		else {
			if(beta.size == 1)
				bitb <- as.vector(tb^2) * as.vector(ittivtt)
			else {
				bitb <- as.double(rep(0, ni))
				bitb <- .C("quadraticform_XAX",
					as.double(ittivtt[lower.tri(ittivtt)]),

						as.double(diag(ittivtt)),
					as.double(as.vector(t(tb))),
					as.integer(ni),
					as.integer(beta.size),
					res = bitb)$res
			}
		}
		if(krige$signal)
			kc.result$krige.var <- as.vector(sill.partial - tv0ivv0 +
				bitb)
		else kc.result$krige.var <- as.vector(sill.tot - tv0ivv0 + bitb
				)
		kc.result$beta.est <- beta.flat
	}
	bitb <- tv0ivv0 <- NULL
	if(any(kc.result$krige.var < 0))
		cat("krige.conv: negative kriging variance found! Investigate why this is happening.\n"
			)
	message <- "krige.conv: Kriging performed using global neighbourhood"
	if(messages.screen)
		cat(paste(message, "\n"))
	if(lambda != 1) {
		if(messages.screen)
			cat("krige.conv: back-transforming the predictions according to the (Box-Cox) parameter lambda\n"
				)
		if(lambda == 0) {
			predict.transf <- kc.result$predict
			kc.result$predict <- exp(predict.transf) - 0.5 * 
				kc.result$krige.var
			kc.result$krige.var <- (exp(2 * predict.transf - 
				kc.result$krige.var)) * (exp(kc.result$
				krige.var) - 1)
			remove(predict.transf)
		}
		if(lambda > 0) {
			if(messages.screen)
				cat("krige.conv: back-transformation done by sampling from the resulting (normal) predictive distribution\n"
					)
			ap.warn <- options()$warn
			options(warn = -1)
			temp.data <- matrix(rnorm(ni * n.samples.backtransform,
				mean = kc.result$predict, sd = sqrt(kc.result$
				krige.var)), nrow = ni)
			options(warn = ap.warn)
			temp.data[(kc.result$krige.var == 0),  ] <- kc.result$
				predict[(kc.result$krige.var == 0)]
			temp.data[temp.data < -1/lambda] <- -1/lambda
			temp.data <- ((temp.data * lambda) + 1)^(1/lambda)
			###      temp.data[is.na(temp.data)] <- Inf
			kc.result$predict <- as.vector(apply(temp.data, 1,
				mean))
			kc.result$krige.var <- as.vector(apply(temp.data, 1,
				var))
		}
		if(lambda < 0) {
			cat("krige.conv: resulting distribution has no mean for lambda < 0 - back transformation not performed\n"
				)
		}
	}
	kc.result <- c(kc.result, list(message = message, call = call))
	############## Sampling from the resulting distribution #####################
	if(n.sim > 0) {
		print("sampling realisations from the resulting distribution")
		varcov <- varcov.spatial(coords = locations, cov.model = 
			cov.model, cov.pars = cov.pars, kappa = kappa, nugget
			 = nugget) - reduce.var
		remove("reduce.var")
		kc.result$samples <- kc.result$predict + (chol(varcov) %*% 
			matrix(rnorm(ni * n.sim), ncol = n.sim))
	}
	#######################################
	if(version$major <= 4) class(kc.result) <- setOldClass(c("kriging"))
		 else class(kc.result) <- c("kriging")
	return(kc.result)
}

"ksline" <- 
function(geodata, coords = geodata$coords, data = geodata$data, locations,
	cov.pars = stop("covariance parameters (sigmasq and phi) needed"),
	nugget = 0, micro.scale = 0, cov.model = c("exponential", "matern",
	"gaussian", "spherical", "cubic",  , "wave", "powered.exponential",
	"cauchy", "gneiting", "gneiting.matern", "pure.nugget"), kappa = 0.5,
	lambda = 1, n.samples.backtransform = 500, m0 = "ok", nwin = "full",
	trend = 1, d = 2, ktedata = NULL, ktelocations = NULL, aniso.pars = F,
	signal = F, messages.screen = T, dist.epsilon = 1e-10)
{
	micro.scale
	# This function performs SK,OK,KT or KTE for global or moving neighbourhood
	# The parameters:
	#      coords : a 2 columns matrix with data coordinates (for transect 1st col cte)
	#      data : data vector such data data[i] is data in position coords[i]
	#      locations : 2 columns matrix with coordinates of points to be estimated
	#      cov.pars : covariance parameters vector (sill,range)
	#      nugget: nugget effect (measurement error)
	#      cov.model : covariance model (a positive definite function)
	#      kappa : kappa parameter for Matern or powered exponential covariance function
	#      m0 : defines the type of kriging:
	#                       numeric value: SK with know mean
	#                                'av': SK with aritimetic mean
	#                                'ok': ordinary kriging
	#                                'kt': kriging with a trend model(universal)
	#                               'kte': kriging with external drift
	#      nwin : number of neighbours for kriging. 'full' = global neighbourhood.
	#      trend : (for KT only) 1 or 2: degree of polynomial trend
	#      d : dimension 1=transect 2=two dimensions 
	#      ktedata : (for KTE only) vector or matrix with external trend variables
	#                such that ktevars[i] is/are external data in position coords[i] 
	#      ktelocations : (for KTE only) vector or matrix with external trend variables
	#                such that ktelocations[i] is/are external data in position locations[i]
	# signal : if T only the signal is predictie, otherwise the variable is predicted. For exact=F, which corresponds to the model based perspective, the predicted value is the same whether signal is T or F. The difference is only in the kriging variances. For signal = F it includes the nugget therefore is bigger.
	#      micro.scale :  REWRITE THIS!!! model-based approach (on data locations predictions don't coincides with data if tausq(nugget) != 0). This is meant to reproduce conventional geostatistical interpolation where the predicted value at sample locations is equal to the observed value even if the nugget is different from 0.
	#
	# TO DO: 1d entrada com vetor!!! e checar 2D  AND  
	# KTE not working - in selection for window!!!  option for NA !!!
	# inserir em kte alteracao igual a de kt para valores altos
	# pode ser modificada para soltar varios resultados ao mesmo tempo : ok, sk, kte...
	#
	##### 1. Reading input information #####
	#  
	call <- match.call()
	if(lambda != 1) {
		if(messages.screen == T)
			print("Data transformation (Box-Cox) performed.")
		if(lambda == 0)
			data <- log(data)
		else data <- ((data^lambda) - 1)/lambda
	}
	cov.model <- match.arg(cov.model)
	coords <- as.matrix(coords)
	locations <- as.matrix(locations)
	dimnames(coords) <- list(NULL, NULL)
	dimnames(locations) <- list(NULL, NULL)
	if(!is.null(ktedata) & !is.null(ktelocations) & m0 != "kte") {
		print("external variable (covariate) provided. Kriging ste to KTE"
			)
		m0 <- "kte"
	}
	#
	#
	# Correcting anisotropy
	#
	if(any(aniso.pars) == T) {
		coords.c <- coords.aniso(coord = coords, aniso.pars = 
			aniso.pars)
		locations.c <- coords.aniso(coord = locations, aniso.pars = 
			aniso.pars)
	}
	else {
		coords.c <- coords
		locations.c <- locations
	}
	##### 2. Preparing KTE matrices #####
	#  
	if(m0 == "kte") {
		ktedata <- as.matrix(ktedata)
		ktelocations <- as.matrix(ktelocations)
		dimnames(ktedata) <- list(NULL, NULL)
		dimnames(ktelocations) <- list(NULL, NULL)
	}
	#
	##### 3. Preparing matrices to store results ######
	#
	n <- length(data)
	ni <- length(locations[, 1])
	tausq <- nugget
	sigmasq <- cov.pars[1]
	phi <- cov.pars[2]
	#
	##### 4. Kriging with full neighbourhood #####
	#
	if(nwin == "full") {
		est <- rep(0, ni)
		dif <- rep(0, ni)
		kvar <- rep(0, ni)
		sumw <- rep(0, ni)
		wofmean <- rep(0, ni)
		iv <- varcov.spatial(coords = coords.c, cov.model = cov.model,
			kappa = kappa, nugget = nugget, cov.pars = cov.pars,
			inv = T, det = F, func.inv = "chol", scaled = F, 
			only.decomp = F)
		av <- mean(data)
		sd <- sqrt(var(data))
		one <- rep(1, n)
		tone <- t(one)
		toneiv <- crossprod(one, iv)
		den <- solve(toneiv %*% one)
		ml <- den %*% toneiv %*% data
		kmsd <- sqrt(den)
		means <- c(average = av, stdev = sd, kmean = ml, kmsd = kmsd)
		if(m0 != "kt") {
			mktlocations <- "Constant trend"
			beta <- ml
		}
		else {
			mktlocations <- rep(0, ni)
			if(m0 == "kt" & trend == 1) {
				if(d == 1) {
					xmat <- cbind(rep(1, n), coords[, 2])
					xmati <- cbind(rep(1, ni), locations[
						, 2])
				}
				else {
					xmat <- cbind(rep(1, n), coords[, 1],
						coords[, 2])
					xmati <- cbind(rep(1, ni), locations[
						, 1], locations[, 2])
				}
				txiv <- crossprod(xmat, iv)
				iviv <- solve(txiv %*% xmat)
				beta <- iviv %*% txiv %*% data
				mkt <- xmat %*% beta
			}
			if(m0 == "kt" & trend == 2) {
				if(d == 1) {
					xmat <- cbind(rep(1, n), coords[, 2],
						(coords[, 2])^2)
					xmati <- cbind(rep(1, ni), locations[
						, 2], (locations[, 2])^2)
				}
				else {
					xmat <- cbind(rep(1, n), coords[, 1],
						coords[, 2], (coords[, 1])^
						2, (coords[, 2])^2, coords[
						, 1] * coords[, 2])
					xmati <- cbind(rep(1, ni), locations[
						, 1], locations[, 2], (
						locations[, 1])^2, (locations[
						, 2])^2, locations[, 1] * 
						locations[, 2])
				}
				txiv <- crossprod(xmat, iv)
				iviv <- solve(txiv %*% xmat)
				beta <- iviv %*% txiv %*% data
				mkt <- xmat %*% beta
			}
		}
		#
		#### 4.1 Preparing matrices for KT (UK) #####
		#
		if(m0 != "kte") mktelocations <- "No external trend" else {
			if(m0 == "kte") {
				#sxy <- paste(coords[,1],coords[,2],sep=',') : the data coordinates
				#sxyi <- paste(coordsi[,1],coordsi[,2],sep=',') : the locations+est.drift
				#row.names(xyi) <- scoordsi 
				#coordsi[scoords,]
				mktelocations <- rep(0, ni)
				xmat <- cbind(rep(1, n), ktedata)
				xmati <- cbind(rep(1, ni), ktelocations)
				txiv <- crossprod(xmat, iv)
				iviv <- solve(txiv %*% xmat)
				beta <- iviv %*% txiv %*% data
				mkte <- xmat %*% beta
			}
		}
		for(i in 1:ni) {
			if(messages.screen == T) {
				if(ni < 11)
					print(paste("kriging location: ", i,
						"out of", ni))
				else {
					if(ni < 101 & (i %% 10 == 1))
						print(paste(
							"kriging location: ",
							i, "out of", ni))
					if(ni > 100 & i %% 100 == 1)
						print(paste(
							"kriging location: ",
							i, "out of", ni))
					if(i == ni)
						print(paste(
							"kriging location: ",
							i, "out of", ni))
				}
			}
			coords0 <- cbind((coords.c[, 1] - locations.c[i, 1]),
				(coords.c[, 2] - locations.c[i, 2]))
			dm0 <- sqrt(coords0[, 1]^2 + coords0[, 2]^2)
			#      coords0 <- rbind(locations[i,  ], coords)
			#      dm0 <- dist(coords0)[1:n]
			v0 <- cov.spatial(obj = dm0, cov.model = cov.model,
				kappa = kappa, cov.pars = cov.pars)
			if(micro.scale == 0)
				v0[dm0 < dist.epsilon] <- sigmasq
			else v0[dm0 < dist.epsilon] <- micro.scale + sigmasq
			tv0 <- t(v0)
			v0iv <- crossprod(v0, iv)
			v0ivv0 <- v0iv %*% v0
			skw <- v0iv
			wofmean[i] <- 1 - sum(skw)
			#
			###### 4.2.1 Simple kriging with known mean
			#
			if(is.numeric(m0) == T) {
				dif[i] <- skw %*% (data - m0)
				est[i] <- m0 + dif[i]
				if(signal == T)
					kvar[i] <- sigmasq - v0ivv0
				else kvar[i] <- tausq + sigmasq - v0ivv0
				sumw[i] <- sum(skw)
			}
			#
			###### 4.2.2 Simple kriging with data average mean
			#
			if(m0 == "av") {
				dif[i] <- skw %*% (data - av)
				est[i] <- av + dif[i]
				if(signal == T)
					kvar[i] <- sigmasq - v0ivv0
				else kvar[i] <- tausq + sigmasq - v0ivv0
				sumw[i] <- sum(((tone/n) + skw - ((skw %*% one %*%
					tone)/n)))
			}
			#
			###### 4.2.3 Ordinary kriging (or SK with G.L.S. mean)
			#
			if(m0 == "ok") {
				dif[i] <- skw %*% (data - ml)
				est[i] <- ml + dif[i]
				redu <- as.vector(1 - toneiv %*% v0)
				if(signal == T)
					kvar[i] <- sigmasq - v0ivv0 + (redu %*%
						den %*% redu)
				else kvar[i] <- tausq + sigmasq - v0ivv0 + (
						redu %*% den %*% redu)
				sumw[i] <- sum((den %*% one + tv0 - v0iv %*% 
					one %*% den %*% tone) %*% iv)
			}
			#
			###### 4.2.4 Universal Kriging (or Kriging with trend model) 
			#
			if(m0 == "kt") {
				dif[i] <- skw %*% (data - mkt)
				est[i] <- xmati[i,  ] %*% beta + dif[i]
				redu <- as.vector(xmati[i,  ]) - as.vector(
					txiv %*% v0)
				if(signal == T)
					kvar[i] <- sigmasq - v0ivv0 + (redu %*%
						iviv %*% redu)
				else kvar[i] <- tausq + sigmasq - v0ivv0 + (
						redu %*% iviv %*% redu)
				sumw[i] <- sum(skw + xmati[i,  ] %*% iviv %*%
					txiv - skw %*% xmat %*% iviv %*% txiv)
				mktlocations[i] <- xmati[i,  ] %*% beta
			}
			#
			###### 4.2.5 Kriging with external trend 
			#
			if(m0 == "kte") {
				dif[i] <- skw %*% (data - mkte)
				est[i] <- xmati[i,  ] %*% beta + dif[i]
				redu <- as.vector(xmati[i,  ]) - as.vector(
					txiv %*% v0)
				if(signal == T)
					kvar[i] <- sigmasq - v0ivv0 - (redu %*%
						iviv %*% redu)
				else kvar[i] <- tausq + sigmasq - v0ivv0 + (
						redu %*% iviv %*% redu)
				sumw[i] <- sum(skw + xmati[i,  ] %*% iviv %*%
					txiv - skw %*% xmat %*% iviv %*% txiv)
				mktelocations[i] <- xmati[i,  ] %*% beta
			}
		}
		message <- "Kriging performed in global neighbourhood"
		if(messages.screen == T)
			print(message)
		results <- list(predict = est, krige.var = kvar, dif = dif,
			summary = means, ktrend = mktlocations, ktetrend = 
			mktelocations, beta = beta, wofmean = wofmean)
	}
	else {
		#
		#### 5.1 Preparing vectors to store results and matrices #####
		#
		nwin <- min(n, nwin)
		avwin <- rep(0, ni)
		sdwin <- rep(0, ni)
		mlwin <- rep(0, ni)
		kmsdwin <- rep(0, ni)
		estwin <- rep(0, ni)
		difwin <- rep(0, ni)
		kvarwin <- rep(0, ni)
		sumwwin <- rep(0, ni)
		wofmean <- rep(0, ni)
		if(m0 != "kt")
			mkt <- "Constant position trend"
		else mkt <- rep(0, ni)
		if(m0 != "kte")
			mkte <- "No external trend"
		else mkte <- rep(0, ni)
		if(m0 != "kt" & m0 != "kte")
			betawin <- "No polynomial or external trend"
		if(m0 == "kt") {
			if(trend == 1) {
				if(d == 1)
					xmati <- cbind(rep(1, ni), locations[
						, 2])
				else xmati <- cbind(rep(1, ni), locations[
						, 1], locations[, 2])
			}
			if(trend == 2) {
				if(d == 1)
					xmati <- cbind(rep(1, ni), locations[
						, 2], locations[, 2]^2)
				else xmati <- cbind(rep(1, ni), locations[
						, 1], locations[, 2], (
						locations[, 1])^2, (locations[
						, 2])^2, locations[, 1] * 
						locations[, 2])
			}
			betawin <- matrix(0, nrow = (ncol(xmati) * ni), ncol = 
				ncol(xmati))
		}
		if(m0 == "kte") {
			xmati <- cbind(rep(1, ni), ktelocations)
			if(is.vector(ktedata) == T)
				betawin <- matrix(0, nrow = (2 * ni), ncol = 2)
			else betawin <- matrix(0, nrow = ((ncol(ktedata) + 1) *
					ni), ncol = (ncol(ktedata) + 1))
		}
		#
		#### 5.2 Kriging calculations:  #####
		#
		for(i in 1:ni) {
			temp.win <- ksline.aux1(coords = coords, coords.c = 
				coords.c, data = data, n = n, locations = 
				locations[i,  ], locations.c = locations.c[
				i,  ], cov.pars = cov.pars, nugget = nugget,
				cov.model = cov.model, kappa = kappa, m0 = m0,
				nwin = nwin, trend = trend, d = d, ktedata = 
				ktedata, ktelocations = ktelocations, 
				micro.scale = micro.scale, location.number = i,
				xmati = xmati[i,  ], mkte = NULL, mkt = NULL,
				betawin = NULL, signal = signal, dist.epsilon
				 = dist.epsilon)
			avwin[i] <- temp.win$avwin
			sdwin[i] <- temp.win$sdwin
			mlwin[i] <- temp.win$mlwin
			kmsdwin[i] <- temp.win$kmsdwin
			estwin[i] <- temp.win$estwin
			difwin[i] <- temp.win$difwin
			kvarwin[i] <- temp.win$kvarwin
			sumwwin[i] <- temp.win$sumwwin
			wofmean[i] <- temp.win$wofmean
			if(m0 == "kt")
				mkt[i] <- temp.win$mkt
			if(m0 == "kte")
				mkte[i] <- temp.win$mkte
			if(m0 == "kt" | m0 == "kte")
				betawin[i,  ] <- temp.win$betawin
			if(messages.screen == T) {
				if(ni < 11)
					print(paste("kriging location: ", i,
						"out of", ni))
				else {
					if(ni < 101 & (i %% 10 == 1))
						print(paste(
							"kriging location: ",
							i, "out of", ni))
					if(ni > 100 & i %% 100 == 1)
						print(paste(
							"kriging location: ",
							i, "out of", ni))
					if(i == ni)
						print(paste(
							"kriging location: ",
							i, "out of", ni))
				}
			}
		}
		message <- "kriging performed in moving neighbourhood"
		if(messages.screen == T)
			print(message)
		results <- list(predict = estwin, krige.var = kvarwin, dif = 
			difwin, avtrend = avwin, sd = sdwin, oktrend = mlwin,
			oksd = kmsdwin, ktrend = mkt, ktetrend = mkte, beta = 
			betawin, sumw = sumwwin, wofmean = wofmean)
	}
	results$krige.var <- round(results$krige.var, dig = 8)
	if(lambda != 1) {
		if(messages.screen == T)
			print("Back-transforming the predictions according to the (Box-Cox) parameter lambda"
				)
		if(lambda == 0) {
			predict.transf <- results$predict
			results$predict <- exp(predict.transf) - 0.5 * results$
				krige.var
			results$krige.var <- (exp(2 * predict.transf - results$
				krige.var)) * (exp(results$krige.var) - 1)
		}
		if(lambda > 0) {
			if(messages.screen == T)
				print("Back-transformation done by sampling from the resulting (normal) predictive distribution"
					)
			ap.warn <- options()$warn
			options(warn = -1)
			temp.data <- matrix(rnorm(ni * n.samples.backtransform,
				mean = results$predict, sd = sqrt(results$
				krige.var)), nrow = ni)
			options(warn = ap.warn)
			temp.data[(results$krige.var == 0),  ] <- results$
				predict[(results$krige.var == 0)]
			temp.data[temp.data < -1/lambda] <- -1/lambda
			temp.data <- ((temp.data * lambda) + 1)^(1/lambda)
			#      temp.data[is.na(temp.data)] <- Inf
			results$predict <- as.vector(apply(temp.data, 1, mean))
			results$krige.var <- as.vector(apply(temp.data, 1,
				var))
		}
		if(lambda < 0) {
			print("Resulting distribution has no mean for lambda < 0 - back transformation not performed"
				)
		}
	}
	results$locations <- locations
	results$message <- message
	results$call <- call
	if(version$major <= 4)
		class(results) <- setOldClass(c("kriging", "ksline"))
	else class(results) <- c("kriging", "ksline")
	return(results)
}

"ksline.aux.1" <- 
function(coords, coords.c, data, locations, locations.c, n, cov.pars, nugget,
	cov.model, kappa, m0, nwin, trend, d, ktedata, ktelocations, 
	micro.scale, messages, location.number, xmati, mkte, mkt, betawin,
	signal, dist.epsilon = dist.epsilon)
{
	i <- location.number
	sigmasq <- cov.pars[1]
	phi <- cov.pars[2]
	tausq <- nugget
	coords0 <- cbind((coords.c[, 1] - locations.c[1]), (coords.c[, 2] -
		locations.c[2]))
	dm0 <- sqrt(coords0[, 1]^2 + coords0[, 2]^2)
	coordswin <- coords[order(dm0)[1:nwin],  ]
	coordswin.c <- coords.c[order(dm0)[1:nwin],  ]
	datawin <- data[order(dm0)[1:nwin]]
	ivwin <- varcov.spatial(coords = coordswin.c, cov.model = cov.model,
		kappa = kappa, nugget = nugget, cov.pars = cov.pars, inv = T,
		det = F, func.inv = "chol", scaled = F, only.decomp = F)
	avwin <- mean(datawin)
	sdwin <- sqrt(var(datawin))
	onewin <- rep(1, nwin)
	toneivwin <- crossprod(onewin, ivwin)
	denwin <- solve(toneivwin %*% onewin)
	mlwin <- denwin %*% toneivwin %*% datawin
	kmsdwin <- sqrt(denwin)
	coords0win <- cbind((coordswin[, 1] - locations[1]), (coordswin[, 2] -
		locations[2]))
	coords0win.c <- cbind((coordswin.c[, 1] - locations.c[1]), (coordswin.c[
		, 2] - locations.c[2]))
	dm0win <- sqrt(coords0win.c[, 1]^2 + coords0win.c[, 2]^2)
	v0win <- cov.spatial(obj = dm0win, cov.model = cov.model, kappa = kappa,
		cov.pars = cov.pars)
	if(micro.scale == T)
		v0win[dm0win < dist.epsilon] <- sigmasq
	else v0win[dm0win < dist.epsilon] <- micro.scale + sigmasq
	skwwin <- crossprod(v0win, ivwin)
	wofmean <- 1 - sum(skwwin)
	if(m0 == "kt" & trend == 1) {
		if(d == 1)
			xmatwin <- cbind(rep(1, nwin), coordswin[, 2])
		else xmatwin <- cbind(rep(1, nwin), coordswin[, 1], coordswin[
				, 2])
		txivwin <- crossprod(xmatwin, ivwin)
		ivivwin <- solve(txivwin %*% xmatwin)
		betawin <- ivivwin %*% txivwin %*% datawin
		mktwin <- xmatwin %*% betawin
		mkt <- xmati %*% betawin
	}
	if(m0 == "kt" & trend == 2) {
		if(d == 1)
			xmatwin <- cbind(rep(1, nwin), coordswin[, 2], (
				coordswin[, 2])^2)
		else xmatwin <- cbind(rep(1, nwin), coordswin[, 1], (coordswin[
				, 1])^2, coordswin[, 2], (coordswin[, 2])^
				2, coordswin[, 1] * coordswin[, 2])
		xmatwin.cent <- xmatwin
		xmatwin.cent[, 2] <- xmatwin.cent[, 2] - mean(xmatwin[, 2])
		xmatwin.cent[, 3] <- xmatwin.cent[, 3] - mean(xmatwin[, 3])
		ivivwin <- solve(crossprod(xmatwin.cent, ivwin) %*% 
			xmatwin.cent)
		txivwin <- crossprod(xmatwin.cent, ivwin)
		betawin <- ivivwin %*% txivwin %*% datawin
		betawin <- mean(datawin) - crossprod(betawin, c(0, mean(xmatwin[
			, 2]), mean(xmatwin[, 3])))
		mktwin <- xmatwin %*% betawin
		mkt <- xmati %*% betawin
	}
	if(m0 == "kte") {
		if(is.vector(ktedata) == T)
			ktedatawin <- ktedata[order(dm0)[1:nwin]]
		else ktedatawin <- ktedata[order(dm0)[1:nwin],  ]
		xmatwin <- cbind(rep(1, nwin), ktedatawin)
		ivivwin <- solve(crossprod(xmatwin, ivwin) %*% xmatwin)
		txivwin <- crossprod(xmatwin, ivwin)
		betawin <- ivivwin %*% txivwin %*% datawin
		mktewin <- xmatwin %*% betawin
		mkte <- xmati %*% betawin
	}
	#
	######  Simple kriging with known mean
	#
	if(is.numeric(m0) == T) {
		difwin <- skwwin %*% (data - m0)
		estwin <- m0win + difwin
		if(signal == T)
			kvarwin <- sigmasq - crossprod(v0win, ivwin) %*% v0win
		else kvarwin <- tausq + sigmasq - crossprod(v0win, ivwin) %*%
				v0win
		sumwwin <- sum(skwwin)
	}
	#
	###### 4.2.2 Simple kriging with data average mean
	#
	if(m0 == "av") {
		difwin <- skwwin %*% (datawin - avwin)
		estwin <- avwin + difwin
		if(signal == T)
			kvarwin <- sigmasq - crossprod(v0win, ivwin) %*% v0win
		else kvarwin <- tausq + sigmasq - crossprod(v0win, ivwin) %*%
				v0win
		sumwwin <- sum(((t(onewin)/nwin) + skwwin - ((skwwin %*% onewin %*%
			t(onewin))/n)))
	}
	#
	###### Ordinary kriging (or SK with G.L.S. mean)
	#
	if(m0 == "ok") {
		difwin <- skwwin %*% (datawin - mlwin)
		estwin <- mlwin + difwin
		redu <- as.vector(1 - toneivwin %*% v0win)
		if(signal == T)
			kvarwin <- sigmasq - v0win %*% ivwin %*% v0win + (
				redu %*% denwin %*% redu)
		else kvarwin <- tausq + sigmasq - v0win %*% ivwin %*% v0win +
				(redu %*% denwin %*% redu)
		sumwwin <- sum((denwin %*% onewin + t(v0win) - crossprod(v0win,
			ivwin) %*% onewin %*% denwin %*% t(onewin)) %*% ivwin)
	}
	#
	###### Universal Kriging (or Kriging with trend model) 
	#
	if(m0 == "kt") {
		difwin <- skwwin %*% (datawin - mktwin)
		estwin <- mkt + difwin
		xmati <- as.vector(xmati)
		redu <- as.vector(xmati) - as.vector(txivwin %*% v0win)
		if(signal == T)
			kvarwin <- sigmasq - v0win %*% ivwin %*% v0win + (
				redu %*% ivivwin %*% redu)
		else kvarwin <- tausq + sigmasq - v0win %*% ivwin %*% v0win +
				(redu %*% ivivwin %*% redu)
		sumwwin <- sum(skwwin + xmati %*% ivivwin %*% txivwin - skwwin %*%
			xmatwin %*% ivivwin %*% txivwin)
	}
	#
	###### Kriging with external trend 
	#
	if(m0 == "kte") {
		difwin <- skwwin %*% (datawin - mktewin)
		estwin <- mkte + difwin
		xmati <- as.vector(xmati)
		redu <- as.vector(xmati) - as.vector(txivwin %*% v0win)
		if(signal == T)
			kvarwin <- sigmasq - v0win %*% ivwin %*% v0win + (
				redu %*% ivivwin %*% redu)
		else kvarwin <- tausq + sigmasq - v0win %*% ivwin %*% v0win +
				(redu %*% ivivwin %*% redu)
		sumwwin <- sum(skwwin + xmati %*% ivivwin %*% txivwin - skwwin %*%
			xmatwin %*% ivivwin %*% txivwin)
	}
	#
	#
	#  
	results <- list(avwin = avwin, sdwin = sdwin, mlwin = mlwin, kmsdwin = 
		kmsdwin, wofmean = wofmean, betawin = betawin, mkt = mkt, mkte
		 = mkte, difwin = difwin, estwin = estwin, kvarwin = kvarwin,
		sumwwin = sumwwin)
	return(results)
}

"ksline.aux1" <- 
function(coords, coords.c, data, locations, locations.c, n, cov.pars, nugget,
	cov.model, kappa, m0, nwin, trend, d, ktedata, ktelocations, 
	micro.scale, messages, location.number, xmati, mkte, mkt, betawin,
	signal, dist.epsilon = dist.epsilon)
{
	i <- location.number
	sigmasq <- cov.pars[1]
	phi <- cov.pars[2]
	tausq <- nugget
	coords0 <- cbind((coords.c[, 1] - locations.c[1]), (coords.c[, 2] -
		locations.c[2]))
	dm0 <- sqrt(coords0[, 1]^2 + coords0[, 2]^2)
	coordswin <- coords[order(dm0)[1:nwin],  ]
	coordswin.c <- coords.c[order(dm0)[1:nwin],  ]
	datawin <- data[order(dm0)[1:nwin]]
	ivwin <- varcov.spatial(coords = coordswin.c, cov.model = cov.model,
		kappa = kappa, nugget = nugget, cov.pars = cov.pars, inv = T,
		det = F, func.inv = "chol", scaled = F, only.decomp = F)
	avwin <- mean(datawin)
	sdwin <- sqrt(var(datawin))
	onewin <- rep(1, nwin)
	toneivwin <- crossprod(onewin, ivwin)
	denwin <- solve(toneivwin %*% onewin)
	mlwin <- denwin %*% toneivwin %*% datawin
	kmsdwin <- sqrt(denwin)
	coords0win <- cbind((coordswin[, 1] - locations[1]), (coordswin[, 2] -
		locations[2]))
	coords0win.c <- cbind((coordswin.c[, 1] - locations.c[1]), (coordswin.c[
		, 2] - locations.c[2]))
	dm0win <- sqrt(coords0win.c[, 1]^2 + coords0win.c[, 2]^2)
	v0win <- cov.spatial(obj = dm0win, cov.model = cov.model, kappa = kappa,
		cov.pars = cov.pars)
	if(micro.scale == T)
		v0win[dm0win < dist.epsilon] <- sigmasq
	else v0win[dm0win < dist.epsilon] <- micro.scale + sigmasq
	skwwin <- crossprod(v0win, ivwin)
	wofmean <- 1 - sum(skwwin)
	if(m0 == "kt" & trend == 1) {
		if(d == 1)
			xmatwin <- cbind(rep(1, nwin), coordswin[, 2])
		else xmatwin <- cbind(rep(1, nwin), coordswin[, 1], coordswin[
				, 2])
		txivwin <- crossprod(xmatwin, ivwin)
		ivivwin <- solve(txivwin %*% xmatwin)
		betawin <- ivivwin %*% txivwin %*% datawin
		mktwin <- xmatwin %*% betawin
		mkt <- xmati %*% betawin
	}
	if(m0 == "kt" & trend == 2) {
		if(d == 1)
			xmatwin <- cbind(rep(1, nwin), coordswin[, 2], (
				coordswin[, 2])^2)
		else xmatwin <- cbind(rep(1, nwin), coordswin[, 1], (coordswin[
				, 1])^2, coordswin[, 2], (coordswin[, 2])^
				2, coordswin[, 1] * coordswin[, 2])
		xmatwin.cent <- xmatwin
		xmatwin.cent[, 2] <- xmatwin.cent[, 2] - mean(xmatwin[, 2])
		xmatwin.cent[, 3] <- xmatwin.cent[, 3] - mean(xmatwin[, 3])
		ivivwin <- solve(crossprod(xmatwin.cent, ivwin) %*% 
			xmatwin.cent)
		txivwin <- crossprod(xmatwin.cent, ivwin)
		betawin <- ivivwin %*% txivwin %*% datawin
		betawin <- mean(datawin) - crossprod(betawin, c(0, mean(xmatwin[
			, 2]), mean(xmatwin[, 3])))
		mktwin <- xmatwin %*% betawin
		mkt <- xmati %*% betawin
	}
	if(m0 == "kte") {
		if(is.vector(ktedata) == T)
			ktedatawin <- ktedata[order(dm0)[1:nwin]]
		else ktedatawin <- ktedata[order(dm0)[1:nwin],  ]
		xmatwin <- cbind(rep(1, nwin), ktedatawin)
		ivivwin <- solve(crossprod(xmatwin, ivwin) %*% xmatwin)
		txivwin <- crossprod(xmatwin, ivwin)
		betawin <- ivivwin %*% txivwin %*% datawin
		mktewin <- xmatwin %*% betawin
		mkte <- xmati %*% betawin
	}
	#
	######  Simple kriging with known mean
	#
	if(is.numeric(m0) == T) {
		difwin <- skwwin %*% (data - m0)
		estwin <- m0win + difwin
		if(signal == T)
			kvarwin <- sigmasq - crossprod(v0win, ivwin) %*% v0win
		else kvarwin <- tausq + sigmasq - crossprod(v0win, ivwin) %*%
				v0win
		sumwwin <- sum(skwwin)
	}
	#
	###### 4.2.2 Simple kriging with data average mean
	#
	if(m0 == "av") {
		difwin <- skwwin %*% (datawin - avwin)
		estwin <- avwin + difwin
		if(signal == T)
			kvarwin <- sigmasq - crossprod(v0win, ivwin) %*% v0win
		else kvarwin <- tausq + sigmasq - crossprod(v0win, ivwin) %*%
				v0win
		sumwwin <- sum(((t(onewin)/nwin) + skwwin - ((skwwin %*% onewin %*%
			t(onewin))/n)))
	}
	#
	###### Ordinary kriging (or SK with G.L.S. mean)
	#
	if(m0 == "ok") {
		difwin <- skwwin %*% (datawin - mlwin)
		estwin <- mlwin + difwin
		redu <- as.vector(1 - toneivwin %*% v0win)
		if(signal == T)
			kvarwin <- sigmasq - v0win %*% ivwin %*% v0win + (
				redu %*% denwin %*% redu)
		else kvarwin <- tausq + sigmasq - v0win %*% ivwin %*% v0win +
				(redu %*% denwin %*% redu)
		sumwwin <- sum((denwin %*% onewin + t(v0win) - crossprod(v0win,
			ivwin) %*% onewin %*% denwin %*% t(onewin)) %*% ivwin)
	}
	#
	###### Universal Kriging (or Kriging with trend model) 
	#
	if(m0 == "kt") {
		difwin <- skwwin %*% (datawin - mktwin)
		estwin <- mkt + difwin
		xmati <- as.vector(xmati)
		redu <- as.vector(xmati) - as.vector(txivwin %*% v0win)
		if(signal == T)
			kvarwin <- sigmasq - v0win %*% ivwin %*% v0win + (
				redu %*% ivivwin %*% redu)
		else kvarwin <- tausq + sigmasq - v0win %*% ivwin %*% v0win +
				(redu %*% ivivwin %*% redu)
		sumwwin <- sum(skwwin + xmati %*% ivivwin %*% txivwin - skwwin %*%
			xmatwin %*% ivivwin %*% txivwin)
	}
	#
	###### Kriging with external trend 
	#
	if(m0 == "kte") {
		difwin <- skwwin %*% (datawin - mktewin)
		estwin <- mkte + difwin
		xmati <- as.vector(xmati)
		redu <- as.vector(xmati) - as.vector(txivwin %*% v0win)
		if(signal == T)
			kvarwin <- sigmasq - v0win %*% ivwin %*% v0win + (
				redu %*% ivivwin %*% redu)
		else kvarwin <- tausq + sigmasq - v0win %*% ivwin %*% v0win +
				(redu %*% ivivwin %*% redu)
		sumwwin <- sum(skwwin + xmati %*% ivivwin %*% txivwin - skwwin %*%
			xmatwin %*% ivivwin %*% txivwin)
	}
	#
	#
	#  
	results <- list(avwin = avwin, sdwin = sdwin, mlwin = mlwin, kmsdwin = 
		kmsdwin, wofmean = wofmean, betawin = betawin, mkt = mkt, mkte
		 = mkte, difwin = difwin, estwin = estwin, kvarwin = kvarwin,
		sumwwin = sumwwin)
	if(version$major <= 4)
		class(results) <- c("kriging")
	else class(results) <- setOldClass(c("kriging"))
	return(kc.result)
	return(results)
}

"latexGraph" <- 
function(file = "junk", height = 3, width = 5, labels = F)
{
	# to import this into LaTeX, use the following commands:
	# \usepackage{graphicx}
	# \begin{figure}
	#   \includegraphics[height=3in, width=5in]{junk.ps}
	# \end{figure}
	assign(".temp.ps.orig", ps.options(), w = 1)
	ps.options(maximize = T, width = width, height = height, horizontal = F,
		region = c(0, 0, 72 * width, 72 * height))
	postscript(file = file, print.it = F)
	if(labels == F)
		par(mar = c(1.8, 1.8, 0, 0), oma = c(0, 0, 1.45, 0), cex = 0.5,
			las = 1, mgp = c(1.3, 0.5, 0))
	else par(mar = c(2.2999999999999998, 2.2999999999999998, 0, 0), oma = c(
			0, 0, 1.45, 0), cex = 0.5, las = 1, mgp = c(1.3, 0.5,
			0))
	return(NULL)
}

"latexGraph.off" <- 
function(which = dev.cur())
{
	if(which == 1)
		stop("Cannot shut down device 1 (the null device)")
	if(exists(".Devices", frame = 0)) {
		.Devices <- get(".Devices", frame = 0)
	}
	else {
		.Devices <- list("null device")
	}
	.Devices[[which]] <- ""
	assign(".Devices", .Devices, frame = 0)
	.C("S_shutdown_device",
		as.integer(which))
	assign(".Device", .Devices[[dev.cur()]], frame = 0)
	ps.options(.temp.ps.orig)
	remove(".temp.ps.orig", w = 1)
	dev.cur()
}

"levelplot.ksline" <- 
function(obj.ksline, grid = obj.ksline$grid, values = c("est", "kvar", 
	"oktrend", "ktrend", "ktetrend"), ...)
{
	#
	# obj.ksline: tipically an object generated by the function ksline
	# grid      : regular grid where the interpolation was performed
	# values    : valeus to be plotted. The options are:
	##        - est: the kriging estimates (default option)
	##        - kvar : the kriging variances
	values <- match.arg(values)
	#  grid <- obj.ksline$grid[order(grid[, 2], grid[, 1]),  ]
	#  if(nx == ny)
	#    par(pty = "s")
	if(values == "est") m <- obj.ksline$pred
	if(values == "kvar")
		m <- obj.ksline$krige.var
	levelplot(m ~ obj.ksline$grid[, 1] * obj.ksline$grid[, 2],  , xlab = 
		"Coord X", ylab = "Coord Y", ...)
}

"likfit" <- 
function(geodata, coords = geodata$coords, data = geodata$data, ini, trend = 
	"cte", fix.nugget = F, nugget = 0, method = "ML", cov.model = c(
	"matern", "exponential", "gaussian", "spherical", "cubic", "wave",
	"powered.exponential", "cauchy", "gneiting", "gneiting.matern", 
	"pure.nugget"), kappa = 0.5, fix.lambda = T, lambda = 1, predicted = F,
	residuals = F, messages.screen = T, automatic.refit = T, control = 
	nlminb.control(iter.max = 1000), range.limits = "default", ...)
{
	#			stop("trend must be expecified as `cte`, `1st`, `2nd` or as formulae of the type: ~ X where X is a matrix with the trend variables (covariates. A formula of the type ~X-1 will define a vector of means (which can be different for each location)"
	call <- match.call()
	ftau <- nugget
	fixtau <- fix.nugget
	cov.model <- match.arg(cov.model)
	if(cov.model == "pure.nugget")
		ini <- 0
	if(!is.null(kappa))
		if(cov.model == "matern" & kappa == 0.5)
			cov.model <- "exponential"
	coords <- as.matrix(coords)
	dists.vec <- as.vector(dist(coords))
	range.dist <- range(dists.vec)
	max.dist <- max(range.dist)
	min.dist <- min(range.dist)
	if(all(range.limits == "default")) {
		lower.phi <- 0
		upper.phi <-  + Inf
	}
	else {
		lower.phi <- range.limits[1]
		upper.phi <- range.limits[2]
	}
	z <- as.vector(data)
	if(fix.lambda == T) {
		if(lambda == 1) {
			log.jacobian <- 0
		}
		else {
			if(any(data <= 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(data^(lambda - 1) <= 0))
				log.jacobian <- log(prod(data^(lambda - 1)))
			else log.jacobian <- sum(log(data^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	n <- length(z)
	if((2 * n) != length(coords))
		stop("Number of locations does not match with number of data")
	reduce.pars <- 0
	if(method == "REML" | method == "reml" | method == "rml")
		method <- "RML"
	if(method == "ML" | method == "ml")
		method <- "ML"
	# Checking initial parameters vector size:
	if(all(ini == 0)) {
		cov.model <- "pure.nugget"
		cat("likfit: all initial values equal to zero. Model without spatial correlation will be fitted\n"
			)
	}
	else {
		if(is.matrix(ini)) {
			inilength <- dim(ini)[2]
			if(fixtau == F & inilength != 3)
				stop("wrong number of columns for ini (must be 3)"
					)
			if(fixtau == T & inilength != 2)
				stop("wrong number of columns for ini (must be 2)"
					)
		}
		else {
			inilength <- length(ini)
			if(fixtau == F & inilength != 3)
				stop("wrong length for ini (must be 3)")
			if(fixtau == T & inilength != 2)
				stop("wrong length for ini (must be 2)")
		}
	}
	xmat <- trend.spatial(trend = trend, coords = coords)
	fit.ols <- lm(z ~ xmat - 1)
	trend.ols <- list(coefficients = fit.ols$coefficients)
	var.z <- sum((fit.ols$residual)^2)/(n - length(fit.ols$coefficients))
	dimnames(xmat) <- list(NULL, NULL)
	txmat <- t(xmat)
	beta.size <- dim(xmat)[2]
	# List with objects to be passed to the minimisation function
	temp.list <- list(z = z, xmat = xmat, txmat = txmat, fixtau = fixtau,
		ftau = ftau, method = method, kappa = kappa, cov.model = 
		cov.model, beta.size = beta.size, lower.phi = lower.phi, 
		dists.lowertri = dists.vec, var.z = var.z, fix.lambda = 
		fix.lambda, n = n, control = control)
	if(fix.lambda == T) {
		temp.list$lambda <- lambda
		temp.list$log.jacobian <- log.jacobian
	}
	# Searching for best initial value
	if((cov.model == "pure.nugget") | all(ini == 0)) {
		lik.results <- likfit.nospatial(temp.list, ...)
		if(fix.nugget == F)
			temp.pars <- c(lik.results$tausqhat, 0, 0)
		else temp.pars <- c(ftau, (lik.results$tausqhat - ftau), 0)
		lambda <- lik.results$lambda
	}
	else {
		if(is.matrix(ini) | is.data.frame(ini)) {
			ini <- as.matrix(ini)
			if(messages.screen == T)
				cat("likfit: searching for the best initial value\n"
					)
			ini.search <- ini
			if(fix.nugget == T)
				ini.search <- cbind(nugget, ini.search)
			if(length(lambda) == 1)
				ini.search <- cbind(ini.search, lambda)
			else {
				temp <- ini.search
				for(i in 1:(length(lambda) - 1)) {
					ini.search <- rbind(ini.search, temp)
				}
				ini.search <- cbind(ini.search, rep(lambda,
					each = dim(temp)[1]))
			}
			dimnames(ini.search) <- list(NULL, NULL)
			loglik.ini <- round(100000000. * apply(ini.search,
				1, loglik.spatial, temp.list))
			ini.max <- as.vector(ini.search[loglik.ini == max(
				loglik.ini),  ])
			if(fixtau == T) {
				ini <- as.vector(ini.max[2:3])
			}
			else {
				ini <- as.vector(ini.max[1:3])
			}
			if(messages.screen == T) {
				cat("likfit: best initial value:\n")
				names(ini.max) <- c("nugget", "sill", "range",
					"lambda")
				print(ini.max)
			}
			if(fix.lambda == F)
				lambda <- as.vector(ini.max[4])
		}
		# Finding the ML/REML estimates using numerical methods
		if(messages.screen == T) {
			cat("likfit: Initialising likelihood maximisation using nlminb.\n"
				)
			cat("likfit: WARNING: This step can be time demanding!\n"
				)
			cat("------------------------------------------------------------\n"
				)
			cat("Consider providing additional (non-default) arguments for minimisation function nlminb using the argument control.\n"
				)
			cat("For more details see documentation for the function nlminb.control.\n"
				)
			cat("------------------------------------------------------------\n"
				)
			cat(paste(
				"likfit: range of values for the parameter phi:",
				nice(lower.phi), "and", nice(upper.phi), "\n"))
		}
		if(fixtau == F | ftau != 0) {
			# case 1: parameters = (\sigma^2, \phi) ; nugget fixed and different from 0
			if(fixtau == T & ftau != 0) {
				#      if(messages.screen == T)
				#        print("covariance parameters used in the minimization algorithm  are $sigma^2$ and $phi$")
				lower.sigmasq <- 0.01 * var.z
				if(fix.lambda == T) {
					lik.results <- nlminb(ini, proflik.ftau,
						lower = c(0, lower.phi), upper
						 = c( + Inf, upper.phi), 
						temp.list = temp.list, control
						 = control, ...)
				}
				else {
					lik.results <- nlminb(c(ini, lambda),
						proflik.ftau, lower = c(0,
						lower.phi, -3), upper = c(
						 + Inf, upper.phi, 3), 
						temp.list = temp.list, control
						 = control, ...)
					lambda <- lik.results$parameters[3]
					lik.results$parameters <- as.vector(
						lik.results$parameters[1:2])
					if(lambda == 0)
						z <- log(as.vector(data))
					else z <- (((as.vector(data))^lambda) -
							1)/lambda
				}
				lik.results$parameters <- temp.pars <- 
					as.vector(c(ftau, lik.results$
					parameters))
				if(messages.screen == T) {
					print(lik.results$message)
				}
			}
			# case 2: parameters = (\tau^2, \sigma^2, \phi)
			if(fixtau == F) {
				ini.m <- c(ini[1]/ini[2], ini[3])
				#      if(messages.screen == T)
				#        print("parameters used in the minimization function are the ratio (\tausq/sigmasq) and phi")
				if(fix.lambda == T) {
					lik.results <- nlminb(ini.m, 
						proflik.nug, lower = c(0, 
						lower.phi), upper = c( + Inf,
						upper.phi), temp.list = 
						temp.list, control = control,
						...)
				}
				else {
					lik.results <- nlminb(c(ini.m, lambda),
						proflik.nug, lower = c(0, 
						lower.phi, -3), upper = c(
						 + Inf, upper.phi, 3), 
						temp.list = temp.list, control
						 = control, ...)
					lambda <- lik.results$parameters[3]
					lik.results$parameters <- as.vector(
						lik.results$parameters[1:2])
					if(lambda == 0)
						z <- log(as.vector(data))
					else z <- (((as.vector(data))^lambda) -
							1)/lambda
				}
				if(messages.screen == T) {
					print(lik.results$message)
				}
				if(automatic.refit == T & (lik.results$
					parameters[1] < 0.01)) {
					if(messages.screen == T)
						cat(paste(
							"likfit: WARNING: ratio of estimates $\tau^2/$sigma^2$ < 0.01 (",
							round(lik.results$
							parameters[1], dig = 4),
							"\n"))
					reduce.pars <- 1
					temp.list$ftau <- 0
					temp.list$fixtau <- T
					ini.m <- ini[3]
					if(fix.lambda == T) {
						lik.results <- nlminb(
							lik.results$parameters[
							2], proflik.phi, lower
							 = lower.phi, upper = 
							upper.phi, temp.list = 
							temp.list, control = 
							control, ...)
					}
					else {
						lik.results <- nlminb(c(
							lik.results$parameters[
							2], lambda), 
							proflik.phi, lower = c(
							lower.phi, -3), upper
							 = c(upper.phi, 3),
							temp.list = temp.list,
							control = control,
							...)
						lambda <- as.vector(lik.results$
							parameters[2])
						lik.results$parameters <- 
							as.vector(lik.results$
							parameters[1])
						if(lambda == 0)
							z <- log(as.vector(
								data))
						else z <- (((as.vector(data))^
								lambda) - 1)/
								lambda
					}
					if(messages.screen == T)
						cat("likfit: model re-fitted without nugget effect (tausq = 0)\n"
							)
					lik.results$parameters <- as.vector(
						c(0, lik.results$parameters))
					if(messages.screen == T) {
						print(lik.results$message)
					}
				}
				nugget.rel <- lik.results$parameters[1]
				if(lik.results$parameters[2] < 1e-08)
					icovhat <- diag(n)
				else icovhat <- varcov.spatial(coords = coords,
						cov.model = cov.model, kappa = 
						kappa, nugget = nugget.rel,
						cov.pars = c(1, lik.results$
						parameters[2]), inv = T, det = 
						F)
				txiv <- crossprod(xmat, icovhat)
				sigmasqhat <- (z %*% (icovhat - crossprod(
					txiv, solve(txiv %*% xmat)) %*% txiv) %*%
					z)/n
				if(method == "RML")
					sigmasqhat <- (sigmasqhat * n)/(n -
						beta.size)
				nuggethat <- lik.results$parameters[1] * 
					sigmasqhat
				lik.results$parameters <- temp.pars <- 
					as.vector(c(nuggethat, sigmasqhat,
					lik.results$parameters[2]))
			}
			if((automatic.refit == T & (lik.results$parameters[
				3] <= lower.phi)) | lik.results$parameters[
				3] < 1e-08) {
				if(messages.screen == TRUE) {
					cat("likfit: WARNING: phi estimate < minimum value allowed\n"
						)
					cat("likfit: model re-fitted without spatial correlation (phi=0)\n"
						)
				}
				reduce.pars <- 2
				lik.results <- likfit.nospatial(temp.list,
					...)
				lambda <- lik.results$lambda
				if(fix.nugget == T) {
					lik.results$parameters <- temp.pars <-
						as.vector(c(ftau, (lik.results$
						tausqhat - ftau), 0))
				}
				else {
					lik.results$parameters <- temp.pars <-
						as.vector(c(lik.results$
						tausqhat, 0, 0))
				}
			}
		}
		else {
			# case 3: parameters = $(\sigma^2, \phi)$ ; fixed nugget: $\tau^2= 0$
			ini.m <- ini[2]
			if(fix.lambda == T) {
				lik.results <- nlminb(ini.m, proflik.phi, lower
					 = lower.phi, upper = upper.phi, 
					temp.list = temp.list, control = 
					control, ...)
			}
			else {
				lik.results <- nlminb(c(ini.m, lambda), 
					proflik.phi, lower = c(lower.phi, -3
					), upper = c(upper.phi, 3), temp.list
					 = temp.list, control = control, ...)
				lambda <- as.vector(lik.results$parameters[
					2])
				lik.results$parameters <- as.vector(lik.results$
					parameters[1])
				if(lambda == 0)
					z <- log(as.vector(data))
				else z <- (((as.vector(data))^lambda) - 1)/
						lambda
			}
			if(messages.screen == T) {
				print(lik.results$message)
			}
			if(automatic.refit == T & (lik.results$parameters <=
				lower.phi)) {
				cat("likfit: WARNING: phi estimate < minimum value allowed\n"
					)
				cat("likfit: model re-fitted without spatial correlation (phi=0)\n"
					)
				reduce.pars <- 1
				lik.results <- likfit.nospatial(temp.list,
					...)
				lambda <- lik.results$lambda
				lik.results$parameters <- temp.pars <- 
					as.vector(c(0, lik.results$tausqhat,
					0))
			}
			else {
				if(lik.results$parameters < 1e-08)
					icovhat <- diag(n)
				else icovhat <- varcov.spatial(coords = coords,
						cov.model = cov.model, kappa = 
						kappa, nugget = 0, cov.pars = c(
						1, lik.results$parameters),
						inv = T, det = F)
				txiv <- crossprod(xmat, icovhat)
				sigmasqhat <- (z %*% (icovhat - crossprod(
					txiv, solve(txiv %*% xmat)) %*% txiv) %*%
					z)/n
				if(method == "RML")
					sigmasqhat <- (sigmasqhat * n)/(n -
						beta.size)
				temp.pars <- as.vector(c(0, sigmasqhat, 
					lik.results$parameters))
				lik.results$parameters <- as.vector(c(
					sigmasqhat, lik.results$parameters))
			}
		}
	}
	if(messages.screen == T) {
		cat("likfit: end of likelihood maximisation\n")
	}
	loglik <-  - lik.results$objective
	npars <- length(trend.ols$coefficients) + length(ini) - reduce.pars
	if(fix.lambda == F)
		npars <- npars + 1
	AIC <- loglik - npars
	BIC <- loglik - 0.5 * log(n) * npars
	if(any(temp.pars[2:3]) != 0)
		invcov <- varcov.spatial(coords = coords, cov.model = cov.model,
			kappa = kappa, nugget = temp.pars[1], cov.pars = 
			temp.pars[2:3], inv = T, det = F)
	else invcov <- diag((1/temp.pars[1]), n)
	txmatinvcov <- crossprod(xmat, invcov)
	beta <- solve(txmatinvcov %*% xmat) %*% txmatinvcov %*% z
	beta.var <- solve(txmatinvcov %*% xmat)
	if(residuals == T | predicted == T) {
		#    print("computing predicted values and residuals")
		trend.est <- as.vector(xmat %*% beta)
		residuals.trend <- as.vector(z - trend.est)
		if(any(temp.pars[2:3]) != 0)
			covmat.signal <- varcov.spatial(coords = coords, 
				cov.model = cov.model, kappa = kappa, nugget = 
				0, cov.pars = temp.pars[2:3])
		else covmat.signal <- diag(n)
		signal.est <- as.vector(covmat.signal %*% invcov %*% 
			residuals.trend)
		predict.est <- trend.est + signal.est
		residuals.est <- as.vector(z - predict.est)
		residuals.std <- as.vector(invcov %*% residuals.est)
		residuals.trend.std <- as.vector(invcov %*% residuals.trend)
		s2.trend <- (crossprod(residuals.trend, invcov) %*% 
			residuals.trend)/(n - length(beta))
		s2 <- (crossprod(residuals.est, invcov) %*% residuals.est)/
			(n - length(beta))
	}
	if(messages.screen == T)
		cat("likfit: preparing output\n")
	results <- list()
	results$cov.model <- cov.model
	results$nugget <- temp.pars[1]
	results$cov.pars <- as.vector(c(sigmasq = temp.pars[2], phi = temp.pars[
		3]))
	if(is.null(kappa))
		results$kappa <- "not used"
	else results$kappa <- kappa
	results$beta <- as.vector(beta)
	results$beta.var <- beta.var
	if(length(results$beta.var) == 1)
		results$beta.var <- as.vector(results$beta.var)
	if(length(results$beta) > 1) {
		if(inherits(trend, "formula"))
			beta.names <- c("intercept", paste("covar", 1:(ncol(
				xmat) - 1), sep = ""))
		else if(trend == "1st")
			beta.names <- c("1", "x", "y")
		else if(trend == "2nd")
			beta.names <- c("1", "x", "y", "x2", "xy", "y2")
		names(results$beta) <- beta.names
	}
	results$lambda <- lambda
	results$loglik <- loglik
	results$npars <- npars
	results$AIC <- AIC
	results$BIC <- BIC
	results$trend.ols <- trend.ols$coefficients
	names(results$trend.ols) <- names(results$beta)
	if(residuals == T) {
		results$s2 <- s2
		results$s2.trend <- s2.trend
	}
	if(predicted == T)
		results$predicted <- cbind(predicted = predict.est, trend.est
			 = trend.est, signal.est = signal.est)
	if(residuals == T)
		results$residuals <- round(cbind(residuals = residuals.est,
			resid.trend = residuals.trend, resid.std = 
			residuals.std, resid.trend.std = residuals.trend.std),
			dig = 12)
	if(fix.lambda == F) {
		if(lambda == 1) {
			log.jacobian <- 0
		}
		else {
			if(any(data^(lambda - 1) <= 0))
				log.jacobian <- log(prod(data^(lambda - 1)))
			else log.jacobian <- sum(log(data^(lambda - 1)))
		}
	}
	results$info.lambda <- list(fix.lambda = fix.lambda, log.jacobian = 
		log.jacobian)
	lik.results$parameters <- NULL
	lik.results$aux <- NULL
	lik.results$objective <- NULL
	results$nlminb.inf <- lik.results
	results$max.dist <- max.dist
	results$trend.matrix <- xmat
	results$method <- method
	results$call <- call
	class(results) <- "variomodel"
	if(messages.screen == T) {
		cat("likfit: model parameters are:\n")
		cat(paste("covariance model:", cov.model))
		if(cov.model == "matern" | cov.model == "powered.exponential" |
			cov.model == "cauchy" | cov.model == "gneiting.matern"
			)
			cat(paste(" with kappa =", kappa))
		if(!is.null(kappa))
			if(cov.model == "matern" & kappa == 0.5)
				cat(" (exponential)")
		cat("\n")
		print(c(nugget = results$nugget, sill = results$cov.pars[1],
			range = results$cov.pars[2]))
		if(fix.lambda == F)
			cat(paste("Box-Cox transformation parameter:", nice(
				results$lambda), "\n"))
		if((results$cov.pars[1] < (0.01 * (results$nugget + results$
			cov.pars[1]))) & (results$cov.pars[2] > 0))
			cat("\nWARNING: estimated sill is less than 1 hundredth of the total variance (nugget + sill). Consider re-examine the model excluding spatial dependence\n"
				)
		if((results$cov.pars[2] > (10 * max.dist)) & (results$cov.pars[
			1] > 0))
			cat("\nWARNING: estimated range is more than 10 times bigger than the biggest distance between two points. Consider re-examine the model:\n 1) excluding spatial dependence if estimated sill is too low and/or \n 2) taking trends (covariates) into account\n"
				)
		if((results$cov.pars[2] < (0.10000000000000001 * min.dist)) &
			(results$cov.pars[1] > 0))
			cat("\nWARNING: estimated range is less than 1 tenth of the minimum distance between two points. Consider re-examine the model excluding spatial dependence\n"
				)
	}
	return(results)
}

"likfit.nospatial" <- 
#
function(temp.list, ...)
{
	results <- list()
	z <- temp.list$z
	n <- temp.list$n
	beta.size <- temp.list$beta.size
	xmat <- temp.list$xmat
	txmat <- temp.list$txmat
	ixx <- solve(crossprod(xmat))
	if(temp.list$fix.lambda == F) {
		results <- nlminb(1, proflik.lambda, lower = -2, upper = 2,
			temp.list = temp.list, control = temp.list$control,
			...)
		results$lambda <- results$parameters
		if(results$lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z <= 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(results$lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(results$
					lambda - 1)))
			else temp.list$log.jacobian <- sum(log(z^(results$
					lambda - 1)))
			if(results$lambda == 0)
				z <- log(z)
			else z <- ((z^results$lambda) - 1)/results$lambda
		}
	}
	else {
		results$lambda <- temp.list$lambda
		results$message <- "RELATIVE FUNCTION CONVERGENCE"
	}
	ssres <- (z %*% (diag(n) - xmat %*% solve(crossprod(xmat)) %*% txmat) %*%
		z)
	if(temp.list$method == "ML") {
		results$tausqhat <- ssres/n
		if(temp.list$fix.lambda == T)
			results$objective <- ((n/2) * log(2 * pi) + (n/2) *
				log(results$tausqhat) + (n/2) - temp.list$
				log.jacobian)
	}
	if(temp.list$method == "RML") {
		results$tausqhat <- (ssres/(n - beta.size))
		if(temp.list$fix.lambda == T) {
			results$objective <- (((n - beta.size)/2) * log(2 *
				pi) + ((n - beta.size)/2) * log(results$
				tausqhat) + (n/2) - temp.list$log.jacobian)
		}
	}
	return(results)
}

"lines.grf" <- 
function(obj, max.dist = max(dist(obj$coords)), length = 100, lwd = 2, ...)
{
	if(obj$cov.model == "matern" | obj$cov.model == "powered.exponential" |
		obj$cov.model == "cauchy" | obj$cov.model == "gneiting-matern"
		)
		kappa <- obj$kappa
	else kappa <- NULL
	distance <- seq(0, max.dist, length = length)
	if(is.vector(obj$cov.pars))
		sill.tot <- obj$nugget + obj$cov.pars[1]
	else sill.tot <- obj$nugget + sum(obj$cov.pars[, 1])
	gamma <- sill.tot - cov.spatial(distance, cov.model = obj$cov.model,
		kappa = kappa, cov.pars = obj$cov.pars)
	lines(distance, gamma, lwd = lwd, ...)
}

"lines.grfsim" <- 
function(obj, max.dist = max(dist(obj$coords)), length = 100, lwd = 2, ...)
{
	if(obj$cov.model == "matern" | obj$cov.model == "powered.exponential" |
		obj$cov.model == "cauchy" | obj$cov.model == "gneiting-matern"
		)
		kappa <- obj$kappa
	else kappa <- NULL
	distance <- seq(0, max.dist, length = length)
	if(is.vector(obj$cov.pars))
		sill.tot <- obj$nugget + obj$cov.pars[1]
	else sill.tot <- obj$nugget + sum(obj$cov.pars[, 1])
	gamma <- sill.tot - cov.spatial(distance, cov.model = obj$cov.model,
		kappa = kappa, cov.pars = obj$cov.pars)
	lines(distance, gamma, lwd = lwd, ...)
}

"lines.grid.bayes" <- 
function(obj, max.dist, length = 100, summary.posterior = c("mode", "median",
	"mean"), ...)
{
	spost <- match.arg(summary.posterior)
	if(is.null(obj$call$cov.model))
		cov.model <- "exponential"
	else {
		cov.model <- obj$call$cov.model
		if(obj$call$cov.model == "matern" | obj$call$cov.model == 
			"powered.exponential" | obj$call$cov.model == "cauchy" |
			obj$call$cov.model == "gneiting-matern")
			kappa <- obj$call$kappa
		else kappa <- NULL
	}
	distance <- seq(0, max.dist, length = length)
	cov.pars <- c(obj$sigmasq.summary[spost], obj$phi.summary[spost])
	names(cov.pars) <- NULL
	if(any(obj$tausq.summary != 0))
		nugget <- obj$tausq.summary[spost] * cov.pars[1]
	else nugget <- 0
	names(nugget) <- NULL
	sill.tot <- nugget + cov.pars[1]
	gamma <- (nugget + cov.pars[1]) - cov.spatial(distance, cov.model = 
		cov.model, kappa = kappa, cov.pars = cov.pars)
	lines(distance, gamma, ...)
}

"lines.krige.bayes" <- 
function(obj, max.dist, length = 100, summary.posterior = c("mode", "median",
	"mean"), ...)
{
	spost <- match.arg(summary.posterior)
	if(is.null(obj$call$cov.model))
		cov.model <- "exponential"
	else {
		cov.model <- obj$call$cov.model
		if(obj$call$cov.model == "matern" | obj$call$cov.model == 
			"powered.exponential" | obj$call$cov.model == "cauchy" |
			obj$call$cov.model == "gneiting-matern")
			kappa <- obj$call$kappa
		else kappa <- NULL
	}
	distance <- seq(0, max.dist, length = length)
	if(summary.posterior == "mode")
		spost1 <- "mode.cond"
	else spost1 <- spost
	cov.pars <- c(obj$sigmasq.summary[spos1], obj$phi.summary[spost])
	names(cov.pars) <- NULL
	if(any(obj$tausq.summary != 0))
		nugget <- obj$tausq.summary[spost] * cov.pars[1]
	else nugget <- 0
	names(nugget) <- NULL
	sill.tot <- nugget + cov.pars[1]
	gamma <- (nugget + cov.pars[1]) - cov.spatial(distance, cov.model = 
		cov.model, kappa = kappa, cov.pars = cov.pars)
	lines(distance, gamma, ...)
}

"lines.variogram" <- 
function(obj, max.dist = max(obj$u), type = "o", scaled = F, ...)
{
	if(scaled == T)
		obj$v <- obj$v/obj$var.mark
	if(!is.matrix(obj$v))
		lines(obj$u[obj$u <= max.dist], obj$v[obj$u <= max.dist], type
			 = type, ...)
	else {
		for(j in 1:ncol(obj$v))
			lines(obj$u[obj$u <= max.dist], obj$v[obj$u <= max.dist,
				j], type = type, ...)
	}
}

"lines.variomodel" <- 
function(obj, max.dist = obj$max.dist, length = 100, scaled = F, ...)
{
	if(scaled == T)
		obj$v <- obj$v/obj$var.mark
	if(is.null(max.dist))
		stop("argument max.dist needed for this object")
	if(obj$cov.model == "matern" | obj$cov.model == "powered.exponential" |
		obj$cov.model == "cauchy" | obj$cov.model == "gneiting-matern"
		)
		kappa <- obj$kappa
	else kappa <- NULL
	distance <- seq(0, max.dist, length = length)
	if(is.vector(obj$cov.pars))
		sill.tot <- obj$nugget + obj$cov.pars[1]
	else sill.tot <- obj$nugget + sum(obj$cov.pars[, 1])
	if(obj$cov.pars[2] < 9.9999999999999998e-13)
		gamma <- rep(sill.tot, length)
	else gamma <- sill.tot - cov.spatial(distance, cov.model = obj$
			cov.model, kappa = kappa, cov.pars = obj$cov.pars)
	lines(distance, gamma, ...)
}

"loglik.spatial" <- 
# This function computes the log-likelihood for a (transformed) Gaussian random field for known parameters
function(pars, temp.list)
{
	pars <- as.vector(pars)
	tausq <- pars[1]
	sigmasq <- pars[2]
	phi <- pars[3]
	lambda <- pars[4]
	z <- temp.list$z
	n <- temp.list$n
	if(temp.list$fix.lambda == F) {
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z < 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	txmat <- temp.list$txmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = tausq,
		cov.pars = c(sigmasq, phi), scale = F, inv = T, det = T, 
		sqrt.inv = T)
	six <- crossprod(covinf$sqrt.inverse, xmat)
	xix <- crossprod(six)
	if(length(as.vector(xix)) == 1) {
		ixix <- 1/xix
		choldet <- 0.5 * log(xix)
	}
	else {
		chol.xix <- chol(xix)
		i.chol.xix <- solve.upper(chol.xix)
		ixix <- i.chol.xix %*% t(i.chol.xix)
		choldet <- sum(log(diag(chol.xix)))
	}
	xi <- crossprod(xmat, covinf$inverse)
	ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
	if(temp.list$method == "ML") {
		loglik <- ( - (n/2) * log(2 * pi) - covinf$log.det.to.half -
			0.5 * ssresmat + temp.list$log.jacobian)
	}
	if(temp.list$method == "RML") {
		xx.eigen <- eigen(crossprod(xmat), symmetric = T, only.values
			 = T)
		loglik <- ( - ((n - beta.size)/2) * log(2 * pi) + 0.5 * sum(
			log(xx.eigen$values)) - covinf$log.det.to.half - (
			0.5) * ssresmat - choldet + temp.list$log.jacobian)
	}
	#	}
	return(as.vector(loglik))
}

"loss.olsvario" <- 
function(theta, data, fix.nugget, nugget, cov.model)
{
	u <- data$u
	v <- data$v
	kappa <- data$kappa
	if(fix.nugget == F) {
		tausq <- theta[1]
		sigmasq <- theta[2]
		phi <- theta[3]
		tau1 <- tausq
	}
	else {
		tausq <- nugget
		sigmasq <- theta[1]
		phi <- theta[2]
		tau1 <- tausq + 1
	}
	sill.tot <- sigmasq + tausq
	if(fix.nugget == F) {
		gamma <- sill.tot - cov.spatial(u, cov.model = cov.model, kappa
			 = kappa, cov.pars = c(theta[2], theta[3]))
	}
	else {
		gamma <- sill.tot - cov.spatial(u, cov.model = cov.model, kappa
			 = kappa, cov.pars = c(theta[1], theta[2]))
	}
	loss <- sum((v - gamma)^2)
	return(loss)
}

"loss.vario" <- 
function(theta, data)
{
	u <- data$u
	v <- data$v
	kappa <- data$kappa
	loss <- sum(v - vario.matern(u = u, tausq = theta[1], sigmasq = theta[
		2], phi = theta[3], kappa = kappa))^2
	return(loss)
}

"loss.wlsvario" <- 
function(theta, data, fix.nugget, nugget, cov.model, weight)
{
	u <- data$u
	v <- data$v
	n <- data$n
	kappa <- data$kappa
	if(fix.nugget == F) {
		tausq <- theta[1]
		sigmasq <- theta[2]
		phi <- theta[3]
		tau1 <- tausq
	}
	else {
		tausq <- nugget
		sigmasq <- theta[1]
		phi <- theta[2]
		tau1 <- tausq + 1
	}
	sill.tot <- sigmasq + tausq
	if(fix.nugget == F) {
		gamma <- sill.tot - cov.spatial(u, cov.model = cov.model, kappa
			 = kappa, cov.pars = c(theta[2], theta[3]))
	}
	else {
		gamma <- sill.tot - cov.spatial(u, cov.model = data$cov.model,
			kappa = data$kappa, cov.pars = c(theta[1], theta[2]))
	}
	if(weight == "npairs")
		loss <- sum(n * (v - gamma)^2)
	if(weight == "cressie")
		loss <- sum((n/(gamma^2)) * (v - gamma)^2)
	return(loss)
}

"matern" <- 
#
function(u, phi, kappa)
{
	if(kappa == 0.5) {
		covs <- exp( - u/phi)
	}
	else {
		if(kappa != floor(kappa)) {
			kappa <- floor(kappa)
			cat("kappa must be an interger value, changed for kappa =",
				kappa, "\n")
		}
		if(!is.loaded("Xwhittlematern"))
			dyn.load("/home/ribeiro/geoS/woodandchan.o")
		covs <- .C("Xwhittlematern",
			u = as.double(u),
			as.integer(length(u)),
			as.double(phi),
			as.double(kappa))$u
		# check == here
		covs[u == 0] <- 1
	}
	return(covs)
}

"model.control" <- 
function(trend.d = "cte", trend.l = "cte", cov.model = c("exponential", 
	"matern", "gaussian", "spherical", "circular", "cubic", "wave", 
	"powered.exponential", "cauchy", "gneiting", "gneiting.matern", 
	"pure.nugget"), kappa = 0.5, aniso.pars = NULL, lambda = 1)
{
	cov.model <- match.arg(cov.model)
	return(list(trend.d = trend.d, trend.l = trend.l, cov.model = cov.model,
		kappa = kappa, aniso.pars = aniso.pars, lambda = lambda))
}

"my.ps" <- 
function(file = "junk", height = 2.5, width = 5, labels = T, cex = 
	0.69999999999999996)
{
	ps.options(horizontal = F)
	ps.options(width = width, height = height)
	postscript(file = file, print.it = F)
	if(labels == F)
		par(mar = c(1.8, 1.8, 0, 0), oma = c(0, 0, 1.45, 0), cex = cex,
			las = 1, mgp = c(1.5, 0.5, 0))
	else par(mar = c(2.7000000000000002, 2.7000000000000002, 0, 
			0.10000000000000001), oma = c(0, 0, 1.45, 0), cex = cex,
			las = 1, mgp = c(2, 0.20000000000000001, 0))
	return(invisible())
}

"mypostscript" <- 
function(file = NULL, width = -1, height = -1, append = F, onefile = T, 
	print.it = NULL, horizontal = F, psize = 11, ...)
{
	.ps.options <- lapply(ps.options(), eval)
	xtras <- list(...)
	if(length(xtras) > 0) {
		# Normalize the names.
		.names <- names(xtras)
		.full.names <- names(.ps.options)
		.matched <- charmatch(.names, .full.names, nomatch = 0)
		.names[.matched > 0] <- paste("ps.", .full.names[.matched],
			sep = "")
		.full.names <- paste("ps.", .full.names, sep = "")
		.matched <- charmatch(.names, .full.names, nomatch = 0)
		.names[.matched > 0] <- .full.names[.matched]
		.dup <- duplicated(.names)
		if(any(.dup)) {
			for(i in names(xtras)[.dup])
				warning(paste("Discarding duplicate option:",
					i))
			xtras <- xtras[!.dup]
			names(xtras) <- .names[!.dup]
		}
		else names(xtras) <- .names
	}
	if(!is.null(xtras$ps.defaults))
		.ps.options <- xtras$ps.defaults
	.paper <- if(!is.null(xtras$ps.paper)) xtras$ps.paper else .ps.options$
			paper
	.paper <- casefold(.paper)
	.pointsize <- if(is.null(xtras$ps.pointsize)) .ps.options$pointsize
		 else xtras$ps.pointsize
	.region <- if(is.null(xtras$ps.region)) .ps.options$region else xtras$
			ps.region
	if(length(.region) == 1 && .region == -1) {
		if(exists("ps.paper.regions")) {
			.reg.names <- casefold(dimnames(ps.paper.regions)[[
				1]])
			.idx <- match(.paper, .reg.names)
			if(!is.na(.idx))
				.region <- ps.paper.regions[.idx,  ]
		}
	}
	.horizontal <- if(is.null(xtras$ps.horizontal)) .ps.options$horizontal
		 else xtras$ps.horizontal
	if(missing(print.it))
		print.it <- -1
	else if(is.logical(print.it))
		print.it <- if(print.it) 1 else 0
	n <- nchar(.paper)
	if(missing(width))
		width <- .ps.options$width
	if(missing(height))
		height <- .ps.options$height
	tpar <- c(width, height, .horizontal, print.it, .pointsize, rep(.region,
		length = 4), n)
	if(n > 0)
		tpar <- c(tpar, as.single(AsciiToInt(.paper)))
	n <- if(is.null(file)) 0 else nchar(file)
	tpar <- c(tpar, n)
	if(n > 0)
		tpar <- c(tpar, as.single(AsciiToInt(file)))
	argv <- c(list(name = "postscript", pars = as.single(tpar), npars = 
		length(tpar)), xtras)
	do.call("pipegraph", argv)
	.C("S_send_ps_int",
		as.integer(if(append) 1 else 0),
		as.character("la"))
	.C("S_send_ps_int",
		as.integer(if(onefile) 1 else 0),
		as.character("lo"))
	ps.options(point.size = psize)
	invisible()
}

"nice" <- 
function(x, decimal = 2, fixed = F)
{
	ergb <- x
	index <- (x != 0) & is.finite(x)
	if(fixed)
		n <- 0
	else n <- floor(log(abs(x[index]))/log(10))
	ergb[index] <- trunc(x[index]/10^(n - decimal)) * 10^(n - decimal)
	return(ergb)
}

"olsfit" <- 
function(vario, ini, cov.model = c("exponential", "matern", "gaussian", 
	"spherical", "cubic", "wave", "powered.exponential", "cauchy", 
	"gneiting", "gneiting.matern", "pure.nugget"), kappa = NULL, 
	simul.number = NULL, max.dist = "all", minimisation.function = "nlminb",
	lower = 0, fix.nugget = F, nugget = 0, messages.screen = TRUE)
{
	#
	# fit variogram model by OLS applied to variogram cloud 
	#
	# function will also work on a binned or kernel-smoothed variogram cloud, 
	# which is much quicker computationally although I can think of
	# no good statistical reason for not using the unsmoothed cloud
	#
	# for binned data, see wlsfit()
	#
	# vario : list with u and v components
	# ini : inial values of matern parameters theta=(tausq, sigmasq, phi)
	#         (PJD's parameterisation)
	# If a matrix of initial bvalues is provided, and minimisation.function="nlminb" a search for the best initial value is performed
	# kappa : FIXED value of kappa (default corresponds to exponential, 
	#         legal non-default values are positive integers for Matern or 0 < kappa =< 2 for powered exponential
	# max.dist : maximum distance in u component
	# minimisation.function : S-PLUS minimization procedure. options 'nls' or 'nlminb' (default)
	# lower : minimum value (or vector of minimum values) for parameters. 
	# This allows constrained minimization when using 'nlminb'
	# 
	# results is list with components:
	#   theta : estimated value of theta
	#   kappa : fixed value of kappa used
	#   value : minimised sum of squares (to allow informal optimisation
	#           over different values of kappa)
	#   minimisation.function : minimization functio used
	#   call : function call
	# simul.number : select one empirical variogram if the object in vario has more than one
	call <- match.call()
	if(minimisation.function != "nlminb" & minimisation.function != "nls")
		stop("minimization functions available are: nlminb or nls")
	if(is.matrix(vario$v) & is.null(simul.number))
		stop("object in vario$v is a matrix. This function works for only 1 empirical variogram"
			)
	if(!is.null(simul.number))
		vario$v <- vario$v[, simul.number]
	if(is.matrix(ini)) {
		inilength <- dim(ini)[2]
		if(fix.nugget == F & inilength != 3)
			stop("wrong number of columns for ini (must be 3)")
		if(fix.nugget == T & inilength != 2)
			stop("wrong number of columns for ini (must be 2)")
	}
	else {
		inilength <- length(ini)
		if(fix.nugget == F & inilength != 3)
			stop("wrong length for ini (must be 3)")
		if(fix.nugget == T & inilength != 2)
			stop("wrong length for ini (must be 2)")
	}
	cov.model <- match.arg(cov.model)
	if(max.dist == "all")
		data <- list(u = vario$u, v = vario$v)
	else data <- list(u = vario$u[vario$u <= max.dist], v = vario$v[vario$
			u <= max.dist])
	data$kappa <- kappa
	data$cov.model <- cov.model
	if(minimisation.function == "nls") {
		if(fix.nugget == F) {
			result <- nls(v ~ ((sigmasq + tausq) - cov.spatial(
				u, cov.model = cov.model, kappa = kappa, 
				cov.pars = c(sigmasq, phi))), start = list(
				tausq = ini[1], sigmasq = ini[2], phi = ini[
				3]), data = data)
			nugget <- result$parameter[1]
			cov.pars <- as.vector(result$parameter[2:3])
			value <- vario$v - ((nugget + cov.pars[1]) - 
				cov.spatial(vario$u, cov.model = cov.model,
				kappa = kappa, cov.pars = cov.pars))
		}
		else {
			data$nugget <- nugget
			result <- nls(v ~ ((sigmasq + nugget) - cov.spatial(
				u, cov.model = cov.model, kappa = kappa, 
				cov.pars = c(sigmasq, phi))), start = list(
				sigmasq = ini[1], phi = ini[2]), data = data)
			cov.pars <- as.vector(result$parameter[1:2])
			value <- vario$v - ((nugget + cov.pars[1]) - 
				cov.spatial(vario$u, cov.model = cov.model,
				kappa = kappa, cov.pars = cov.pars))
		}
		value <- sum(value * value)
		message <- "nls does not provides convergence message"
	}
	if(minimisation.function == "nlminb") {
		if(is.matrix(ini) | is.data.frame(ini)) {
			ini <- as.matrix(ini)
			cat("olsfit: searching for the best initial value\n")
			ini.search <- ini
			dimnames(ini.search) <- list(NULL, NULL)
			ols.ini <- round(100000000. * apply(ini.search, 1,
				loss.olsvario, data = data, fix.nugget = 
				fix.nugget, nugget = nugget, cov.model = 
				cov.model))
			ini <- as.vector(ini.search[ols.ini == min(ols.ini),
				])
			cat("olsfit: best initial value:\n")
			if(fix.nugget == T) {
				names(ini) <- c("sill", "range")
				print(ini)
			}
			else {
				names(ini) <- c("nugget", "sill", "range")
				print(ini)
			}
		}
		result <- nlminb(ini, loss.olsvario, lower = lower, data = data,
			fix.nugget = fix.nugget, nugget = nugget, kappa = kappa,
			cov.model = cov.model)
		value <- result$objective
		message <- result$message
		if(fix.nugget == F) {
			nugget <- result$parameter[1]
			cov.pars <- as.vector(result$parameter[2:3])
		}
		else {
			cov.pars <- as.vector(result$parameter[1:2])
		}
	}
	dmax <- max(vario$u)
	estimation <- list(nugget = nugget, cov.pars = cov.pars, cov.model = 
		cov.model, kappa = kappa, value = value, max.dist = dmax, 
		minimisation.function = minimisation.function, message = 
		message)
	estimation$method <- "OLS"
	estimation$call <- call
	class(estimation) <- "variomodel"
	if(messages.screen) {
		cat("olsfit: estimated model parameters are:\n")
		cat(paste("covariance model:", cov.model))
		if(cov.model == "matern" | cov.model == "powered.exponential" |
			cov.model == "cauchy" | cov.model == "gneiting.matern"
			)
			cat(paste(" with kappa =", kappa))
		if(!is.null(kappa))
			if(cov.model == "matern" & kappa == 0.5)
				cat(" (exponential)")
		cat("\n")
		print(c(nugget = estimation$nugget, sill = estimation$cov.pars[
			1], range = estimation$cov.pars[2]))
	}
	return(estimation)
}

"output.control" <- 
function(n.posterior = 1000, n.predictive = NULL, simulations.predictive = TRUE,
	keep.simulations = TRUE, mean.estimator = TRUE, quantile.estimator = 
	NULL, probability.estimator = NULL, messages.screen = TRUE, signal = 
	FALSE, moments = TRUE)
{
	return(list(n.posterior = n.posterior, n.predictive = n.predictive,
		moments = moments, simulations.predictive = 
		simulations.predictive, keep.simulations = keep.simulations,
		mean.estimator = mean.estimator, quantile.estimator = 
		quantile.estimator, probability.estimator = 
		probability.estimator, signal = signal, messages.screen = 
		messages.screen))
}

"persp.grf" <- 
function(obj, sim.number = 1, ...)
{
	# plots an 'perspective plot' from a (selected) simulation of a random field generated by the function 'grf'
	# The simulated random filed must be regularly spaced
	x <- as.numeric(levels(as.factor(obj$coords[, 1])))
	nx <- length(x)
	y <- as.numeric(levels(as.factor(obj$coords[, 2])))
	ny <- length(y)
	obj$data <- as.matrix(obj$data)
	n <- nrow(obj$data)
	if(nx * ny != n)
		stop("Probably irregular grid")
	m <- matrix(obj$data[, sim.number], ncol = ny)
	persp(x, y, m, ...)
	return(invisible())
}

"persp.grfsim" <- 
function(obj, sim.number = 1, ...)
{
	# plots an 'perspective plot' from a (selected) simulation of a random field generated by the function 'grf'
	# The simulated random filed must be regularly spaced
	x <- as.numeric(levels(as.factor(obj$coords[, 1])))
	nx <- length(x)
	y <- as.numeric(levels(as.factor(obj$coords[, 2])))
	ny <- length(y)
	obj$data <- as.matrix(obj$data)
	n <- nrow(obj$data)
	if(nx * ny != n)
		stop("Probably irregular grid")
	m <- matrix(obj$data[, sim.number], ncol = ny)
	persp(x, y, m, ...)
	return(invisible())
}

"plot.geodata" <- 
function(geodata, coords = geodata$coords, data = geodata$data, trend = "cte",
	lambda = 1, col.data = 1, window = F, ...)
{
	#
	# This function provides a 2*2 graphic display for geostatistical data (typically objects from the class "geodata")
	# The plots returned are:
	## - A plot with data locations. Symbols, theirs sizes (and colors) sepatates data from diferent quartiles as follows
	###  blue (circles) : 1st quantile
	###  green (triangles) : 2nd
	###  yellow (plus)  : 3rd
	###  red (crosses)  : 4th
	## - A plot with a empirical variogram. Arguments for the function variog can be passed through ...
	## - A plot with data-values agains coordinate X
	## - A plot with data-values agains coordinate Y
	par.ori <- par()
	coords <- as.matrix(coords)
	data <- as.matrix(data)
	data <- data[, col.data]
	if(lambda != 1) {
		if(lambda == 0)
			data <- log(data)
		else data <- ((data^lambda) - 1)/lambda
	}
	if(window == T)
		trellis.device()
	#    motifA4()
	par(mfrow = c(2, 2))
	par(pty = "s")
	#  plot(coords, xlab = "Coord X", ylab = "Coord Y", pch = 3, cex = 1.2)	
	#,main='\nData locations'
	#  plot(coords, xlab = "Coord X", ylab = "Coord Y",main='\nData locations',pch=3)
	data.quantile <- quantile(data)
	plot(coords, xlab = "Coord X", ylab = "Coord Y", type = "n")
	if(length(unique(data.quantile)) == length(data.quantile)) {
		points(coords[(data <= data.quantile[2]),  ], pch = 1, cex = 
			0.59999999999999998, col = 2)
		points(coords[((data > data.quantile[2]) & (data <= 
			data.quantile[3])),  ], pch = 2, cex = 
			1.3999999999999999, col = 4)
		points(coords[((data > data.quantile[3]) & (data <= 
			data.quantile[4])),  ], pch = 3, cex = 1.7, col = 7)
		points(coords[(data > data.quantile[4]),  ], pch = 4, cex = 2,
			col = 8)
	}
	else {
		points(coords, pch = 16)
	}
	par(pty = "m")
	par(mar = c(0, 1, 0, 0.5))
	xyzplot(coords = coords, data = data, ...)
	#  plot(variog(coords = coords, data = data, trend = trend, ...))
	par(mar = c(5, 5, 1, 0.5))
	plot(coords[, 1], data, xlab = "Coord X", cex = 1)
	plot(coords[, 2], data, xlab = "Coord Y", cex = 1)
	par(par.ori)
	return(invisible())
}

"plot.grf" <- 
function(obj, grid = F, model = T, ylim = "default", ...)
{
	## This function plots the variograms for simulated data generated using the function 'grf'
	nsim <- dim(obj[[2]])[2]
	if(grid == T)
		plot(obj$coords[, 1], obj$coords[, 2], xlab = "Coord X", ylab
			 = "Coord Y")
	if(is.vector(obj$cov.pars))
		sill.tot <- obj$nugget + obj$cov.pars[1]
	else sill.tot <- obj$nugget + sum(obj$cov.pars[, 1])
	sim.bin <- variog(obj)
	if(is.numeric(ylim))
		plot(sim.bin, ylim = ylim, var.lines = F, ...)
	else {
		ymax <- max(1.1000000000000001 * (sill.tot), sim.bin$v)
		plot(sim.bin, ylim = c(0, ymax), var.lines = F, ...)
	}
	if(model == T) {
		max.dist <- max(sim.bin$u)
		var.model <- list(nugget = obj$nugget, cov.pars = obj$cov.pars,
			kappa = obj$kappa, max.dist = max.dist, cov.model = obj$
			cov.model)
		lines.variomodel(var.model, lwd = 3, ...)
	}
	return(invisible())
}

"plot.grfsim" <- 
function(obj, grid = F, model = T, ylim = "default", ...)
{
	# This function plots the variograms for simulated data generated using the function 'grf'
	nsim <- dim(obj[[2]])[2]
	if(grid == T)
		plot(obj$coords[, 1], obj$coords[, 2], xlab = "Coord X", ylab
			 = "Coord Y")
	if(is.vector(obj$cov.pars))
		sill.tot <- obj$nugget + obj$cov.pars[1]
	else sill.tot <- obj$nugget + sum(obj$cov.pars[, 1])
	sim.bin <- variog(obj)
	if(is.numeric(ylim))
		plot(sim.bin, ylim = ylim, var.lines = F, ...)
	else {
		ymax <- max(1.1000000000000001 * (sill.tot), sim.bin$v)
		plot(sim.bin, ylim = c(0, ymax), var.lines = F, ...)
	}
	if(model == T) {
		max.dist <- max(sim.bin$u)
		var.model <- list(nugget = obj$nugget, cov.pars = obj$cov.pars,
			kappa = obj$kappa, max.dist = max.dist, cov.model = obj$
			cov.model)
		lines.variomodel(var.model, lwd = 3, ...)
	}
	return(invisible())
}

"plot.proflik" <- 
function(obj.proflik, pages = c("user", "one", "two"), uni.only = F, bi.only = 
	F, type.bi = c("contour", "persp"), conf.int = c(0.90000000000000002,
	0.94999999999999996), yaxis.lims = c("conf.int", "as.computed"), by.col
	 = T, log.scale = F, par.mar.persp = c(9.5, 1, 1, 0), ...)
{
	#  par.mfrow.ori <- par()$mfrow
	#  par.mfcol.ori <- par()$mfcol
	pages <- match.arg(pages)
	if(all(is.character(yaxis.lims)))
		yaxis.lims <- match.arg(yaxis.lims)
	type.bi <- match.arg(type.bi)
	n.uni <- obj.proflik$n.uni
	n.bi <- obj.proflik$n.bi
	if(n.bi == 0)
		uni.only <- T
	if((uni.only == F) & (bi.only == F))
		np <- n.uni + n.bi
	if((uni.only == T) & (bi.only == F))
		np <- n.uni
	if((uni.only == F) & (bi.only == T))
		np <- n.bi
	if(n.uni == 0 & np > 0)
		bi.only <- T
	if(n.bi == 0 & np > 0)
		uni.only <- T
	if(pages == "one") {
		if(np >= 1 & np < 4)
			par(mfrow = c(np, 1))
		if(np >= 4) {
			if(by.col == T)
				par(mfcol = c(ceiling(np/2), 2))
			else par(mfrow = c(ceiling(np/2), 2))
		}
	}
	if(pages == "two") {
		if(n.uni > 1 & n.uni < 4)
			par(mfrow = c(n.uni, 1))
		if(n.uni >= 4)
			par(mfrow = c(ceiling(n.uni/2), 2))
	}
	if(bi.only == F) {
		for(i in 1:n.uni) {
			if(obj.proflik$method == "ML")
				ylabm <- "profile log-likelihood"
			else ylabm <- "profile log-(restricted) likelihood"
			if(all(conf.int) != F) {
				if(!is.numeric(conf.int) | any(conf.int > 1))
					stop("argument conf.int must be numerical (scalar or vector) with values between 0 and 1"
						)
				conf.int.drop <- obj.proflik[[i]][[3]][2] -
					0.5 * qchisq(conf.int, 1)
			}
			if(all(is.character(yaxis.lims))) {
				if(yaxis.lims == "conf.int")
					lik.lims <- c(min(conf.int.drop), 
						obj.proflik[[i]][[3]][2])
				else lik.lims <- c(min(obj.proflik[[i]][[2]]),
						obj.proflik[[i]][[3]][2])
			}
			else lik.lims <- yaxis.lims
			if(log.scale == T) {
				plot(log(obj.proflik[[i]][[1]]), obj.proflik[[
					i]][[2]], type = "b", xlab = paste(
					"log-", names(obj.proflik[[i]])[1]),
					ylab = ylabm, ylim = lik.lims, ...)
				lines(log(c(obj.proflik[[i]][[3]][1], 
					obj.proflik[[i]][[3]][1])), c(min(
					lik.lims), obj.proflik[[i]][[3]][2]),
					lty = 2)
			}
			else {
				plot(obj.proflik[[i]][[1]], obj.proflik[[i]][[
					2]], type = "b", xlab = names(
					obj.proflik[[i]])[1], ylab = ylabm,
					ylim = lik.lims, ...)
				lines(c(obj.proflik[[i]][[3]][1], obj.proflik[[
					i]][[3]][1]), c(min(lik.lims), 
					obj.proflik[[i]][[3]][2]), lty = 2)
			}
			abline(h = conf.int.drop, lty = 3)
		}
	}
	if(uni.only == F) {
		if(pages == "two") {
			if(n.bi >= 1 & n.bi < 4)
				par(mfrow = c(n.bi, 1))
			if(n.bi >= 4)
				par(mfrow = c(ceiling(n.bi/2), 2))
		}
		for(i in 1:n.bi) {
			if(type.bi == "contour") {
				if(log.scale == T) {
					contour(log(obj.proflik[[(n.uni + i)]][[
						1]]), log(obj.proflik[[(n.uni +
						i)]][[2]]), matrix(obj.proflik[[
						(n.uni + i)]][[3]], ncol = 
						length(obj.proflik[[(n.uni +
						i)]][[2]])), xlab = paste(
						"log-", names(obj.proflik[[
						(n.uni + i)]][1])), ylab = 
						paste("log-", names(obj.proflik[[
						(n.uni + i)]][2]), ...))
					points(log(t(obj.proflik[[(n.uni + i)]][[
						4]][1:2])))
				}
				else {
					contour(obj.proflik[[(n.uni + i)]][[
						1]], obj.proflik[[(n.uni + i)]][[
						2]], matrix(obj.proflik[[(
						n.uni + i)]][[3]], ncol = 
						length(obj.proflik[[(n.uni +
						i)]][[2]])), xlab = names(
						obj.proflik[[(n.uni + i)]][
						1]), ylab = names(obj.proflik[[
						(n.uni + i)]][2]), ...)
					points(t(obj.proflik[[(n.uni + i)]][[
						4]][1:2]))
				}
			}
			if(type.bi == "persp") {
				par.mar.ori <- par()$mar
				par(mar = par.mar.persp)
				if(obj.proflik$method == "ML")
					zlabm <- "profile log-likelihood"
				else zlabm <- "profile (restricted) log-likelihood"
				zlimm <- range(obj.proflik[[(n.uni + i)]][[
					3]])
				zlimm[1] <- 1.01 * zlimm[1]
				minlik <- min(obj.proflik[[(n.uni + i)]][[3]])
				if(log.scale == T) {
					pp <- persp(log(obj.proflik[[(n.uni +
						i)]][[1]]), log(obj.proflik[[
						(n.uni + i)]][[2]]), matrix(
						obj.proflik[[(n.uni + i)]][[
						3]], ncol = length(obj.proflik[[
						(n.uni + i)]][[2]])), xlab = 
						paste("log-", names(obj.proflik[[
						(n.uni + i)]][1])), ylab = 
						paste("log-", names(obj.proflik[[
						(n.uni + i)]][2])), zlab = 
						zlabm, box = F, ...)
					pp1 <- perspp(x = log(c(obj.proflik[[
						(n.uni + i)]][[4]][1], min(
						obj.proflik[[(n.uni + i)]][[
						1]]))[c(1, 1, 1, 2)]), y = log(
						c(obj.proflik[[(n.uni + i)]][[
						4]][2], min(obj.proflik[[(
						n.uni + i)]][[2]]))[c(1, 1,
						2, 1)]), z = c(minlik, 
						obj.proflik[[(n.uni + i)]][[
						4]][3])[c(1, 2, 1, 1)], pp)
					segments(log(pp1$x[1]), log(pp1$y[
						1]), log(pp1$x[2]), log(pp1$
						y[2]), lwd = 2)
				}
				else {
					pp <- persp(obj.proflik[[(n.uni + i)]][[
						1]], obj.proflik[[(n.uni + i)]][[
						2]], matrix(obj.proflik[[(
						n.uni + i)]][[3]], ncol = 
						length(obj.proflik[[(n.uni +
						i)]][[2]])), xlab = names(
						obj.proflik[[(n.uni + i)]][
						1]), ylab = names(obj.proflik[[
						(n.uni + i)]][2]), zlab = zlabm,
						box = F, ...)
					pp1 <- perspp(x = c(obj.proflik[[(
						n.uni + i)]][[4]][1], min(
						obj.proflik[[(n.uni + i)]][[
						1]]))[c(1, 1, 1, 2)], y = c(
						obj.proflik[[(n.uni + i)]][[
						4]][2], min(obj.proflik[[(
						n.uni + i)]][[2]]))[c(1, 1,
						2, 1)], z = c(minlik, 
						obj.proflik[[(n.uni + i)]][[
						4]][3])[c(1, 2, 1, 1)], pp)
					segments(pp1$x[1], pp1$y[1], pp1$x[
						2], pp1$y[2], lwd = 2)
				}
				par(mar = par.mar.ori)
			}
		}
	}
	#  par(mfrow=par.mfrow.ori)
	#  par(mfcol=par.mfcol.ori)
	return(invisible())
}

"plot.symbols" <- 
function(obj, data = obj$data, col.data = 1, polygon = NULL, window = F, ...)
{
	if(class(obj) == "geodata" | class(obj) == "grf") {
		coords <- as.matrix(obj$coords)
		data <- as.matrix(data)
		data <- data[, col.data]
	}
	else {
		if(is.list(obj)) {
			coords <- as.matrix(obj$coords)
			data <- as.matrix(obj$data)
			data <- data[, col.data]
			print("Object assumed to be a list with point coordinates in the first component and data in the second."
				)
		}
		else {
			if(is.data.frame(obj) | is.matrix(obj)) {
				coords <- as.matrix(obj[, 1:2])
				data <- obj[, 2 + col.data]
				print("Object assumed to be a matrix with point coordinates in the first and second columns."
					)
			}
		}
	}
	if(window == T)
		trellis.device()
	#  par(mfrow = c(2, 2))
	#  par.ori <- par()
	par(pty = "s")
	par(mar = c(5, 5, 0, 0))
	#  plot(coords, xlab = "Coord X", ylab = "Coord Y", pch = 3, cex = 1.2)	
	#,main='\nData locations'
	#  plot(coords, xlab = "Coord X", ylab = "Coord Y",main='\nData locations',pch=3)
	data.quantile <- quantile(data)
	if(!is.null(polygon)) {
		x.lim <- range(polygon[, 1])
		y.lim <- range(polygon[, 2])
		if(diff(x.lim) > diff(y.lim))
			y.lim <- y.lim + c( - ((diff(x.lim) - diff(y.lim))/
				2), (diff(x.lim) - diff(y.lim))/2)
		if(diff(y.lim) > diff(x.lim))
			x.lim <- x.lim + c( - ((diff(y.lim) - diff(x.lim))/
				2), (diff(y.lim) - diff(x.lim))/2)
		plot(x.lim, y.lim, xlab = "Coord X", ylab = "Coord Y", type = 
			"n")
		lines(polygon)
	}
	else plot(coords, xlab = "Coord X", ylab = "Coord Y", type = "n")
	points(coords[(data <= data.quantile[2]),  ], pch = 1, cex = 
		0.59999999999999998, col = 2)
	points(coords[((data > data.quantile[2]) & (data <= data.quantile[
		3])),  ], pch = 2, cex = 1.3999999999999999, col = 4)
	points(coords[((data > data.quantile[3]) & (data <= data.quantile[
		4])),  ], pch = 3, cex = 1.7, col = 7)
	points(coords[(data > data.quantile[4]),  ], pch = 4, cex = 2, col = 8)
	#  par(par.ori)
	return(invisible())
}

"plot.variogram" <- 
function(obj, max.dist = max(obj$u), ylim = "default", type = "b", scaled = F,
	var.lines = F, envelope.obj = NULL, bin.cloud = F, ...)
{
	if(obj$output.type == "bin")
		type <- "b"
	if(obj$output.type == "smooth")
		type <- "l"
	if(obj$output.type == "cloud")
		type <- "p"
	if(bin.cloud == T && type != "b")
		stop("object must be a binned variogram with option bin.cloud=T"
			)
	if(bin.cloud == T && all(is.na(obj$bin.cloud)))
		stop("object must be a binned variogram with option bin.cloud=T"
			)
	if(bin.cloud == T && any(!is.na(obj$bin.cloud)))
		boxplot(obj$bin.cloud, varwidth = T, names = as.character(
			obj$u), xlab = "midpoints of distance class", ylab = 
			paste("variogram values / ", obj$estimator.type, 
			"estimator"), ...)
	else {
		u <- obj$u[obj$u <= max.dist]
		if(!is.matrix(obj$v)) {
			v <- obj$v[obj$u <= max.dist]
			ymax <- max(obj$v[obj$u <= max.dist])
		}
		else {
			v <- obj$v[obj$u <= max.dist, 1]
			ymax <- max(obj$v[obj$u <= max.dist,  ])
		}
		if(!is.null(envelope.obj))
			ymax <- max(envelope.obj$v.upper, ymax)
		if(scaled == T)
			v <- v/obj$var.mark
		if(is.numeric(ylim))
			plot(u, v, xlim = c(0, max.dist), ylim = ylim, xlab = 
				"Distance", ylab = "Semi-variance", type = type,
				...)
		else plot(u, v, xlim = c(0, max.dist), ylim = c(0, ymax), xlab
				 = "Distance", ylab = "Semi-variance", type = 
				type, ...)
		if(var.lines == T) {
			if(scaled == T)
				abline(h = 1, lty = 3)
			else abline(h = obj$var.mark, lty = 3)
		}
		if(is.matrix(obj$v)) {
			for(k in 2:ncol(obj$v)) {
				v <- obj$v[obj$u <= max.dist, k]
				lines(u, v, type = "b")
			}
		}
		if(!is.null(envelope.obj)) {
			lines(u, envelope.obj$v.lower, lty = 4)
			lines(u, envelope.obj$v.upper, lty = 4)
		}
	}
	return(invisible())
}

"points.dots" <- 
function(obj, data = NULL, col.data = 1, ...)
{
	if(class(obj) == "geodata" | class(obj) == "grfsim") {
		coords <- as.matrix(obj$coords)
		data <- as.matrix(data)
		data <- data[, col.data]
	}
	else {
		coords <- obj
		if(is.matrix(data))
			data <- data[, col.data]
	}
	data.quantile <- quantile(data)
	points(coords[(data <= data.quantile[2]),  ], cex = 0.40000000000000002,
		col = 2)
	points(coords[((data > data.quantile[2]) & (data <= data.quantile[
		3])),  ], cex = 0.65000000000000002, col = 4)
	points(coords[((data > data.quantile[3]) & (data <= data.quantile[
		4])),  ], cex = 0.90000000000000002, col = 7)
	points(coords[(data > data.quantile[4]),  ], cex = 1.2, col = 8)
	#  par(par.ori)
	return(invisible())
}

"points.geodata" <- 
function(geodata, coords = geodata$coords, data = geodata$data, col.data = 1,
	pt.sizes = c("data.proportional", "rank.proportional", "quintiles",
	"quartiles", "deciles", "equal"), cex.min, cex.max, pch.seq, col.seq,
	add.to.plot = F, round.quantiles = F, graph.pars = T, xlab = "Coords X",
	ylab = "Coord Y", ...)
{
	par.now <- par()
	on.exit(par(par.now))
	pt.sizes <- match.arg(pt.sizes)
	if(add.to.plot == F) {
		plot(apply(coords, 2, range), type = "n", xlab = xlab, ylab = 
			ylab)
	}
	if(missing(cex.min))
		cex.min <- 0.5
	if(missing(cex.max))
		cex.max <- 2
	if(is.matrix(data))
		data <- as.vector(data[, col.data])
	graph.list <- list()
	if(pt.sizes == "quintiles" | pt.sizes == "quartiles" | pt.sizes == 
		"deciles") {
		if(pt.sizes == "quintiles") {
			n.quant <- 5
			if(missing(col.seq))
				col.seq <- c(6, 4, 7, 5, 8)
		}
		if(pt.sizes == "quartiles") {
			n.quant <- 4
			if(missing(col.seq))
				col.seq <- c(6, 4, 7, 8)
		}
		if(pt.sizes == "deciles") {
			n.quant <- 10
			if(missing(col.seq))
				col.seq <- c(6, 4, 2, 9, 11, 14, 12, 7, 5,
					8)
		}
		if(missing(pch.seq))
			pch.seq <- rep(16, n.quant)
		cex.pt <- seq(cex.min, cex.max, l = n.quant)
		data.quantile <- quantile(data, probs = seq(0, 1, by = (1/
			n.quant)))
		if(round.quantiles == T) {
			data.quantile[1] <- floor(data.quantile[1])
			data.quantile[n.quant + 1] <- ceiling(data.quantile[
				n.quant + 1])
			data.quantile <- round(data.quantile)
		}
		graph.list$quantiles <- data.quantile
		graph.list$cex <- cex.pt
		graph.list$col <- col.seq
		graph.list$pch <- pch.seq
		data.quantile[1] <- 0.90000000000000002 * data.quantile[1]
		for(i in 1:n.quant) {
			points(coords[((data > data.quantile[i]) & (data <= 
				data.quantile[i + 1])),  ], pch = pch.seq[
				i], cex = cex.pt[i], col = col.seq[i], ...)
			points(coords[((data > data.quantile[i]) & (data <= 
				data.quantile[i + 1])),  ], pch = 1, cex = 1.2 *
				cex.pt[i], col = 1, ...)
		}
	}
	if(pt.sizes == "rank.proportional") {
		if(missing(pch.seq))
			pch.seq <- 16
		data.quantile <- range(data)
		n <- length(data)
		coords.order <- coords[order(data),  ]
		data.order <- data[order(data)]
		pt.size <- seq(cex.min, cex.max, l = n)
		graph.list$cex <- c(cex.min, cex.max)
		graph.list$pch <- unique(range(pch.seq))
		for(i in 1:n)
			points(coords.order[i,  , drop = F], cex = pt.size[
				i], pch = pch.seq, ...)
	}
	if(pt.sizes == "data.proportional") {
		if(missing(pch.seq))
			pch.seq <- 16
		r.y <- range(data)
		n <- length(data)
		size <- cex.min + ((data - r.y[1]) * (cex.max - cex.min))/
			(r.y[2] - r.y[1])
		graph.list$cex <- c(cex.min, cex.max)
		graph.list$pch <- unique(range(pch.seq))
		for(i in 1:n)
			points(coords[i,  , drop = F], cex = size[i], pch = 
				pch.seq, ...)
	}
	if(pt.sizes == "equal") {
		if(missing(pch.seq))
			pch.seq <- 16
		points(coords, pch = pch.seq, ...)
	}
	if(graph.pars == T)
		return(graph.list)
	else return(invisible())
}

"points.symbols" <- 
function(geodata, coords = geodata$coords, data = geodata$data, col.data = 1,
	pt.sizes = c("proportional", "quartiles", "deciles", "equal"), cex.min,
	cex.max, pch.seq, col.seq, ...)
{
	pt.sizes <- match.arg(pt.sizes)
	if(missing(cex.min))
		cex.min <- 0.5
	if(missing(cex.max))
		cex.max <- 2
	if(is.matrix(data))
		data <- as.vector(data[, col.data])
	if(pt.sizes == "quartiles") {
		if(missing(pch.seq))
			pch.seq <- rep(16, 4)
		if(missing(col.seq))
			col.seq <- c(2, 4, 7, 8)
		cex.pt <- seq(cex.min, cex.max, l = 4)
		data.quantile <- quantile(data)
		for(i in 1:4) {
			points(coords[((data > data.quantile[i]) & (data <= 
				data.quantile[i + 1])),  ], pch = pch.seq[
				i], cex = cex.pt[i], col = col.seq[i], ...)
		}
	}
	if(pt.sizes == "deciles") {
		if(missing(pch.seq))
			pch.seq <- rep(16, 10)
		#    if(missing(col.seq))
		#      col.seq <- c(2,4,7,8)
		cex.pt <- seq(cex.min, cex.max, l = 10)
		data.quantile <- quantile(data, probs = seq(0, 1, by = 
			0.10000000000000001))
		for(i in 1:10) {
			points(coords[((data > data.quantile[i]) & (data <= 
				data.quantile[i + 1])),  ], pch = pch.seq[
				i], cex = cex.pt[i], ...)
		}
	}
	if(pt.sizes == "proportional") {
		if(missing(pch.seq))
			pch.seq <- 16
		data.quantile <- range(data)
		n <- length(data)
		coords.order <- coords[order(data),  ]
		data.order <- data[order(data)]
		pt.size <- seq(cex.min, cex.max, l = n)
		for(i in 1:n)
			points(coords.order[i,  , drop = F], cex = pt.size[
				i], pch = pch.seq, ...)
	}
	if(pt.sizes == "equal")
		if(missing(pch.seq))
			pch.seq <- 16
	points(coords, pch = pch.seq, ...)
	return(invisible())
}

"polygrid" <- 
function(xgrid, ygrid, poly, vec.inout = F)
{
	xygrid <- expand.grid(x = xgrid, y = ygrid)
	ind <- as.vector(inout(xygrid, poly))
	xypoly <- xygrid[ind == T,  ]
	if(vec.inout == F)
		return(xypoly)
	else return(list(xypoly = xypoly, vec.inout = ind))
}

"prior.control" <- 
function(beta.prior = c("flat", "normal", "fixed"), beta = NULL, beta.var = 
	NULL, sill.prior = c("reciprocal", "fixed"), sill = NULL, range.prior
	 = c("uniform", "exponential", "fixed", "squared.reciprocal", 
	"reciprocal"), exponential.prior.par = 1, range = NULL, range.discrete
	 = NULL, nugget.prior = c("fixed", "uniform"), nugget = 0, 
	nugget.discrete = NULL)
{
	beta.prior <- match.arg(beta.prior)
	sill.prior <- match.arg(sill.prior)
	range.prior <- match.arg(range.prior)
	nugget.prior <- match.arg(nugget.prior)
	return(list(beta.prior = beta.prior, beta = beta, beta.var = beta.var,
		sill.prior = sill.prior, sill = sill, range.prior = range.prior,
		exponential.prior.par = exponential.prior.par, range = range,
		range.discrete = range.discrete, nugget.prior = nugget.prior,
		nugget = nugget, nugget.discrete = nugget.discrete))
}

"proflik" <- 
function(obj.likfit, geodata, coords = geodata$coords, data = geodata$data,
	nugget.values, nugget.rel.values, sill.values, range.values, 
	lambda.values, sillrange.values = T, sillnugget.values = T, 
	rangenugget.values = T, sillnugget.rel.values = F, 
	rangenugget.rel.values = F, silllambda.values = F, rangelambda.values
	 = T, nuggetlambda.values = F, nugget.rellambda.values = F, uni.only = 
	F, bi.only = F, control = nlminb.control(iter.max = 1000))
{
	#
	# 1. setting arguments
	#
	call <- match.call()
	n.cov.pars <- obj.likfit$npars - length(obj.likfit$beta)
	if(obj.likfit$info.lambda$fix.lambda == F)
		n.cov.pars <- n.cov.pars - 1
	if(missing(sill.values))
		sill.values <- F
	if(missing(range.values))
		range.values <- F
	if(!is.null(obj.likfit$call$fix.nugget))
		if(obj.likfit$call$fix.nugget == T)
			nugget.values <- nugget.rel.values <- F
	if(missing(nugget.values))
		nugget.values <- F
	if(missing(nugget.rel.values))
		nugget.rel.values <- F
	if(missing(lambda.values) | obj.likfit$info.lambda$fix.lambda == T)
		lambda.values <- F
	if(uni.only == T) {
		sillrange.values <- sillnugget.values <- rangenugget.values <-
			sillnugget.rel.values <- rangenugget.rel.values <- 
			silllambda.values <- rangelambda.values <- 
			nugget.rellambda.values <- nuggetlambda.values <- F
	}
	else {
		if(all(sillrange.values == T)) {
			if(all(sill.values == F) | all(range.values == F)) {
				sillrange.values <- F
				stop("if argument sillrange.values = T sill.values and range.values must be provided. Alternatively a matrix can be provided in sillrange.values"
					)
			}
			else sillrange.values <- as.matrix(expand.grid(
					sill.values, range.values))
		}
		if(n.cov.pars == 2) {
			sillnugget.values <- rangenugget.values <- 
				sillnugget.rel.values <- rangenugget.rel.values <-
				nugget.rellambda.values <- nuggetlambda.values <-
				F
		}
		else {
			if(all(sillnugget.values == T)) {
				if(all(sill.values == F) | all(nugget.values ==
					F)) {
					sillnugget.values <- F
					stop("if argument sillnugget.values = T sill.values and nugget.values must be provided. Alternatively a matrix can be provided in sillnugget.values"
						)
				}
				else sillnugget.values <- as.matrix(expand.grid(
						sill.values, nugget.values))
			}
			if(all(rangenugget.values == T)) {
				if(all(range.values == F) | all(nugget.values ==
					F)) {
					rangenugget.values <- F
					stop("if argument rangenugget.values = T range.values and nugget.values must be provided. Alternatively a matrix can be provided in rangenugget.values"
						)
				}
				else rangenugget.values <- as.matrix(
						expand.grid(range.values, 
						nugget.values))
			}
			if(all(sillnugget.rel.values == T)) {
				if(all(sill.values == F) | all(
					nugget.rel.values == F)) {
					sillnugget.rel.values <- F
					stop("if argument sillnugget.rel.values = T sill.values and nugget.rel.values must be provided. Alternatively a matrix can be provided in sillnugget.rel.values"
						)
				}
				else sillnugget.rel.values <- as.matrix(
						expand.grid(sill.values, 
						nugget.rel.values))
			}
			if(all(rangenugget.rel.values == T)) {
				if(all(range.values == F) | all(
					nugget.rel.values == F)) {
					rangenugget.rel.values <- F
					stop("if argument rangenugget.rel.values = T range.values and nugget.rel.values must be provided. Alternatively a matrix can be provided in rangenugget.rel.values"
						)
				}
				else rangenugget.rel.values <- as.matrix(
						expand.grid(range.values, 
						nugget.rel.values))
			}
			if(obj.likfit$info.lambda$fix.lambda == T) {
				if(all(nuggetlambda.values == T)) {
					if(all(lambda.values == F) | all(
						nugget.values == F)) {
						nuggetlambda.values <- F
						stop("if argument nuggetlambda.values = T lambda.values and nugget.values must be provided. Alternatively a matrix can be provided in nuggetlambda.values"
							)
					}
					else nuggetlambda.values <- as.matrix(
							expand.grid(
							lambda.values, 
							nugget.values))
				}
				if(all(nugget.rellambda.values == T)) {
					if(all(lambda.values == F) | all(
						nugget.rel.values == F)) {
						nugget.rellambda.values <- F
						stop("if argument nugget.rellambda.values = T lambda.values and nugget.rel.values must be provided. Alternatively a matrix can be provided in nugget.rellambda.values"
							)
					}
					else nugget.rellambda.values <- 
							as.matrix(expand.grid(
							lambda.values, 
							nugget.rel.values))
				}
			}
		}
		if(obj.likfit$info.lambda$fix.lambda == T)
			silllambda.values <- rangelambda.values <- F
		else {
			if(all(silllambda.values == T)) {
				if(all(sill.values == F) | all(lambda.values ==
					F)) {
					silllambda.values <- F
					stop("if argument silllambda.values = T sill.values and lambda.values must be provided. Alternatively a matrix can be provided in silllambda.values"
						)
				}
				else silllambda.values <- as.matrix(expand.grid(
						sill.values, lambda.values))
			}
			if(all(rangelambda.values == T)) {
				if(all(range.values == F) | all(lambda.values ==
					F)) {
					rangelambda.values <- F
					stop("if argument rangelambda.values = T range.values and lambda.values must be provided. Alternatively a matrix can be provided in rangelambda.values"
						)
				}
				else rangelambda.values <- as.matrix(
						expand.grid(range.values, 
						lambda.values))
			}
		}
	}
	#
	# 2. data preparation
	#
	trend <- as.matrix(obj.likfit$trend.matrix)
	data <- as.vector(data)
	dimnames(trend) <- list(NULL, NULL)
	if(obj.likfit$info.lambda$fix.lambda == T) {
		if(obj.likfit$lambda != 1) {
			if(any(data <= 0))
				stop("Data transformation not allowed when there are zeros or negative data"
					)
			if(obj.likfit$lambda == 0)
				data <- log(data)
			else data <- ((data^obj.likfit$lambda) - 1)/obj.likfit$
					lambda
		}
	}
	n <- length(data)
	dists.vec <- dist(coords)
	d <- range(dists.vec)
	min.dist <- d[1]
	max.dist <- d[2]
	tausq <- obj.likfit$nugget
	sigmasq <- obj.likfit$cov.pars[1]
	tausq.rel <- tausq/sigmasq
	phi <- obj.likfit$cov.pars[2]
	lambda <- obj.likfit$lambda
	loglik <- obj.likfit$loglik
	sill.tot <- sigmasq + tausq
	n.uni <- 0
	n.bi <- 0
	lower.phi <- 0
	upper.phi <- 1000 * max.dist
	lower.sigmasq <- 0.01 * sill.tot
	result <- list()
	temp.list <- list(n = n, z = data, beta.size = dim(trend)[2], kappa = 
		obj.likfit$kappa, xmat = trend, txmat = t(trend), method = 
		obj.likfit$method, dists.lowertri = dists.vec, cov.model = 
		obj.likfit$cov.model, fix.lambda = obj.likfit$info.lambda$
		fix.lambda, lambda = obj.likfit$lambda, lower.phi = lower.phi,
		upper.phi = upper.phi, lower.sigmasq = lower.sigmasq, phi.est
		 = phi, tausq.rel.est = tausq.rel, tausq.est = tausq, 
		sigmasq.est = sigmasq, control = control)
	if(obj.likfit$info.lambda$fix.lambda == T) {
		temp.list$log.jacobian <- obj.likfit$info.lambda$log.jacobian
	}
	#
	# 3. One-dimentional profile likelihoods
	#
	#  
	# 3.1 Profile for sigmasq
	#  
	if(bi.only == F) {
		if(any(sill.values != F)) {
			n.uni <- n.uni + 1
			cat("proflik: computing profile likelihood for the sill\n"
				)
			if(n.cov.pars == 2) {
				if(tausq == 0) {
					if(any(lambda.values != F)) {
						temp.list$ini.grid <- as.matrix(
							expand.grid(seq(min(
							range.values), max(
							range.values), l = 10),
							seq(min(lambda.values),
							max(lambda.values),
							l = 5)))
					}
					else {
						temp.list$ini.grid <- as.matrix(
							seq(min(range.values),
							max(range.values),
							l = 10))
					}
					dimnames(temp.list$ini.grid) <- list(
						NULL, NULL)
					pl.sigmasq <- apply(matrix(sill.values,
						ncol = 1), 1, proflik.aux2,
						temp.list)
					temp.list$ini.grid <- NULL
				}
				else {
					stop("not yet implemented for fixed nugget != 0"
						)
				}
			}
			if(n.cov.pars == 3) {
				if(any(lambda.values != F)) {
					ini.grid <- as.matrix(expand.grid(
						seq(min(range.values), max(
						range.values), l = 6), seq(
						0, 2 * tausq.rel, l = 4), seq(
						-1, 1, l = 5)))
				}
				else {
					ini.grid <- as.matrix(expand.grid(
						seq(min(range.values), max(
						range.values), l = 10), seq(
						0, 2 * tausq.rel, l = 4)))
				}
				dimnames(ini.grid) <- list(NULL, NULL)
				temp.list$ini.grid <- ini.grid
				pl.sigmasq <- apply(matrix(sill.values, ncol = 
					1), 1, proflik.aux9, temp.list)
				temp.list$ini.grid <- NULL
			}
			v.ord <- order(c(sigmasq, sill.values))
			result$sill <- list(sill = c(sigmasq, sill.values)[
				v.ord], proflik.sill = c(loglik, pl.sigmasq)[
				v.ord], est.sill = c(sigmasq, loglik))
		}
		#  
		# 3.2 Profile for phi
		#
		if(any(range.values != F)) {
			n.uni <- n.uni + 1
			cat("proflik: computing profile likelihood for the range\n"
				)
			if(n.cov.pars == 2) {
				if(tausq == 0) {
					temp.list$nugget <- 0
					pl.phi <- apply(matrix(range.values,
						ncol = 1), 1, proflik.aux0,
						temp.list)
					temp.list$nugget <- NULL
				}
				else {
					stop("not yet implemented for fixed nugget != 0"
						)
				}
			}
			if(n.cov.pars == 3) {
				pl.phi <- apply(matrix(range.values, ncol = 1),
					1, proflik.aux7, temp.list)
			}
			v.ord <- order(c(phi, range.values))
			result$range <- list(range = c(phi, range.values)[
				v.ord], proflik.range = c(loglik, pl.phi)[
				v.ord], est.range = c(phi, loglik))
		}
		#  
		# 3.3 Profile for \tau^2
		#  
		if(n.cov.pars == 3) {
			if(any(nugget.values != F)) {
				n.uni <- n.uni + 1
				cat("proflik: computing profile likelihood for the nugget\n"
					)
				pl.tausq <- apply(matrix(nugget.values, ncol = 
					1), 1, proflik.aux11, temp.list)
				v.ord <- order(c(tausq, nugget.values))
				result$nugget <- list(nugget = c(tausq, 
					nugget.values)[v.ord], proflik.nugget
					 = c(loglik, pl.tausq)[v.ord], 
					est.nugget = c(tausq, loglik))
			}
			#  
			# 3.4 Profile for relative \tau^2
			#
			if(any(nugget.rel.values != F)) {
				cat("proflik: computing profile likelihood for the relative nugget\n"
					)
				n.uni <- n.uni + 1
				pl.tausq.rel <- apply(matrix(nugget.rel.values,
					ncol = 1), 1, proflik.aux5, temp.list)
				v.ord <- order(c(tausq.rel, nugget.rel.values))
				result$nugget.rel <- list(nugget.rel = c(
					tausq.rel, nugget.rel.values)[v.ord],
					proflik.nugget = c(loglik, pl.tausq.rel
					)[v.ord], est.nugget.rel = c(tausq.rel,
					loglik))
			}
		}
		#  
		# 3.5 Profile for \lambda
		#
		if(any(lambda.values != F)) {
			temp.list$coords <- coords
			if(dim(temp.list$xmat)[2] == 1 & all(temp.list$xmat ==
				1))
				temp.list$trend.mat <- "cte"
			else {
				.temp.xmat <<- (temp.list$xmat[, -1])
				temp.list$trend.mat <-  ~ .temp.xmat
			}
			n.uni <- n.uni + 1
			cat("proflik: computing profile likelihood for the transformation parameter\n"
				)
			if(n.cov.pars == 2) {
				if(tausq == 0) {
					temp.list$fixtau <- T
					temp.list$ini <- c(sigmasq, phi)
					pl.lambda <- apply(as.matrix(
						lambda.values), 1, 
						proflik.aux23, temp.list)
					temp.list$fixtau <- temp.list$ini <-
						NULL
				}
				else {
					stop("not yet implemented for fixed nugget != 0"
						)
				}
			}
			if(n.cov.pars == 3) {
				temp.list$fixtau <- F
				temp.list$ini <- phi
				pl.lambda <- apply(matrix(lambda.values, ncol
					 = 1), 1, proflik.aux23, temp.list)
				temp.list$fixtau <- temp.list$ini <- NULL
			}
			v.ord <- order(c(lambda, lambda.values))
			result$lambda <- list(lambda = c(lambda, lambda.values)[
				v.ord], proflik.lambda = c(loglik, pl.lambda)[
				v.ord], est.lambda = c(lambda, loglik))
			if(exists(".temp.xmat", w = 1))
				remove(".temp.xmat", w = 1)
		}
	}
	#
	# 4. Two-dimentional profile likelihoods
	#
	#  
	# 4.1 Profile for \sigma^2 and \phi
	#
	#
	#
	if(uni.only == F) {
		if(any(sillrange.values != F)) {
			n.bi <- n.bi + 1
			cat("proflik: computing 2-D profile likelihood for the sill and range parameters\n"
				)
			if(n.cov.pars == 2) {
				if(tausq == 0) {
					temp.list$nugget <- 0
					if(temp.list$fix.lambda == T) {
						pl.sigmasqphi <- apply(cbind(
							0, sillrange.values,
							1), 1, loglik.spatial,
							temp.list = temp.list)
					}
					else {
						pl.sigmasqphi <- apply(
							sillrange.values, 1,
							proflik.aux28, 
							temp.list)
					}
					temp.list$nugget <- NULL
				}
				else {
					stop("not yet implemented for fixed nugget != 0"
						)
				}
			}
			if(n.cov.pars == 3) {
				pl.sigmasqphi <- apply(sillrange.values, 1,
					proflik.aux13, temp.list)
			}
			names(pl.sigmasqphi) <- NULL
			result$sillrange <- list(sill = as.numeric(levels(
				as.factor(sillrange.values[, 1]))), range = 
				as.numeric(levels(as.factor(sillrange.values[
				, 2]))), proflik.sillrange = pl.sigmasqphi,
				est.sillrange = c(sigmasq, phi, loglik))
		}
		#  
		# 4.2 Profile for \sigma^2 and \tau^2
		#  
		if(any(sillnugget.values != F)) {
			n.bi <- n.bi + 1
			cat("proflik: computing 2-D profile likelihood for the sill and nugget\n"
				)
			if(obj.likfit$info.lambda$fix.lambda == F)
				ini.grid <- as.matrix(expand.grid(seq(min(
					range.values), max(range.values), l = 
					10), seq(-1, 1, l = 5)))
			else ini.grid <- as.matrix(seq(min(range.values), max(
					range.values), l = 10))
			dimnames(ini.grid) <- list(NULL, NULL)
			temp.list$ini.grid <- ini.grid
			pl.sigmasqtausq <- apply(sillnugget.values, 1, 
				proflik.aux15, temp.list)
			temp.list$ini.grid <- NULL
			names(pl.sigmasqtausq) <- NULL
			result$sillnugget <- list(sill = as.numeric(levels(
				as.factor(sillnugget.values[, 1]))), nugget = 
				as.numeric(levels(as.factor(sillnugget.values[
				, 2]))), proflik.sillnugget = pl.sigmasqtausq,
				est.sillrange = c(sigmasq, tausq, loglik))
		}
		#  
		# 4.3 Profile for \phi and \tau^2
		#
		if(any(rangenugget.values != F)) {
			n.bi <- n.bi + 1
			cat("proflik: computing 2-D profile likelihood for the range and nugget\n"
				)
			temp.list$ini.grid <- as.matrix(seq(sigmasq/4, 5 * 
				sigmasq, l = 15))
			pl.phitausq <- apply(rangenugget.values, 1, 
				proflik.aux17, temp.list)
			temp.list$ini.grid <- NULL
			names(pl.phitausq) <- NULL
			result$rangenugget <- list(range = as.numeric(levels(
				as.factor(rangenugget.values[, 1]))), nugget = 
				as.numeric(levels(as.factor(rangenugget.values[
				, 1]))), proflik.rangenugget = pl.phitausq,
				est.rangenugget = c(phi, tausq, loglik))
		}
		#  
		# 4.4 Profile for \sigma^2 and \tau^2_{rel}
		#
		if(any(sillnugget.rel.values != F)) {
			n.bi <- n.bi + 1
			cat("proflik: computing 2-D profile likelihood for the sill and relative nugget parameters\n"
				)
			if(temp.list$fix.lambda == F)
				ini.grid <- as.matrix(expand.grid(seq(min(
					range.values), max(range.values), l = 
					10), seq(-1, 1, l = 5)))
			else ini.grid <- as.matrix(seq(min(range.values), max(
					range.values), l = 10))
			dimnames(ini.grid) <- list(NULL, NULL)
			temp.list$ini.grid <- ini.grid
			pl.sigmasqtausq.rel <- apply(sillnugget.rel.values,
				1, proflik.aux19, temp.list)
			names(pl.sigmasqtausq.rel) <- NULL
			temp.list$ini.grid <- NULL
			result$sillnugget.rel <- list(sill = as.numeric(levels(
				as.factor(sillnugget.rel.values[, 1]))), 
				nugget.rel = as.numeric(levels(as.factor(
				sillnugget.rel.values[, 2]))), 
				proflik.sillnugget.rel = pl.sigmasqtausq.rel,
				est.sillrange.rel = c(sigmasq, tausq.rel, 
				loglik))
		}
		#  
		# 4.5 Profile for \phi and \tau^2_{rel}
		#
		if(any(rangenugget.rel.values != F)) {
			n.bi <- n.bi + 1
			cat("proflik: computing 2-D profile likelihood for the range and relative nugget parameters\n"
				)
			pl.phitausq.rel <- apply(rangenugget.rel.values, 1,
				proflik.aux30, temp.list)
			names(pl.phitausq.rel) <- NULL
			result$rangenugget.rel <- list(range = as.numeric(
				levels(as.factor(rangenugget.rel.values[, 1]))),
				nugget.rel = as.numeric(levels(as.factor(
				rangenugget.rel.values[, 2]))), 
				proflik.rangenugget.rel = pl.phitausq.rel,
				"est.rangenugget .rel" = c(phi, tausq.rel,
				loglik))
		}
	}
	#  
	# 4.6 Profile for \sigma^2 and \lambda
	#
	if(any(silllambda.values != F)) {
		n.bi <- n.bi + 1
		cat("proflik: computing 2-D profile likelihood for the sill and transformation parameters\n"
			)
		if(n.cov.pars == 2) {
			ini.grid <- as.matrix(seq(min(range.values), max(
				range.values), l = 10))
			dimnames(ini.grid) <- list(NULL, NULL)
			temp.list$ini.grid <- ini.grid
			if(tausq == 0) {
				temp.list$nugget <- 0
				pl.sigmasqlambda <- apply(silllambda.values,
					1, proflik.aux24, temp.list)
				temp.list$ini.grid <- temp.list$nugget <- NULL
			}
			else {
				stop("not yet implemented for fixed nugget != 0"
					)
			}
		}
		if(n.cov.pars == 3) {
			ini.grid <- as.matrix(expand.grid(seq(min(range.values),
				max(range.values), l = 10), seq(0, 1, l = 5)))
			dimnames(ini.grid) <- list(NULL, NULL)
			temp.list$ini.grid <- ini.grid
			pl.sigmasqlambda <- apply(sigmasqlambda.values, 1,
				proflik.aux27, temp.list)
			temp.list$ini.grid <- NULL
		}
		names(pl.sigmasqlambda) <- NULL
		result$silllambda <- list(sill = as.numeric(levels(as.factor(
			silllambda.values[, 1]))), lambda = as.numeric(levels(
			as.factor(silllambda.values[, 2]))), proflik.silllambda
			 = pl.sigmasqlambda, est.silllambda = c(sigmasq, lambda,
			loglik))
	}
	#  
	# 4.7 Profile for \phi and \lambda
	#
	if(any(rangelambda.values != F)) {
		n.bi <- n.bi + 1
		cat("proflik: computing 2-D profile likelihood for the range and transformation parameters\n"
			)
		if(n.cov.pars == 2) {
			if(tausq == 0) {
				temp.list$nugget <- 0
				pl.philambda <- apply(rangelambda.values, 1,
					proflik.aux29, temp.list)
				temp.list$nugget <- NULL
			}
			else {
				stop("not yet implemented for fixed nugget != 0"
					)
			}
		}
		if(n.cov.pars == 3) {
			pl.philambda <- apply(rangelambda.values, 1, 
				proflik.aux31, temp.list)
		}
		names(pl.philambda) <- NULL
		result$rangelambda <- list(range = as.numeric(levels(as.factor(
			rangelambda.values[, 1]))), lambda = as.numeric(levels(
			as.factor(rangelambda.values[, 2]))), 
			proflik.rangelambda = pl.philambda, est.rangelambda = c(
			phi, lambda, loglik))
	}
	#  
	#                                        # 4.8 Profile for \tau^2 and \lambda
	#                                        #
	if(any(nuggetlambda.values != F)) {
		n.bi <- n.bi + 1
		cat("proflik: computing 2-D profile likelihood for the nugget and transformation parameters\n"
			)
		pl.nuggetlambda <- apply(nuggetlambda.values, 1, proflik.aux32,
			temp.list)
		names(pl.nuggetlambda) <- NULL
		result$nuggetlambda <- list(nugget = as.numeric(levels(
			as.factor(nuggetlambda.values[, 1]))), lambda = 
			as.numeric(levels(as.factor(nuggetlambda.values[, 2]))),
			proflik.nuggetlambda = pl.nuggetlambda, 
			est.nuggetlambda = c(tausq, lambda, loglik))
	}
	#  
	# 4.9 2-D Profile for \tau^2_{rel} and \lambda
	#
	if(any(nugget.rellambda.values != F)) {
		n.bi <- n.bi + 1
		pl.nugget.rellambda <- apply(nugget.rellambda.values, 1, 
			proflik.aux33, temp.list)
		names(pl.nugget.rellambda) <- NULL
		result$nugget.rellambda <- list(nugget.rel = as.numeric(levels(
			as.factor(nugget.rellambda.values[, 1]))), lambda = 
			as.numeric(levels(as.factor(nugget.rellambda.values[
			, 2]))), proflik.nugget.rellambda = pl.nugget.rellambda,
			est.nugget.rellambda = c(tausq.rel, lambda, loglik))
	}
	result$n.uni <- n.uni
	result$n.bi <- n.bi
	result$method <- obj.likfit$method
	result$call <- call
	class(result) <- "proflik"
	return(result)
}

"proflik.aux0" <- 
function(phi, temp.list)
{
	# This function computes the value of the profile likelihood for the correlation parameter \phi when nugget effect is not included in the model.
	# It requires the minimisation of the function wrt \phi for each value of \lambda, if this transformation parameter is included in the model
	# This is an auxiliary function called by likfit.proflik
	#
	if(temp.list$fix.lambda == T) proflik <- proflik.aux1(phi = phi, 
			temp.list = temp.list) else {
		temp.list$phi <- phi
		proflik <-  - (nlminb(temp.list$lambda, proflik.aux1.1, lower
			 = -3, upper = 3, temp.list = temp.list)$objective)
	}
	return(proflik)
}

"proflik.aux1" <- 
function(phi, temp.list)
{
	# This function computes the value of the profile likelihood for the correlation function scale parameter \phi when nugget effect = 0
	z <- temp.list$z
	n <- temp.list$n
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$nugget, cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		proflik <-  - (n/2) * log(2 * pi) - covinf$log.det.to.half -
			(n/2) * log(ssresmat/n) - (n/2) + temp.list$
			log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		proflik <-  - ((n - beta.size)/2) * log(2 * pi) - covinf$
			log.det.to.half - ((n - beta.size)/2) * log(ssresmat/
			n) - (n/2) + 0.5 * sum(log(eigentrem$values)) + 
			temp.list$log.jacobian
	}
	return(proflik)
}

"proflik.aux1.1" <- 
function(lambda, temp.list)
{
	# This function computes the value of the profile likelihood for the correlation function scale parameter \phi when nugget effect = 0
	z <- temp.list$z
	n <- temp.list$n
	phi <- temp.list$phi
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$nugget, cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
			n) + (n/2) - 0.5 * sum(log(eigentrem$values)) - 
			temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux10" <- 
function(phitausq.rel.lambda, temp.list)
{
	phitausq.rel.lambda <- as.vector(phitausq.rel.lambda)
	if(length(phitausq.rel.lambda) == 3)
		include.lambda <- TRUE
	else include.lambda <- FALSE
	z <- temp.list$z
	n <- length(z)
	phi <- phitausq.rel.lambda[1]
	tausq <- phitausq.rel.lambda[2]
	if(include.lambda) {
		lambda <- phitausq.rel.lambda[3]
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z <= 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	sigmasq <- temp.list$sigmasq
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = tausq,
		cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(sigmasq) + (0.5/sigmasq) * ssresmat - 
			temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(sigmasq) +
			(0.5/sigmasq) * ssresmat - 0.5 * sum(log(eigentrem$
			values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux10.1" <- 
function(phitausq.rel.lambda, temp.list)
{
	phitausq.rel.lambda <- as.vector(phitausq.rel.lambda)
	z <- temp.list$z
	n <- length(z)
	phi <- phitausq.rel.lambda[1]
	tausq <- phitausq.rel.lambda[2]
	lambda <- phitausq.rel.lambda[3]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	sigmasq <- temp.list$sigmasq
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = tausq,
		cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(sigmasq) + (0.5/sigmasq) * ssresmat - 
			temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(sigmasq) +
			(0.5/sigmasq) * ssresmat - 0.5 * sum(log(eigentrem$
			values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux11" <- 
function(tausq, temp.list)
{
	# This function computes the value of the profile likelihood for the parameter \tau^2.
	# It requires the minimisation of the function wrt \sigma^2, \phi and \lambda (if the case)  for each value of \tau^2.
	# This is an auxiliary function called by proflik.
	temp.list$nugget <- as.vector(tausq)
	if(temp.list$fix.lambda == T) {
		sigmasqphi.res <- nlminb(c(temp.list$sigmasq.est, temp.list$
			phi.est), proflik.aux12, lower = c(temp.list$
			lower.sigmasq, temp.list$lower.phi), temp.list = 
			temp.list)
	}
	else {
		sigmasqphi.res <- nlminb(c(temp.list$sigmasq.est, temp.list$
			phi.est, temp.list$lambda), proflik.aux12, lower = c(
			temp.list$lower.sigmasq, temp.list$lower.phi, -3),
			upper = c( + Inf,  + Inf, 3), temp.list = temp.list)
	}
	return( - sigmasqphi.res$objective)
}

"proflik.aux12" <- 
function(sigmasqphi.lambda, temp.list)
{
	## This function computes the value of the profile likelihood for the nugget parameter \tau^2, minimizing the likelihood wrt correlation function scale parameter \phi (range), the random field scale parameter \sigma^2 (sill) and the transformation parameter \lambda. 
	sigmasqphi.lambda <- as.vector(sigmasqphi.lambda)
	if(length(sigmasqphi.lambda) == 3)
		include.lambda <- TRUE
	else include.lambda <- FALSE
	z <- temp.list$z
	n <- temp.list$n
	sigmasq <- sigmasqphi.lambda[1]
	phi <- sigmasqphi.lambda[2]
	if(include.lambda) {
		lambda <- sigmasqphi.lambda[3]
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z <= 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$nugget, cov.pars = c(sigmasq, phi), inv = T, det = T,
		sca = F)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
			(ssresmat) - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + 0.5 * (ssresmat) - 0.5 * sum(log(
			eigentrem$values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux12.1" <- 
function(sigmasqphi.lambda, temp.list)
{
	# This function computes the value of the profile likelihood for the nugget parameter \tau^2, minimizing the likelihood wrt correlation function scale parameter \phi (range), the random field scale parameter \sigma^2 (sill) and the transformation parameter \lambda. 
	sigmasqphi.lambda <- as.vector(sigmasqphi.lambda)
	z <- temp.list$z
	n <- temp.list$n
	sigmasq <- sigmasqphi.lambda[1]
	phi <- sigmasqphi.lambda[2]
	lambda <- sigmasqphi.lambda[3]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$nugget, cov.pars = c(sigmasq, phi), inv = T, det = T,
		sca = F)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
			(ssresmat) - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + 0.5 * (ssresmat) - 0.5 * sum(log(
			eigentrem$values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux13" <- 
function(sigmasqphi, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the parameters \sigma^2 and \phi when the nugget is included.
	# It requires the minimisation of the function wrt \tau^2 and \lambda (if the case) for each value of (\sigma^2, \phi)
	# This is an auxiliary function called by likfit.proflik
	temp.list$sigmasqphi <- as.vector(sigmasqphi)
	if(temp.list$fix.lambda == T) {
		tausq.res <- nlminb(temp.list$tausq.est, proflik.aux14, lower
			 = 0, temp.list = temp.list)
	}
	else {
		tausq.res <- nlminb(c(temp.list$tausq.est, temp.list$lambda),
			proflik.aux14, lower = c(0, -3), upper = c( + Inf,
			3), temp.list = temp.list)
	}
	return( - tausq.res$objective)
}

"proflik.aux14" <- 
function(tausq.lambda, temp.list)
{
	## This function computes the value of the 2-D profile likelihood for the parameters (\sigma^2, \phi), minimizing the likelihood wrt the nugget parameter \tau^2.
	## This functions is called by the auxiliary function proflik.aux13
	if(length(tausq.lambda) == 2) include.lambda <- TRUE else 
			include.lambda <- FALSE
	z <- temp.list$z
	n <- temp.list$n
	tausq <- tausq.lambda[1]
	if(include.lambda) {
		lambda <- tausq.lambda[2]
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z <= 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = tausq,
		cov.pars = temp.list$sigmasqphi, inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
			ssresmat - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + 0.5 * ssresmat - 0.5 * sum(log(
			eigentrem$values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux14.1" <- 
function(tausq.lambda, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the parameters (\sigma^2, \phi), minimizing the likelihood wrt the nugget parameter \tau^2.
	# This functions is called by the auxiliary function proflik.aux13
	z <- temp.list$z
	n <- temp.list$n
	tausq <- tausq.lambda[1]
	lambda <- tausq.lambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = tausq,
		cov.pars = temp.list$sigmasqphi, inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
			ssresmat - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + 0.5 * ssresmat - 0.5 * sum(log(
			eigentrem$values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux15" <- 
function(sigmasqtausq, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the parameters \sigma^2 and \tau^2
	# It requires the minimisation of the function wrt \phi and also \lambda (if the case) for each value of (\sigma^2, \tau^2) 
	# This is an auxiliary function called by likfit.proflik
	temp.list$sigmasqtausq <- as.vector(sigmasqtausq)
	if(temp.list$fix.lambda == T) {
		phi.res <- nlminb(temp.list$phi.est, proflik.aux16, lower = 
			temp.list$lower.phi, temp.list = temp.list)
	}
	else {
		phi.res <- nlminb(c(temp.list$phi.est, temp.list$lambda), 
			proflik.aux16, lower = c(temp.list$lower.phi, -3),
			upper = c( + Inf, 3), temp.list = temp.list)
	}
	return( - phi.res$objective)
}

"proflik.aux16" <- 
function(phi.lambda, temp.list)
{
	## This function computes the value of the 2-D profile likelihood for the sill and nugget parameters (\sigma^2,\tau^2), minimising the profile likelihood wrt correlation function scale parameter \phi
	## This is an auxiliary function called by likfit.aux15  
	if(length(phi.lambda) == 2) include.lambda <- TRUE else include.lambda <-
			FALSE
	z <- temp.list$z
	n <- temp.list$n
	phi <- phi.lambda[1]
	if(include.lambda) {
		lambda <- phi.lambda[2]
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z <= 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$sigmasqtausq[2], cov.pars = c(temp.list$sigmasqtausq[
		1], phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
			ssresmat - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + 0.5 * ssresmat - 0.5 * sum(log(
			eigentrem$values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux16.1" <- 
function(phi.lambda, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the sill and nugget parameters (\sigma^2,\tau^2), minimising the profile likelihood wrt correlation function scale parameter \phi
	# This is an auxiliary function called by likfit.aux15  
	z <- temp.list$z
	n <- temp.list$n
	phi <- phi.lambda[1]
	lambda <- phi.lambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$sigmasqtausq[2], cov.pars = c(temp.list$sigmasqtausq[
		1], phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
			ssresmat - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + 0.5 * ssresmat - 0.5 * sum(log(
			eigentrem$values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux17" <- 
function(phitausq, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the parameters (\phi, \tau^2)
	# It requires the minimisation of the function wrt \sigma^2 and \lambda (if the case) for each value of (\phi, \tau^2) 
	# This is an auxiliary function called by likfit.proflik
	temp.list$phitausq <- as.vector(phitausq)
	if(temp.list$fix.lambda == T) {
		sigmasq.res <- nlminb(temp.list$sigmasq.est, proflik.aux18,
			lower = temp.list$lower.sigmasq, temp.list = temp.list)
	}
	else {
		sigmasq.res <- nlminb(c(temp.list$sigmasq.est, temp.list$lambda
			), proflik.aux18, lower = c(temp.list$lower.sigmasq,
			-3), upper = c( + Inf, 3), temp.list = temp.list)
	}
	return( - sigmasq.res$objective)
}

"proflik.aux18" <- 
function(sigmasq.lambda, temp.list)
{
	## This function computes the value of the 2-D profile likelihood for the range and nugget parameters (\phi, \tau^2), minimising the likelihood wrt the random field scale parameter \sigma^2 (sill) ant the transformation parameter \lambda. 
	## This is an auxiliary function called by likfit.aux17.
	if(length(sigmasq.lambda) == 2) include.lambda <- TRUE else 
			include.lambda <- FALSE
	z <- temp.list$z
	n <- temp.list$n
	sigmasq <- sigmasq.lambda[1]
	if(include.lambda) {
		lambda <- sigmasq.lambda[2]
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z <= 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$phitausq[2], cov.pars = c(sigmasq, temp.list$phitausq[
		1]), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
			ssresmat - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + 0.5 * ssresmat - 0.5 * sum(log(
			eigentrem$values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux18.1" <- 
function(sigmasq.lambda, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the range and nugget parameters (\phi, \tau^2), minimising the likelihood wrt the random field scale parameter \sigma^2 (sill) ant the transformation parameter \lambda. 
	# This is an auxiliary function called by likfit.aux17.
	z <- temp.list$z
	n <- temp.list$n
	sigmasq <- sigmasq.lambda[1]
	lambda <- sigmasq.lambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$phitausq[2], cov.pars = c(sigmasq, temp.list$phitausq[
		1]), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
			ssresmat - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + 0.5 * ssresmat - 0.5 * sum(log(
			eigentrem$values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux19" <- 
function(sigmasqtausq.rel, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the parameters (\sigma^2, \tau^2_{rel})
	# It requires the minimisation of the function wrt \phi and \lambda (if the case) for each value of (\sigma^2, \tau^2_{rel})
	# This is an auxiliary function called by likfit.proflik
	temp.list$sigmasqtausq.rel <- as.vector(sigmasqtausq.rel)
	if(temp.list$fix.lambda == T) {
		phi.res <- nlminb(temp.list$phi.est, proflik.aux20, lower = 
			temp.list$lower.phi, temp.list = temp.list)
	}
	else {
		phi.res <- nlminb(c(temp.list$phi.est, temp.list$lambda), 
			proflik.aux20, lower = c(temp.list$lower.phi, -3),
			upper = c( + Inf, 3), temp.list = temp.list)
	}
	return( - phi.res$objective)
}

"proflik.aux2" <- 
function(sigmasq, temp.list)
{
	# This function computes the value of the profile likelihood for the random field scale (variance) parameter \sigma^2 when nugget effect is not included in the model.
	# It requires the minimisation of the function wrt \phi for each value of \sigma^2
	# This is an auxiliary function called by likfit.proflik
	#
	temp.list$sigmasq <- as.vector(sigmasq)
	#    phi.res <- nlminb(temp.list$phi.est, proflik.aux3, lower = 
	#                      temp.list$lower.phi, temp.list = temp.list)
	if(temp.list$fix.lambda == T) {
		ini.lik <- round(100000000. * apply(temp.list$ini.grid, 1,
			proflik.aux3, temp.list))
		ini <- as.vector(temp.list$ini.grid[ini.lik == min(ini.lik),
			])
		phi.res <- nlminb(ini, proflik.aux3, lower = temp.list$
			lower.phi, temp.list = temp.list)
	}
	else {
		ini.lik <- round(100000000. * apply(temp.list$ini.grid, 1,
			proflik.aux3, temp.list))
		ini <- as.vector(temp.list$ini.grid[ini.lik == min(ini.lik),
			])
		phi.res <- nlminb(ini, proflik.aux3, lower = c(temp.list$
			lower.phi, -3), upper = c( + Inf, 3), temp.list = 
			temp.list)
	}
	return( - phi.res$objective)
}

"proflik.aux20" <- 
function(phi.lambda, temp.list)
{
	## This function computes the value of the 2-D profile likelihood for the sill and relative nugget parameters (\sigma^2, \tau^2_{rel}), minimising the likelihood wrt the correlation function scale parameter \phi and the transformation parameter \lambda.
	## This is an auxiliary function called by likfit.aux19.
	if(length(phi.lambda) == 2) include.lambda <- TRUE else include.lambda <-
			FALSE
	phi.lambda <- as.vector(phi.lambda)
	sigmasqtausq.rel <- as.vector(temp.list$sigmasqtausq.rel)
	sigmasq <- sigmasqtausq.rel[1]
	tausq.rel <- sigmasqtausq.rel[2]
	phi <- phi.lambda[1]
	lambda <- phi.lambda[2]
	z <- temp.list$z
	n <- temp.list$n
	if(include.lambda) {
		lambda <- phi.lambda[2]
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z <= 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		tausq.rel, cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(sigmasq) + (0.5/sigmasq) * ssresmat
		ssresmat - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + (n/2) * log(sigmasq) + (0.5/sigmasq) *
			ssresmat - 0.5 * sum(log(eigentrem$values)) - temp.list$
			log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux20.1" <- 
function(phi.lambda, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the sill and relative nugget parameters (\sigma^2, \tau^2_{rel}), minimising the likelihood wrt the correlation function scale parameter \phi and the transformation parameter \lambda.
	# This is an auxiliary function called by likfit.aux19.
	phi.lambda <- as.vector(phi.lambda)
	sigmasqtausq.rel <- as.vector(temp.list$sigmasqtausq.rel)
	sigmasq <- sigmasqtausq.rel[1]
	tausq.rel <- sigmasqtausq.rel[2]
	phi <- phi.lambda[1]
	lambda <- phi.lambda[2]
	z <- temp.list$z
	n <- temp.list$n
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		tausq.rel, cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(sigmasq) + (0.5/sigmasq) * ssresmat
		ssresmat - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + (n/2) * log(sigmasq) + (0.5/sigmasq) *
			ssresmat - 0.5 * sum(log(eigentrem$values)) - temp.list$
			log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux21" <- 
function(phitausq.rel, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the parameters (\phi, \tau^2_{rel})
	# This is an auxiliary function called by likfit.proflik
	phitausq.rel <- as.vector(phitausq.rel)
	phi <- phitausq.rel[1]
	tausq.rel <- phitausq.rel[2]
	z <- temp.list$z
	n <- temp.list$n
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	txmat <- temp.list$txmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		tausq.rel, cov.pars = c(1, phi), inv = T, det = T)
	ixix <- solve(txmat %*% covinf$inverse %*% xmat)
	ssresmat <- z %*% (covinf$inverse - covinf$inverse %*% xmat %*% ixix %*%
		txmat %*% covinf$inverse) %*% z
	sigmasq.hat <- ssresmat/n
	if(temp.list$method == "ML") {
		proflik <-  - (n/2) * log(2 * pi) - covinf$log.det.to.half -
			(n/2) * log(sigmasq.hat) - (n/2) - temp.list$
			log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		proflik <-  - ((n - beta.size)/2) * log(2 * pi) - covinf$
			log.det.to.half - (n/2) * log(sigmasq.hat) - (n/2) +
			0.5 * sum(log(eigentrem$values)) - temp.list$
			log.jacobian
	}
	return(proflik)
}

"proflik.aux21.1" <- 
function(lambda, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the parameters (\phi, \tau^2_{rel})
	# This requires minimasation wrt to the transformation parameter \lambda
	# This is an auxiliary function called by likfit.proflik
	phitausq.rel <- as.vector(temp.list$phitausq.rel)
	phi <- phitausq.rel[1]
	tausq.rel <- phitausq.rel[2]
	z <- temp.list$z
	n <- temp.list$n
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	txmat <- temp.list$txmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		tausq.rel, cov.pars = c(1, phi), inv = T, det = T)
	ixix <- solve(txmat %*% covinf$inverse %*% xmat)
	ssresmat <- z %*% (covinf$inverse - covinf$inverse %*% xmat %*% ixix %*%
		txmat %*% covinf$inverse) %*% z
	sigmasq.hat <- ssresmat/n
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(sigmasq.hat) + (n/2) + temp.list$
			log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + (n/2) * log(sigmasq.hat) + (n/2) -
			0.5 * sum(log(eigentrem$values)) + temp.list$
			log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux22" <- 
function(sigmasq, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the range and nugget parameters (\phi, \tau^2), minimising the likelihood wrt the random field scale parameter \sigma^2 (sill) 
	# This is an auxiliary function called by likfit.aux17
	z <- temp.list$z
	n <- temp.list$n
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$phitausq.rel[2], cov.pars = c(sigmasq, temp.list$
		phitausq[1]), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
			ssresmat - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + 0.5 * ssresmat - 0.5 * sum(log(
			eigentrem$values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux23" <- 
function(lambda, temp.list)
{
	# This function computes the value of the profile likelihood for the transformation parameter \lambda
	# It requires the minimisation of the function wrt \phi and \tau^2 and sigma^2 for each value of \lambda
	# This is an auxiliary function called by proflik
	lambda <- as.vector(lambda)
	if(temp.list$fixtau == F) {
		if(lambda == 0)
			data.l <- log(temp.list$z)
		else data.l <- ((temp.list$z^lambda) - 1)/lambda
		var.l <- var(data.l)
		ini.l <- c(0.10000000000000001 * var.l, var.l, temp.list$ini)
	}
	else ini.l <- temp.list$ini
	#  if(dim(temp.list$xmat)[2] == 1 & all(temp.list$xmat == 1))
	#    trend.mat <- "cte"
	#  else{
	#    .temp.xmat <<-  (temp.list$xmat[, -1])
	#    trend.mat <- ~.temp.xmat
	#  }
	lambda.res <- likfit(coords = temp.list$coords, data = temp.list$z,
		ini = ini.l, trend = temp.list$trend.mat, fix.nugget = 
		temp.list$fixtau, method = temp.list$method, cov.model = 
		temp.list$cov.model, kappa = temp.list$kappa, fix.lambda = T,
		lambda = lambda, messages.screen = F)$loglik
	return(lambda.res)
}

"proflik.aux24" <- 
function(sigmasqlambda, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the parameters (\sigma^2, \lambda) when there is no nugget effect (\tau^2 = 0, fixed)
	# It requires the minimisation of the function wrt \phi for each value of (\sigma^2, \lambda)
	# This is an auxiliary function called by proflik
	sigmasqlambda <- as.vector(sigmasqlambda)
	temp.list$sigmasq <- sigmasqlambda[1]
	lambda <- sigmasqlambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(temp.list$z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		temp.list$log.jacobian <- sum(log(temp.list$z^(lambda - 1)))
		if(lambda == 0)
			temp.list$z <- log(temp.list$z)
		else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
	}
	phi.res <- nlminb(temp.list$phi.est, proflik.aux3, lower = temp.list$
		lower.phi, temp.list = temp.list)
	return( - phi.res$objective)
}

"proflik.aux25" <- 
function(sigmasqlambda, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the parameters (\sigma^2, \lambda), when the nugget (\tau^2) is included in the model.
	# It requires the minimisation of the function wrt \phi and \tau^2 for each value of (\sigma^2, \lambda)
	# This is an auxiliary function called by proflik
	sigmasqlambda <- as.vector(sigmasqlambda)
	temp.list$sigmasq <- sigmasqlambda[1]
	lambda <- sigmasqlambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		stop("Transformation option not allowed when there are zeros or negative data"
			)
		temp.list$log.jacobian <- sum(log(temp.list$z^(lambda - 1)))
		if(lambda == 0)
			temp.list$z <- log(temp.list$z)
		else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
	}
	phitausq.rel.res <- nlminb(c(temp.list$phi.est, temp.list$tausq.rel.est
		), proflik.aux10, lower = c(temp.list$lower.phi, 0), temp.list
		 = temp.list)
	return( - phitausq.rel.res$objective)
}

"proflik.aux27" <- 
function(sigmasqlambda, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for sill \sigma^2 and the transformation parameter \lambda
	# It requires the minimisation of the function wrt \phi and \tau^2 and for each value of (\sigma^2,\lambda)
	# This is an auxiliary function called by proflik.
	sigmasqlambda <- as.vector(sigmasqlambda)
	temp.list$sigmasq <- sigmasqlambda[1]
	lambda <- sigmasqlambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		temp.list$fix.lambda <- T
		if(any(temp.list$z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(temp.list$z^(lambda -
				1)))
		else temp.list$log.jacobian <- sum(log(temp.list$z^(lambda -
				1)))
		if(lambda == 0)
			temp.list$z <- log(temp.list$z)
		else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
	}
	phitausq.rel.res <- nlminb(c(temp.list$phi.est, temp.list$tausq.rel.est
		), proflik.aux10, lower = c(temp.list$lower.phi, 0), temp.list
		 = temp.list)
	return( - phitausq.rel.res$objective)
}

"proflik.aux28" <- 
function(sigmasqphi, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for the random field scale (variance) parameter \sigma^2  and the correlation function parameter \phi when nugget effect is not included in the model.
	# It requires the minimisation of the function wrt \lambda for each value of (\sigma^2, \phi)
	# This is an auxiliary function called by likfit.proflik
	#
	ini.seq <- seq(-2, 2, l = 10)
	temp.list$sigmasqphi <- sigmasqphi
	lambda.lik <- apply(as.matrix(ini.seq), 1, proflik.aux4, temp.list = 
		temp.list)
	ini <- ini.seq[lambda.lik == max(lambda.lik)]
	lambda.res <- nlminb(ini, proflik.aux4, lower = -3, upper = 3, 
		temp.list = temp.list)
	temp.list$sigmasqphi <- NULL
	return( - lambda.res$objective)
}

"proflik.aux29" <- 
function(philambda, temp.list)
{
	# This function computes the value of the profile likelihood for the parameters (\phi, \lambda) when nugget effect is not included in the model.
	# This is an auxiliary function called by proflik.
	#
	philambda <- as.vector(philambda)
	phi <- philambda[1]
	lambda <- philambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(temp.list$z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(temp.list$z^(lambda -
				1)))
		else temp.list$log.jacobian <- sum(log(temp.list$z^(lambda -
				1)))
		if(lambda == 0)
			temp.list$z <- log(temp.list$z)
		else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
	}
	proflik <- proflik.aux1(phi = phi, temp.list = temp.list)
	temp.list$z <- temp.list$data
	return(proflik)
}

"proflik.aux3" <- 
function(phi.lambda, temp.list)
{
	## This function computer the negative of the likelihood function for the correlation function scale parameter \phi and tenasformation parameter \lambda only for models with fixed nugget effect (i.e., when it is not a parameter to be estimated) 
	## This function is used when computing the profile likelihood for \sigma^2
	## This is an auxiliary function called by proflik.aux2
	##  phi <- pmax(phi, temp.list$lower.phi)
	if(length(phi.lambda) == 2) include.lambda <- TRUE else include.lambda <-
			FALSE
	sigmasq <- temp.list$sigmasq
	z <- temp.list$z
	n <- length(z)
	phi <- phi.lambda[1]
	if(include.lambda) {
		lambda <- phi.lambda[2]
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z <= 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 0,
		cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- as.vector(z %*% (covinf$inverse - crossprod(txiv, ixix) %*%
		txiv) %*% z)
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(sigmasq) + (0.5/sigmasq) * ssresmat - 
			temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(sigmasq) +
			(0.5/sigmasq) * ssresmat - 0.5 * sum(log(eigentrem$
			values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux3.1" <- 
function(phi.lambda, temp.list)
{
	# This function computer the negative of the likelihood function for the correlation function scale parameter \phi and tenasformation parameter \lambda only for models with fixed nugget effect (i.e., when it is not a parameter to be estimated) 
	# This function is used when computing the profile likelihood for \sigma^2
	# This is an auxiliary function called by proflik.aux2
	#  phi <- pmax(phi, temp.list$lower.phi)
	sigmasq <- temp.list$sigmasq
	z <- temp.list$z
	n <- length(z)
	phi <- phi.lambda[1]
	lambda <- phi.lambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 0,
		cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- as.vector(z %*% (covinf$inverse - crossprod(txiv, ixix) %*%
		txiv) %*% z)
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(sigmasq) + (0.5/sigmasq) * ssresmat - 
			temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(sigmasq) +
			(0.5/sigmasq) * ssresmat - 0.5 * sum(log(eigentrem$
			values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux30" <- 
function(phitausq.rel, temp.list)
{
	# This function computes the value of the profile likelihood for the correlation parameter \phi when nugget effect is not included in the model.
	# It requires the minimisation of the function wrt \phi for each value of \lambda, if this transformation parameter is included in the model
	# This is an auxiliary function called by likfit.proflik
	#
	phitausq.rel <- as.vector(phitausq.rel)
	if(temp.list$fix.lambda == T)
		proflik <- proflik.aux21(phitausq.rel = phitausq.rel, temp.list
			 = temp.list)
	else {
		temp.list$phitausq.rel <- phitausq.rel
		proflik <-  - (nlminb(temp.list$lambda, proflik.aux21.1, lower
			 = -3, upper = 3, temp.list = temp.list)$objective)
		temp.list$phitausq.rel <- NULL
	}
	return(proflik)
}

"proflik.aux31" <- 
function(philambda, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for range \phi and the transformation parameter \lambda.
	# It requires the minimisation of the function wrt \tau^2_{rel} and for each value of (\phi,\lambda).
	# This is an auxiliary function called by proflik.
	philambda <- as.vector(philambda)
	temp.list$phi <- philambda[1]
	lambda <- philambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(temp.list$z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(temp.list$z^(lambda -
				1)))
		else temp.list$log.jacobian <- sum(log(temp.list$z^(lambda -
				1)))
		if(lambda == 0)
			temp.list$z <- log(temp.list$z)
		else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
	}
	tausq.rel.res <- nlminb(temp.list$tausq.rel.est, proflik.aux8, lower = 
		0, temp.list = temp.list)
	return( - tausq.rel.res$objective)
}

"proflik.aux32" <- 
function(tausqlambda, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for nugget \tau^2 and the transformation parameter \lambda.
	# It requires the minimisation of the function wrt \phi and \sigma^2 and for each value of (\tau^2,\lambda).
	# This is an auxiliary function called by proflik.
	tausqlambda <- as.vector(tausqlambda)
	temp.list$nugget <- tausqlambda[1]
	lambda <- tausqlambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(temp.list$z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(temp.list$z^(lambda -
				1)))
		else temp.list$log.jacobian <- sum(log(temp.list$z^(lambda -
				1)))
		if(lambda == 0)
			temp.list$z <- log(temp.list$z)
		else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
	}
	sigmasqphi.res <- nlminb(c(temp.list$sigmasq.est, temp.list$phi.est),
		proflik.aux12, lower = c(temp.list$lower.sigmasq, temp.list$
		lower.phi), temp.list = temp.list)
	return( - sigmasqphi.res$objective)
}

"proflik.aux33" <- 
function(tausq.rellambda, temp.list)
{
	# This function computes the value of the 2-D profile likelihood for nugget \tau^2 and the transformation parameter \lambda.
	# It requires the minimisation of the function wrt \phi for each value of (\tau^2,\lambda).
	# This is an auxiliary function called by proflik.
	tausq.rellambda <- as.vector(tausq.rellambda)
	temp.list$nugget.rel <- tausq.rellambda[1]
	lambda <- tausq.rellambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(temp.list$z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(temp.list$z^(lambda -
				1)))
		else temp.list$log.jacobian <- sum(log(temp.list$z^(lambda -
				1)))
		if(lambda == 0)
			temp.list$z <- log(temp.list$z)
		else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
	}
	phi.res <- nlminb(temp.list$phi.est, proflik.aux6, lower = temp.list$
		lower.phi, temp.list = temp.list)
	return( - phi.res$objective)
}

"proflik.aux4" <- 
function(lambda, temp.list)
{
	# This function computer the values of the profile likelihood function for the parameters \phi  and \sigma^2 for models with nugget effect = 0, including the tranformation parameter \lambda
	# This is an auxiliary function called by proflik.aux28
	#
	sigmasqphi <- as.vector(temp.list$sigmasqphi)
	sigmasq <- sigmasqphi[1]
	phi <- sigmasqphi[2]
	z <- temp.list$z
	n <- temp.list$n
	if(lambda > 0.94999999999999996 & lambda < 1.05)
		lambda <- 1
	if(lambda == 1) {
		log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			log.jacobian <- log(prod(z^(lambda - 1)))
		else log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$nugget, cov.pars = c(sigmasq, phi), scale = F, inv = 
		T, det = T, sqrt.inv = T)
	six <- crossprod(covinf$sqrt.inverse, xmat)
	xix <- crossprod(six)
	if(length(as.vector(xix)) == 1) {
		ixix <- 1/xix
		choldet <- 0.5 * log(xix)
	}
	else {
		chol.xix <- chol(xix)
		ixix <- chol2inv(chol.xix)
		choldet <- sum(log(diag(chol.xix)))
	}
	xi <- crossprod(xmat, covinf$inverse)
	ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
	if(temp.list$method == "ML") {
		neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
			ssresmat - log.jacobian)
	}
	if(temp.list$method == "RML") {
		xx.eigen <- eigen(crossprod(xmat), symmetric = T, only.values
			 = T)
		neglik <- (((n - beta.size)/2) * log(2 * pi) - 0.5 * sum(log(
			xx.eigen$values)) + covinf$log.det.to.half + (0.5) *
			ssresmat + choldet + log.jacobian)
	}
	return(as.vector(neglik))
}

"proflik.aux4.1" <- 
function(lambda, temp.list)
{
	# This function computer the values of the profile likelihood function for the parameters \phi  and \sigma^2 for models with nugget effect = 0, including the tranformation parameter \lambda
	# This is an auxiliary function called by proflik.aux28
	#
	sigmasqphi <- as.vector(temp.list$sigmasqphi)
	sigmasq <- sigmasqphi[1]
	phi <- sigmasqphi[2]
	z <- temp.list$z
	n <- temp.list$n
	if(lambda > 0.94999999999999996 & lambda < 1.05)
		lambda <- 1
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$nugget, cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(sigmasq) + (0.5/sigmasq) * ssresmat + 
			temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(sigmasq) +
			(0.5/sigmasq) * ssresmat - (n/2) - 0.5 * sum(log(
			eigentrem$values)) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux5" <- 
function(tausq.rel, temp.list)
{
	# This function computes the value of the profile likelihood for the parameter \tau^2_{rel}.
	# It requires the minimisation of the function wrt \phi and \lambda (if the case) for each value of \tau^2_{rel}.
	# This is an auxiliary function called by proflik.
	temp.list$nugget.rel <- as.vector(tausq.rel)
	if(temp.list$fix.lambda == T) {
		phi.res <- nlminb(temp.list$phi.est, proflik.aux6, lower = 
			temp.list$lower.phi, temp.list = temp.list)
	}
	else {
		phi.res <- nlminb(c(temp.list$phi.est, temp.list$lambda), 
			proflik.aux6, lower = c(temp.list$lower.phi, -3), upper
			 = c( + Inf, 3), temp.list = temp.list)
	}
	return( - phi.res$objective)
}

"proflik.aux6" <- 
function(phi.lambda, temp.list)
{
	## This function computes the value of the profile likelihood for the relative nugget parameter \tau^2_{rel}, minimizing the likelihood wrt correlation function scale parameter \phi (range) and the transformation parameter \lambda.
	if(length(phi.lambda) == 2) include.lambda <- TRUE else include.lambda <-
			FALSE
	phi.lambda <- as.vector(phi.lambda)
	phi <- phi.lambda[1]
	z <- temp.list$z
	n <- temp.list$n
	if(include.lambda) {
		lambda <- phi.lambda[2]
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z <= 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$nugget.rel, cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
			n) + (n/2) - 0.5 * sum(log(eigentrem$values)) - 
			temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux6.1" <- 
function(phi.lambda, temp.list)
{
	# This function computes the value of the profile likelihood for the relative nugget parameter \tau^2_{rel}, minimizing the likelihood wrt correlation function scale parameter \phi (range) and the transformation parameter \lambda.
	phi.lambda <- as.vector(phi.lambda)
	phi <- phi.lambda[1]
	lambda <- phi.lambda[2]
	z <- temp.list$z
	n <- temp.list$n
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		temp.list$nugget.rel, cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
			n) + (n/2) - 0.5 * sum(log(eigentrem$values)) - 
			temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux7" <- 
function(phi, temp.list)
{
	# This function computes the value of the profile likelihood for the parameter \phi when the nugget \tau^2 is included in the model
	# It requires the minimisation of the function wrt relative \tau^2_{rel} for each value of \phi
	# This is an auxiliary function called by proflik.
	temp.list$phi <- as.vector(phi)
	if(temp.list$fix.lambda == T) {
		tausq.rel.res <- nlminb(temp.list$tausq.rel.est, proflik.aux8,
			lower = 0, temp.list = temp.list)
	}
	else {
		tausq.rel.res <- nlminb(c(temp.list$tausq.rel.est, temp.list$
			lambda), proflik.aux8, lower = c(0, -3), upper = c(
			 + Inf, 3), temp.list = temp.list)
	}
	return( - tausq.rel.res$objective)
}

"proflik.aux8" <- 
function(tausq.rel.lambda, temp.list)
{
	## This function computes the value of the profile likelihood for the correlation function scale parameter \phi, minimizing the likelihood wrt relative nugget parameter \tau^2_{rel}
	if(length(tausq.rel.lambda) == 2) include.lambda <- TRUE else 
			include.lambda <- FALSE
	z <- temp.list$z
	n <- temp.list$n
	phi <- temp.list$phi
	tausq.rel <- tausq.rel.lambda[1]
	if(include.lambda) {
		lambda <- tausq.rel.lambda[2]
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z <= 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		tausq.rel, cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
			n) + (n/2) - 0.5 * sum(log(eigentrem$values))) - 
			temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux8.1" <- 
function(tausq.rel.lambda, temp.list)
{
	# This function computes the value of the profile likelihood for the correlation function scale parameter \phi, minimizing the likelihood wrt relative nugget parameter \tau^2_{rel}
	z <- temp.list$z
	n <- temp.list$n
	phi <- temp.list$phi
	tausq.rel <- tausq.rel.lambda[1]
	lambda <- tausq.rel.lambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		tausq.rel, cov.pars = c(1, phi), inv = T, det = T)
	txiv <- crossprod(xmat, covinf$inverse)
	ixix <- solve(txiv %*% xmat)
	ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
		z
	if(temp.list$method == "ML") {
		neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
	}
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixix, symmetric = T, only.values = T)
		neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
			n) + (n/2) - 0.5 * sum(log(eigentrem$values))) - 
			temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.aux9" <- 
function(sigmasq, temp.list)
{
	# This function computes the value of the profile likelihood for the parameter \sigma^2 when \tau^2 is included in the model
	# It requires the minimisation of the function wrt \phi and \tau^2 for each value of \sigma^2
	# This is an auxiliary function called by likfit.proflik
	temp.list$sigmasq <- as.vector(sigmasq)
	if(temp.list$fix.lambda == T) {
		phitausq.rel.res <- nlminb(c(temp.list$phi.est, temp.list$
			tausq.rel.est), proflik.aux10, lower = c(temp.list$
			lower.phi, 0), temp.list = temp.list)
	}
	else {
		phitausq.rel.res <- nlminb(c(temp.list$phi.est, temp.list$
			tausq.rel.est, temp.list$lambda), proflik.aux10, lower
			 = c(temp.list$lower.phi, 0, -3), upper = c( + Inf,
			 + Inf, 3), temp.list = temp.list)
	}
	return( - phitausq.rel.res$objective)
}

"proflik.cov" <- 
function(theta, temp.list)
{
	z <- temp.list$z
	n <- temp.list$n
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	if(temp.list$fixtau == T) {
		tausq <- temp.list$ftau
		sigmasq <- theta[1]
		phi <- theta[2]
	}
	else {
		tausq <- theta[1]
		sigmasq <- theta[2]
		phi <- theta[3]
	}
	sill.tot <- sigmasq + tausq
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = (
		tausq/sigmasq), cov.pars = c(1, phi), sqrt.inv = T, inv = T,
		det = T)
	six <- crossprod(covinf$sqrt.inverse, xmat)
	xix <- crossprod(six)
	if(length(as.vector(xix)) == 1) {
		ixix <- 1/xix
		choldet <- 0.5 * log(xix)
	}
	else {
		chol.xix <- chol(xix)
		i.chol.xix <- solve.upper(chol.xix)
		ixix <- i.chol.xix %*% t(i.chol.xix)
		choldet <- sum(log(diag(chol.xix)))
	}
	xi <- crossprod(xmat, covinf$inverse)
	ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
	if(temp.list$method == "ML") {
		neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(ssresmat/n) + (n/2)) - temp.list$
			log.jacobian
	}
	if(temp.list$method == "RML") {
		neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
			n) + (n/2) + choldet) - temp.list$log.jacobian
	}
	return(as.vector(neglik))
}

"proflik.ftau" <- 
function(theta, temp.list)
{
	if(length(theta) == 3)
		include.lambda <- TRUE
	else include.lambda <- FALSE
	z <- temp.list$z
	n <- temp.list$n
	lambda <- theta[3]
	if(include.lambda) {
		lambda <- theta[3]
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z < 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	txmat <- temp.list$txmat
	ftau <- temp.list$ftau
	sigmasq <- theta[1]
	sill.tot <- ftau + sigmasq
	phi <- theta[2]
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = ftau,
		cov.pars = c(sigmasq, phi), inv = T, det = T, sqrt.inv = T)
	six <- crossprod(covinf$sqrt.inverse, xmat)
	xix <- crossprod(six)
	if(length(as.vector(xix)) == 1) {
		ixix <- 1/xix
		choldet <- 0.5 * log(xix)
	}
	else {
		chol.xix <- chol(xix)
		i.chol.xix <- solve.upper(chol.xix)
		ixix <- i.chol.xix %*% t(i.chol.xix)
		choldet <- sum(log(diag(chol.xix)))
	}
	xi <- crossprod(xmat, covinf$inverse)
	ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
	if(temp.list$method == "ML") {
		neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
			ssresmat - temp.list$log.jacobian)
	}
	if(temp.list$method == "RML") {
		xx.eigen <- eigen(crossprod(xmat), symmetric = T, only.values
			 = T)
		neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + 0.5 * ssresmat + choldet - 0.5 * sum(
			log(xx.eigen$values)) - temp.list$log.jacobian)
	}
	return(as.vector(neglik))
}

"proflik.ftau1" <- 
function(theta, temp.list)
{
	z <- temp.list$z
	n <- temp.list$n
	lambda <- theta[3]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z < 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	txmat <- temp.list$txmat
	ftau <- temp.list$ftau
	sigmasq <- theta[1]
	sill.tot <- ftau + sigmasq
	phi <- theta[2]
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = ftau,
		cov.pars = c(sigmasq, phi), inv = T, det = T, sqrt.inv = T)
	six <- crossprod(covinf$sqrt.inverse, xmat)
	xix <- crossprod(six)
	if(length(as.vector(xix)) == 1) {
		ixix <- 1/xix
		choldet <- 0.5 * log(xix)
	}
	else {
		chol.xix <- chol(xix)
		i.chol.xix <- solve.upper(chol.xix)
		ixix <- i.chol.xix %*% t(i.chol.xix)
		choldet <- sum(log(diag(chol.xix)))
	}
	xi <- crossprod(xmat, covinf$inverse)
	ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
	if(temp.list$method == "ML") {
		neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
			ssresmat - temp.list$log.jacobian)
	}
	if(temp.list$method == "RML") {
		xx.eigen <- eigen(crossprod(xmat), symmetric = T, only.values
			 = T)
		neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + 0.5 * ssresmat + choldet - 0.5 * sum(
			log(xx.eigen$values)) - temp.list$log.jacobian)
	}
	return(as.vector(neglik))
}

"proflik.lambda" <- 
function(lambda, temp.list)
{
	z <- temp.list$z
	n <- temp.list$n
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z < 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	txmat <- temp.list$txmat
	ixx <- solve(crossprod(xmat))
	tausqhat <- (z %*% (diag(n) - xmat %*% ixx %*% txmat) %*% z)/n
	if(temp.list$method == "ML")
		neglik <- ((n/2) * log(2 * pi) + (n/2) * log(tausqhat) + (
			n/2) - temp.list$log.jacobian)
	if(temp.list$method == "RML") {
		eigentrem <- eigen(ixx, symmetric = T, only.values = T)
		neglik <- (((n - beta.size)/2) * log(2 * pi) + ((n - beta.size)/
			2) * log(tausqhat) + (n/2) - 0.5 * sum(log(eigentrem$
			values)) - temp.list$log.jacobian)
	}
	return(as.vector(neglik))
}

"proflik.nug" <- 
function(theta, temp.list)
{
	if(length(theta) == 3)
		include.lambda <- TRUE
	else include.lambda <- FALSE
	z <- temp.list$z
	n <- temp.list$n
	if(include.lambda) {
		lambda <- theta[3]
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z < 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	txmat <- temp.list$txmat
	tausq.rel <- theta[1]
	phi <- theta[2]
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		tausq.rel, cov.pars = c(1, phi), det = T, inv = T, sqrt.inv = T
		)
	six <- crossprod(covinf$sqrt.inverse, xmat)
	xix <- crossprod(six)
	if(length(as.vector(xix)) == 1) {
		ixix <- 1/xix
		choldet <- 0.5 * log(xix)
	}
	else {
		chol.xix <- chol(xix)
		i.chol.xix <- solve.upper(chol.xix)
		ixix <- i.chol.xix %*% t(i.chol.xix)
		choldet <- sum(log(diag(chol.xix)))
	}
	xi <- crossprod(xmat, covinf$inverse)
	ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
	if(temp.list$method == "ML") {
		neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
			)
	}
	if(temp.list$method == "RML") {
		xx.eigen <- eigen(crossprod(xmat), symmetric = T, only.values
			 = T)
		neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
			(n - beta.size)) + (n/2) + choldet - 0.5 * sum(log(
			xx.eigen$values)) - temp.list$log.jacobian)
	}
	return(as.vector(neglik))
}

"proflik.nug1" <- 
function(theta, temp.list)
{
	z <- temp.list$z
	n <- temp.list$n
	lambda <- theta[3]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z < 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	txmat <- temp.list$txmat
	tausq.rel <- theta[1]
	phi <- theta[2]
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 
		tausq.rel, cov.pars = c(1, phi), det = T, inv = T, sqrt.inv = T
		)
	six <- crossprod(covinf$sqrt.inverse, xmat)
	xix <- crossprod(six)
	if(length(as.vector(xix)) == 1) {
		ixix <- 1/xix
		choldet <- 0.5 * log(xix)
	}
	else {
		chol.xix <- chol(xix)
		i.chol.xix <- solve.upper(chol.xix)
		ixix <- i.chol.xix %*% t(i.chol.xix)
		choldet <- sum(log(diag(chol.xix)))
	}
	xi <- crossprod(xmat, covinf$inverse)
	ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
	if(temp.list$method == "ML") {
		neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
			)
	}
	if(temp.list$method == "RML") {
		xx.eigen <- eigen(crossprod(xmat), symmetric = T, only.values
			 = T)
		neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
			(n - beta.size)) + (n/2) + choldet - 0.5 * sum(log(
			xx.eigen$values)) - temp.list$log.jacobian)
	}
	return(as.vector(neglik))
}

"proflik.phi" <- 
function(theta, temp.list)
{
	if(length(theta) == 2)
		include.lambda <- TRUE
	else include.lambda <- FALSE
	z <- as.vector(temp.list$z)
	n <- temp.list$n
	phi <- theta[1]
	if(include.lambda) {
		lambda <- theta[2]
		if(lambda == 1) {
			temp.list$log.jacobian <- 0
		}
		else {
			if(any(z <= 0))
				stop("Transformation option not allowed when there are zeros or negative data"
					)
			if(any(z^(lambda - 1) <= 0))
				temp.list$log.jacobian <- log(prod(z^(lambda -
					1)))
			else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
			if(lambda == 0)
				z <- log(z)
			else z <- ((z^lambda) - 1)/lambda
		}
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	txmat <- temp.list$txmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 0,
		cov.pars = c(1, phi), inv = T, det = T, sqrt.inv = T)
	six <- crossprod(covinf$sqrt.inverse, xmat)
	xix <- crossprod(six)
	if(length(as.vector(xix)) == 1) {
		ixix <- 1/xix
		choldet <- 0.5 * log(xix)
	}
	else {
		chol.xix <- chol(xix)
		i.chol.xix <- solve.upper(chol.xix)
		ixix <- i.chol.xix %*% t(i.chol.xix)
		choldet <- sum(log(diag(chol.xix)))
	}
	xi <- crossprod(xmat, covinf$inverse)
	ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
	if(temp.list$method == "ML") {
		neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
			)
	}
	if(temp.list$method == "RML") {
		xx.eigen <- eigen(crossprod(xmat), symmetric = T, only.values
			 = T)
		neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
			(n - beta.size)) + (n/2) + choldet - 0.5 * sum(log(
			xx.eigen$values)) - temp.list$log.jacobian)
	}
	return(as.vector(neglik))
}

"proflik.phi1" <- 
function(phi.lambda, temp.list)
{
	z <- as.vector(temp.list$z)
	n <- temp.list$n
	phi <- phi.lambda[1]
	lambda <- phi.lambda[2]
	if(lambda == 1) {
		temp.list$log.jacobian <- 0
	}
	else {
		if(any(z <= 0))
			stop("Transformation option not allowed when there are zeros or negative data"
				)
		if(any(z^(lambda - 1) <= 0))
			temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
		else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
		if(lambda == 0)
			z <- log(z)
		else z <- ((z^lambda) - 1)/lambda
	}
	beta.size <- temp.list$beta.size
	kappa <- temp.list$kappa
	xmat <- temp.list$xmat
	txmat <- temp.list$txmat
	covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
		cov.model = temp.list$cov.model, kappa = kappa, nugget = 0,
		cov.pars = c(1, phi), inv = T, det = T, sqrt.inv = T)
	six <- crossprod(covinf$sqrt.inverse, xmat)
	xix <- crossprod(six)
	if(length(as.vector(xix)) == 1) {
		ixix <- 1/xix
		choldet <- 0.5 * log(xix)
	}
	else {
		chol.xix <- chol(xix)
		i.chol.xix <- solve.upper(chol.xix)
		ixix <- i.chol.xix %*% t(i.chol.xix)
		choldet <- sum(log(diag(chol.xix)))
	}
	xi <- crossprod(xmat, covinf$inverse)
	ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
	if(temp.list$method == "ML") {
		neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + (
			n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
			)
	}
	if(temp.list$method == "RML") {
		xx.eigen <- eigen(crossprod(xmat), symmetric = T, only.values
			 = T)
		neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
			log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
			(n - beta.size)) + (n/2) + choldet - 0.5 * sum(log(
			xx.eigen$values)) - temp.list$log.jacobian)
	}
	return(as.vector(neglik))
}

"rfm.bin" <- 
function(cloud, l = 15, uvec = "default", nugget.tolerance = 0, estimator.type
	 = c("classical", "robust"), bin.cloud = F, max.dist)
{
	#
	# binned f*.rfm function
	#
	# cloud : output from f*.rfm function with option 'cloud'
	# uvec : n-element vector of values to define the binning; the values of uvec defines the center of the bins 
	## if uvec[1] > 0 first bin is for 0< u <= uvec[2]-0.5*(uvec[2]-uvec[1])
	## if uvec[1] = 0 first bin is for 0< u <= 0.5*uvec[1] and uvec[1] is replaced by the midpoint of this interval
	# nugget.tolerance: pairs separated for a distance less than that va;lue are considered to be at same location (for direct estimation of nugget effect)
	# result is a list with components:
	#   u : uvec
	#   v : averages of empirical (semi)variogram ordinates within bins
	#   n : numbers of empirical (semi)variogram ordinates contributing
	#       to each bin
	##    if (all(uvec == "default")) 
	##        uvec <- seq(0, max(cloud$u), l = l)
	estimator.type <- match.arg(estimator.type)
	##  if(nugget.tolerance > 0) {
	##    dnug <- mean(cloud$u[cloud$u <= nugget.tolerance])
	##    cloud$u[cloud$u <= nugget.tolerance] <- 0
	##    uvec <- uvec[uvec > nugget.tolerance]
	##  }
	##  u <- c(0, uvec)
	##  n <- length(u)
	if(all(uvec == "default")) uvec <- seq(0, max.dist, l = 15)
	ubin <- c(0, uvec)
	nvec <- length(ubin)
	d <- 0.5 * diff(ubin[2:nvec])
	bins.lim <- c(0, (ubin[2:(nvec - 1)] + d), (d[nvec - 2] + ubin[nvec]))
	if(uvec[1] == 0 & nugget.tolerance == 0)
		uvec[1] <- (bins.lim[1] + bins.lim[2])/2
	if(nugget.tolerance > 0) {
		bins.lim <- c(0, nugget.tolerance, bins.lim[bins.lim > 
			nugget.tolerance])
		uvec <- c(0, (bins.lim[ - (1:2)] - 0.5 * diff(bins.lim)[-1]))
	}
	nbins <- nc <- length(bins.lim) - 1
	if(is.null(max.dist))
		max.dist <- max(bins.lim)
	min.dist <- min(cloud$u)
	##	d <- 0.5 * (u[3:n] - u[2:(n - 1)])
	##	low <- c(0, (u[2:(n - 1)] + d))
	##	high <- c((u[3:n] - d), (d[n - 2] + u[n]))
	##	nc <- n - 1
	if(!is.matrix(cloud$v)) {
		vbin <- rep(0, nc)
		nbin <- rep(0, nc)
		sdbin <- rep(0, nc)
		if(bin.cloud == T)
			bins.clouds <- list()
		for(i in 1:nc) {
			ind <- (cloud$u > bins.lim[i]) & (cloud$u <= bins.lim[
				i + 1])
			vbin[i] <- mean(cloud$v[ind])
			if(bin.cloud == T)
				bins.clouds[[i]] <- cloud$v[ind]
			nbin[i] <- sum(ind)
			if(estimator.type == "robust")
				vbin[i] <- ((vbin[i])^4)/(0.91400000000000003 +
					(0.98799999999999999/nbin[i]))
			if(nbin[i] > 0)
				sdbin[i] <- sqrt(var(cloud$v[ind]))
			else sdbin[i] <- NA
			NULL
		}
		# check == here
		if(uvec[1] == 0) uvec[1] <- (bins.lim[1] + bins.lim[2])/2
		# check == here
		if(min.dist == 0) {
			# check == here
			ind <- (cloud$u == 0)
			n.zero <- sum(ind)
			v.zero <- mean(cloud$v[ind])
			if(bin.cloud == T) {
				bins.clouds[[2:(length(bins.clouds) + 1)]] <-
					bins.clouds[1:n]
				bins.clouds[[1]] <- cloud$v[ind]
			}
			if(estimator.type == "robust")
				v.zero <- ((v.zero)^4)/(0.91400000000000003 +
					(0.98799999999999999/n.zero))
			if(n.zero > 0)
				sd.zero <- sqrt(var(cloud$v[ind]))
			else sd.zero <- NA
			uvec <- c(0, uvec)
			vbin <- c(v.zero, vbin)
			nbin <- c(n.zero, nbin)
			sdbin <- c(sd.zero, sdbin)
		}
		u <- uvec[!is.na(vbin)]
		v <- vbin[!is.na(vbin)]
		n <- nbin[!is.na(vbin)]
		sd <- sdbin[!is.na(vbin)]
		if(bin.cloud == T)
			bins.clouds <- bins.clouds[!is.na(vbin)]
	}
	else {
		if(bin.cloud == T)
			stop("option bins.cloud=T allowed only for 1 variable")
		nvcols <- ncol(cloud$v)
		vbin <- matrix(0, nrow = nc, ncol = nvcols)
		nbin <- rep(0, nc)
		sdbin <- matrix(0, nrow = nc, ncol = nvcols)
		for(i in 1:nc) {
			ind <- (cloud$u >= bins.lim[i]) & (cloud$u < bins.lim[
				i + 1])
			nbin[i] <- sum(ind)
			for(j in 1:nvcols) {
				vbin[i, j] <- mean(cloud$v[ind, j])
				if(estimator.type == "robust")
					vbin[i, j] <- ((vbin[i, j])^4)/(
						0.91400000000000003 + (
						0.98799999999999999/nbin[i]))
				if(nbin[i] > 0)
					sdbin[i, j] <- sqrt(var(cloud$v[ind,
						j]))
				else sdbin[i, j] <- NA
			}
			NULL
		}
		# check == here
		if(uvec[1] == 0) uvec[1] <- (bins.lim[1] + bins.lim[2])/2
		# check == here
		if(min.dist == 0) {
			v.zero <- rep(0, nvcols)
			n.zero <- rep(0, nvcols)
			sd.zero <- rep(0, nvcols)
			for(j in 1:nvcols) {
				# check == here
				ind <- (cloud$u == 0)
				n.zero[j] <- sum(ind)
				v.zero[j] <- mean(cloud$v[ind, j])
				if(estimator.type == "robust")
					v.zero[j] <- ((v.zero[j])^4)/(
						0.91400000000000003 + (
						0.98799999999999999/n.zero[
						j]))
				if(n.zero[j] > 0)
					sd.zero[j] <- sqrt(var(cloud$v[ind,
						j]))
				else sd.zero[j] <- NA
				uvec <- c(0, uvec)
				vbin <- rbind(v.zero, vbin)
				nbin <- c(n.zero, nbin)
				sdbin <- rbind(sd.zero, sdbin)
			}
		}
		u <- uvec[!is.na(vbin[, 1])]
		n <- nbin[!is.na(vbin[, 1])]
		v <- matrix(0, nrow = length(u), ncol = nvcols)
		sd <- matrix(0, nrow = length(u), ncol = nvcols)
		for(j in 1:nvcols) {
			v[, j] <- vbin[!is.na(vbin[, j]), j]
			#			sd[, j] <- sdbin[!is.na(sdbin[, j]), j]
			sd[, j] <- sdbin[!is.na(vbin[, j]), j]
		}
	}
	if(nugget.tolerance > 0) {
		u[1] <- nugget.tolerance
	}
	result <- list(u = u, v = v, n = n, sd = sd, output.type = "bin")
	if(!is.matrix(cloud$v) && bin.cloud == T)
		result$bin.cloud <- bins.clouds
	if(!is.null(class(cloud)))
		class(result) <- class(cloud)
	return(result)
}

"rfm.tests" <- 
function(f.data, f.data.mean, f.values, f.mean, f.n, test, outliers.p, 
	cloud.data, cloud.sim)
{
	results <- list()
	f.total <- sum(f.n)
	results$limits <- apply(f.values, 1, range)
	results$tests <- NULL
	pondera <- function(obj, w = f.n)
	{
		(obj %*% w)/sum(w)
	}
	if(any(test == 1)) {
		ind.out <- (f.data >= results$limits[2,  ] | f.data <= results$
			limits[1,  ])
		t1 <- sum(ind.out)
		ind.out <- NULL
		results$tests <- c(results$tests, T1 = t1)
	}
	if(any(test == 2)) {
		ind.out <- apply(f.values, 2, range)
		ind.out <- ind.out[2,  ] - ind.out[1,  ]
		ind.data <- range(f.data)
		ind.data <- ind.data[2] - ind.data[1]
		t2 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
		results$tests <- c(results$tests, T2 = t2)
		ind.out <- NULL
		ind.data <- NULL
	}
	if(any(test > 2)) {
		f.values <- abs(t(t(f.values) - f.mean))
		t.f.values <- t(f.values)
		f.data <- abs(f.data - f.data.mean)
	}
	if(any(test == 3)) {
		ind.data <- max(f.data)
		ind.out <- apply(f.values, 2, max)
		t3 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
		results$tests <- c(results$tests, T3 = t3)
		ind.out <- NULL
		ind.data <- NULL
	}
	if(any(test == 4 | test == 5 | test == 8 | test == 9)) {
		f.data2 <- f.data^2
		f.values2 <- f.values^2
		if(any(test == 4)) {
			ind.data <- sum(f.data2)
			ind.out <- apply(f.values2, 2, sum)
			t4 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
			results$tests <- c(results$tests, T4 = t4)
			ind.out <- NULL
			ind.data <- NULL
		}
		if(any(test == 5)) {
			ind.data <- pondera(f.data2, w = f.n)
			ind.out <- apply(f.values2, 2, pondera, w = f.n)
			t5 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
			results$tests <- c(results$tests, T5 = t5)
		}
		ind.out <- NULL
		ind.data <- NULL
	}
	if(any(test == 6 | test == 7)) {
		if(any(test == 6)) {
			ind.data <- sum(f.data)
			ind.out <- apply(f.values, 2, sum)
			t6 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
			results$tests <- c(results$tests, T6 = t6)
			ind.out <- NULL
			ind.data <- NULL
		}
		if(any(test == 7)) {
			ind.data <- pondera(f.data, w = f.n)
			ind.out <- apply(f.values, 2, pondera, w = f.n)
			t7 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
			results$tests <- c(results$tests, T7 = t7)
		}
		ind.out <- NULL
		ind.data <- NULL
	}
	if(any(test == 8 | test == 9)) {
		if(outliers.p == 0)
			ref.data <- max(f.data)
		else ref.data <- quantile(f.data, prob = (1 - outliers.p))
		f.data.r <- ifelse(f.data <= ref.data, (0.5 * f.data2), (
			ref.data * f.data - (ref.data^2/2)))
		if(outliers.p == 0)
			ref.values <- apply(f.values, 2, max)
		else ref.values <- apply(f.values, 2, quantile, prob = (1 -
				outliers.p))
		f.values.r <- ifelse((t(t.f.values <= ref.values)), (0.5 * 
			f.values2), (t(t.f.values * ref.values - (ref.values^
			2/2))))
		if(any(test == 8)) {
			ind.data <- sum(f.data.r)
			ind.out <- apply(f.values.r, 2, sum)
			t8 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
			results$tests <- c(results$tests, T8 = t8)
			ind.out <- NULL
			ind.data <- NULL
		}
		if(any(test == 9)) {
			ind.data <- pondera(f.data.r, w = f.n)
			ind.out <- apply(f.values.r, 2, pondera, w = f.n)
			t9 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
			results$tests <- c(results$tests, T9 = t9)
		}
		ind.out <- NULL
		ind.data <- NULL
	}
	if(any(test > 9)) {
		cloud.sim <- abs(t(t(cloud.sim) - f.mean))
		t.cloud.sim <- t(cloud.sim)
		cloud.data <- abs(cloud.data - f.data.mean)
		if(any(test == 10 | test == 12)) {
			cloud.data2 <- cloud.data^2
			cloud.sim2 <- cloud.sim^2
		}
		if(any(test == 10)) {
			ind.data <- sum(cloud.data2)
			ind.out <- apply(cloud.sim2, 2, sum)
			t10 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
			results$tests <- c(results$tests, T10 = t10)
			ind.out <- NULL
			ind.data <- NULL
		}
		if(any(test == 11)) {
			ind.data <- sum(cloud.data)
			ind.out <- apply(cloud.sim, 2, sum)
			t11 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
			results$tests <- c(results$tests, T11 = t11)
			ind.out <- NULL
			ind.data <- NULL
		}
		if(any(test == 12)) {
			if(outliers.p == 0)
				ref.data <- max(f.data)
			else ref.data <- quantile(cloud.data, prob = (1 - 
					outliers.p))
			cloud.data.r <- ifelse(cloud.data < ref.data, (0.5 *
				cloud.data2), (ref.data * cloud.data - (
				ref.data^2/2)))
			if(outliers.p == 0)
				ref.data <- apply(cloud.sim, 2, max)
			else ref.values <- apply(cloud.sim, 2, quantile, prob
					 = (1 - outliers.p))
			cloud.sim.r <- ifelse((t(t.cloud.sim <= ref.values)),
				(0.5 * cloud.sim2), (t(t.cloud.sim * ref.values -
				(ref.values^2/2))))
			ind.data <- sum(cloud.data.r)
			ind.out <- apply(cloud.sim.r, 2, sum)
			t12 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
			results$tests <- c(results$tests, T12 = t12)
			ind.out <- NULL
			ind.data <- NULL
		}
	}
	results$tests <- round(results$tests)
	return(results)
}

"summary.variomodel" <- 
#
function(obj)
{
	cat("Summary of the parameter estimation\n")
	cat("-----------------------------------\n")
	if(obj$method == "ML")
		method <- "maximum likelihood"
	if(obj$method == "RML")
		method <- "restricted maximum likelihood"
	if(obj$method == "OLS")
		method <- "ordinary least squares"
	if(obj$method == "WLS")
		method <- "weighted least squares"
	cat(paste("Estimation method:", method, "\n"))
	cat("\n")
	cat(paste("Covariance model:", obj$cov.model))
	if(obj$cov.model == "matern" | obj$cov.model == "powered.exponential" |
		obj$cov.model == "cauchy" | obj$cov.model == "gneiting.matern"
		)
		cat(paste(" with kappa = ", obj$kappa))
	if(!is.null(kappa))
		if(obj$cov.model == "matern" & obj$kappa == 0.5)
			cat(" (exponential)")
	cat("\n")
	cat("Covariance parameters:\n")
	print(c(nugget = obj$nugget, sill = obj$cov.pars[1], range = obj$
		cov.pars[2]))
	cat("\n")
	if(obj$method == "ML" | obj$method == "RML") {
		if(obj$lambda != 1) {
			if(obj$info.lambda$fix.lambda == F)
				cat(paste(
					"Estimated Box-Cox transformation parameter:",
					nice(obj$lambda)))
			else cat(paste("Fixed Box-Cox transformation parameter:",
					obj$lambda))
			cat("\n")
			cat("\n")
		}
		cat("mean parameter(s): ")
		if(length(obj$beta) == 1) {
			cat(obj$beta)
			cat("\n")
		}
		else {
			cat("\n")
			print(obj$beta)
		}
		cat("\n")
		print(c(log.L = obj$loglik, AIC = obj$AIC, BIC = obj$BIC, 
			n.params = round(obj$npars)))
	}
	return(invisible())
}

"trend.spatial" <- 
function(trend, coords = NULL)
{
	if(inherits(trend, "formula")) {
		trend.mat <- model.matrix(trend)
	}
	else {
		if(trend == "cte")
			trend.mat <- as.matrix(rep(1, nrow(coords)))
		else if(trend == "1st")
			trend.mat <- cbind(1, coords)
		else if(trend == "2nd")
			trend.mat <- cbind(1, coords, coords[, 1]^2, coords[
				, 2]^2, coords[, 1] * coords[, 2])
		else stop("external trend must be provided for data locations to be estimated using the argments `trend.d` and `trend.l`. Both (trend.d and trend.l) must be the same `cte`, `1st`, `2nd` or given by a formula of the type ~X where X is a matrix or vector of covariates."
				)
	}
	return(trend.mat)
}

"validation.spatial" <- 
#
function(coords, observed, predicted, variance.predicted = NULL, n.worse = 5)
{
	resid <- observed - predicted
	results <- cbind(coords, observed = observed, predicted = predicted,
		residuals = resid)
	if(!is.null(variance.predicted)) {
		sd.resid <- resid/sqrt(variance.predicted)
		results <- cbind(results, sd.resid = sd.resid)
	}
	return(results)
}

"varcov.chol" <- 
function()
{
}

"varcov.spatial" <- 
function(coords = NULL, dists.lowertri = NULL, cov.model = c("exponential",
	"matern", "gaussian", "spherical", "cubic", "wave", 
	"powered.exponential", "cauchy", "gneiting", "gneiting.matern", 
	"pure.nugget"), kappa = NULL, nugget = 0, cov.pars = stop(
	"no cov.pars argument"), inv = F, det = F, func.inv = "cholesky", 
	scaled = F, only.decomposition = F, sqrt.inv = F, test.decomp = FALSE,
	only.inv.lower.diag = FALSE)
{
	## This function builts the variance-covariance matrix for (2d) spatial data, and optionally returns also the inverse and the determinant of this matrix
	## coords : data coordinates
	## cov.model : see options in cov.spatial
	## kappa : parameter for matern, cauchy, powered.exponential or gneiting.matern covariance functions
	## nugget : nugget effect variance (tausq)
	## cov.pars: (sigmasq,phi) If cov.pars is a matrix, each line identify 1 structure
	## inv: if T returns the inverse of the covariance matrix (instead of the matrix)
	## det : if T returns the log of determinant of covariance matrix ^(0.5)
	## func.inv: method to invert the covariance matrix (cholesky, svd, solve, eigen)
	if(func.inv == "chol") func.inv <- "cholesky"
	if(only.inv.lower.diag)
		inv <- TRUE
	if(is.null(coords) & is.null(dists.lowertri))
		stop("either argument, coords or dists.lowertri must be provided"
			)
	if(!is.null(coords) & !is.null(dists.lowertri))
		stop("only ONE argument, either coords or dists.lowertri must be provided"
			)
	cov.model <- match.arg(cov.model)
	if(!is.null(coords))
		n <- nrow(coords)
	if(!is.null(dists.lowertri))
		n <- 0.5 * (1 + sqrt(1 + 8 * length(dists.lowertri)))
	tausq <- nugget
	if(is.vector(cov.pars))
		sigmasq <- cov.pars[1]
	else sigmasq <- cov.pars[, 1]
	if(!is.null(coords)) {
		dists.lowertri <- as.vector(dist(coords))
	}
	if(round(1000000000000. * min(dists.lowertri)) == 0)
		warning("Two or more pairs of data at coincident (or very close) locations. This can cause crashes in the matrices operations!\n"
			)
	##
	## Building the covariance matrix
	## 
	varcov <- matrix(0, n, n)
	if(scaled == F) {
		if(cov.pars[2] < 9.9999999999999998e-13)
			varcov <- diag(n) * cov.pars[1]
		else {
			covvec <- cov.spatial(obj = dists.lowertri, cov.model
				 = cov.model, kappa = kappa, cov.pars = 
				cov.pars)
			varcov[lower.tri(varcov)] <- covvec
			varcov <- t(varcov)
			varcov[lower.tri(varcov)] <- covvec
			diag(varcov) <- tausq + sum(sigmasq)
		}
	}
	else {
		if(cov.pars[2] < 9.9999999999999998e-13)
			varcov <- diag(n)
		else {
			if(is.vector(cov.pars))
				cov.pars.sc <- c(1, cov.pars[2])
			else cov.pars.sc <- cbind(1, cov.pars[, 2])
			covvec <- cov.spatial(obj = dists.lowertri, cov.model
				 = cov.model, kappa = kappa, cov.pars = 
				cov.pars.sc)
			varcov[lower.tri(varcov)] <- covvec
			varcov <- t(varcov)
			varcov[lower.tri(varcov)] <- covvec
			diag(varcov) <- 1 + (tausq/sum(sigmasq))
		}
	}
	##
	## Computing decompositions, inverses and determinants
	## 
	if(test.decomp == TRUE & func.inv == "cholesky") {
		varcov.eig <- eigen(varcov, symmetric = TRUE)
		if(any(varcov.eig$val < 1e-10))
			func.inv <- "eigen"
	}
	if(inv | det | only.decomposition | sqrt.inv | only.inv.lower.diag) {
		if(func.inv == "cholesky") {
			varcov.sqrt <- chol(varcov)
			if(only.decomposition == F) {
				if(inv) {
					invcov.sqrt <- solve.upper(varcov.sqrt)
					invcov <- invcov.sqrt %*% t(invcov.sqrt
						)
				}
				if(det)
					cov.logdeth <- sum(log(diag(varcov.sqrt
						)))
				if(sqrt.inv)
					inverse.sqrt <- solve(varcov.sqrt)
			}
		}
		if(func.inv == "svd") {
			varcov.svd <- svd(varcov, nv = 0)
			if(only.decomposition)
				varcov.sqrt <- t(varcov.svd$u %*% (t(varcov.svd$
					u) * sqrt(varcov.svd$d)))
			if(inv)
				invcov <- t(varcov.svd$u %*% (t(varcov.svd$
					u) * (1/varcov.svd$d)))
			if(sqrt.inv)
				inverse.sqrt <- t(varcov.svd$u %*% (t(
					varcov.svd$u) * (1/sqrt(varcov.svd$
					d))))
			if(det)
				cov.logdeth <- sum(log(sqrt(varcov.svd$d)))
		}
		if(func.inv == "eigen") {
			varcov.eig <- eigen(varcov, symmetric = T)
			if(only.decomposition)
				varcov.sqrt <- (varcov.eig$vec %*% diag(sqrt(
					varcov.eig$val)) %*% t(varcov.eig$
					vec))
			if(inv)
				invcov <- (varcov.eig$vec %*% diag(1/varcov.eig$
					val) %*% t(varcov.eig$vec))
			if(sqrt.inv)
				inverse.sqrt <- (varcov.eig$vec %*% diag(1/
					sqrt(varcov.eig$val)) %*% t(varcov.eig$
					vec))
			if(det)
				cov.logdeth <- sum(log(sqrt(varcov.eig$val)))
		}
		if(func.inv == "solve") {
			if(det)
				stop("this does not compute determinants. Use func.inv = cholesky, svd or eigen"
					)
			invcov <- solve(varcov)
		}
	}
	##
	## Preparing output
	##
	if(only.decomposition == FALSE) {
		if(det) {
			if(inv) {
				if(only.inv.lower.diag)
					result <- list(lower.inverse = invcov[
						lower.tri(invcov)], 
						diag.inverse = diag(invcov),
						log.det.to.half = cov.logdeth)
				else result <- list(inverse = invcov, 
						log.det.to.half = cov.logdeth)
			}
			else {
				result <- list(varcov = varcov, log.det.to.half
					 = cov.logdeth)
			}
			if(sqrt.inv)
				result$sqrt.inverse <- inverse.sqrt
		}
		else {
			if(inv)
				if(only.inv.lower.diag)
					result <- list(lower.inverse = invcov[
						lower.tri(invcov)], 
						diag.inverse = diag(invcov))
				else result <- invcov
			else result <- varcov
		}
	}
	else result <- t(varcov.sqrt)
	return(result)
}

"vario.matern" <- 
function(u, tausq, sigmasq, phi, kappa)
{
	#
	# variogram corresponding to matern correlation function
	#
	tausq + sigmasq * (1 - matern.d(u, phi, kappa))
}

"variog" <- 
function(geodata, coords = geodata$coords, data = geodata$data, uvec = 
	"default", trend = "cte", lambda = 1, option = c("bin", "cloud", 
	"smooth"), estimator.type = c("classical", "robust"), nugget.tolerance
	 = 0, max.dist = NULL, pairs.min = 2, bin.cloud = F, ...)
{
	#
	## coords : n by 2 matrix of locations
	## data   : n-vector of values
	## options: the output can be one of the the options provided
	## IF option='bin' the following argumants can be used:
	## uvec : n-element vector of values to define the binning; the values of uvec defines the center of the bins 
	### if uvec[1] > 0 first bin is for 0< u <= uvec[2]-0.5*(uvec[2]-uvec[1])
	### if uvec[1] = 0 first bin is for 0< u <= 0.5*uvec[1] and uvec[1] is replaced by the midpoint of this interval
	## nugget.tolerance: pairs separated for a distance less than that va;lue are considered to be at same location (for direct estimation of nugget effect)
	## max.dist: pairs separated for less than this distance are ignored
	## bins with less than this number of points are ignored
	## bin.cloud: the cloud values for each class are included in the output
	## trend: "cte", "1st", "2nd" or a formulae like ~X, where X is a matrix with covariates. If trend is provided the variogram for OLS residuals us plotted  
	#
	## IF option='smooth' the following argumant can be used:
	##
	## result is list with components:
	##   u : distances between all distinct pairs of locations
	##   v : corresponding empirical variogram values
	##
	## Extra arguments nugget.tolerance and uvec can be passed for rfm.bin
	##
	##
	## reading and checking input
	##
	call <- match.call()
	option <- match.arg(option)
	estimator.type <- match.arg(estimator.type)
	coords <- as.matrix(coords)
	data <- as.matrix(data)
	data.var <- apply(data, 2, var)
	n.data <- nrow(coords)
	n.datasets <- ncol(data)
	if(ncol(data) == 1)
		data <- as.vector(data)
	##
	## Box-Cox transformation
	##
	if(lambda != 1) {
		if(lambda == 0)
			data <- log(data)
		else data <- ((data^lambda) - 1)/lambda
	}
	##
	## Removing OLS trend
	##
	xmat <- trend.spatial(trend = trend, coords = coords)
	if(trend != "cte") {
		if(is.vector(data)) {
			data <- lm(data ~ xmat - 1)$residuals
			names(data) <- NULL
		}
		else {
			only.res <- function(y, x)
			{
				lm(y ~ x - 1)$residuals
			}
			data <- apply(data, 2, only.res, x = xmat)
		}
	}
	u <- dist(as.matrix(coords))
	if(option == "bin" & bin.cloud == FALSE) {
		if(!is.null(max.dist))
			umax <- max(u[u < max.dist])
		else umax <- max(u)
		if(all(uvec == "default"))
			uvec <- seq(0, umax, l = 15)
		ubin <- c(0, uvec)
		nvec <- length(ubin)
		d <- 0.5 * diff(ubin[2:nvec])
		bins.lim <- c(0, (ubin[2:(nvec - 1)] + d), (d[nvec - 2] + ubin[
			nvec]))
		if(uvec[1] == 0 & nugget.tolerance == 0)
			uvec[1] <- (bins.lim[1] + bins.lim[2])/2
		if(nugget.tolerance > 0) {
			bins.lim <- c(0, nugget.tolerance, bins.lim[bins.lim >
				nugget.tolerance])
			uvec <- c(0, (bins.lim[ - (1:2)] - 0.5 * diff(bins.lim)[
				-1]))
		}
		nbins <- length(bins.lim) - 1
		if(is.null(max.dist))
			max.dist <- max(bins.lim)
		temp.list <- list(n.data = n.data, coords = coords, data = data,
			nbins = nbins, bins.lim = bins.lim, estimator.type = 
			estimator.type, max.dist = max.dist)
		bin.f <- function(data, temp.list)
		{
			cbin <- vbin <- sdbin <- rep(0, temp.list$nbins)
			result <- .C("binit",
				as.integer(temp.list$n.data),
				as.double(as.vector(temp.list$coords[, 1])),

					as.double(as.vector(temp.list$coords[
					, 2])),
				as.double(as.vector(data)),
				as.integer(temp.list$nbins),
				as.double(as.vector(temp.list$bins.lim)),
				as.integer(temp.list$estimator.type == "robust"
					),
				as.double(temp.list$max.dist),
				cbin = as.integer(cbin),
				vbin = as.double(vbin),
				as.integer(TRUE),
				sdbin = as.double(sdbin))[c("vbin", "cbin",
				"sdbin")]
			return(result)
		}
		result <- array(unlist(lapply(as.data.frame(data), bin.f, 
			temp.list = temp.list)), dim = c(nbins, 3, n.datasets))
		indp <- (result[, 2, 1] >= pairs.min)
		result <- list(u = uvec[indp], v = result[indp, 1,  ], n = 
			result[indp, 2, 1], sd = result[indp, 3,  ], bins.lim
			 = bins.lim, ind.bin = indp)
	}
	else {
		if(is.matrix(data)) {
			v <- matrix(0, nrow = length(u), ncol = ncol(data))
			for(i in 1:ncol(data)) {
				v[, i] <- as.vector(dist(data[, i]))
				if(estimator.type == "robust")
					v[, i] <- v[, i]^(0.5)
				else v[, i] <- (v[, i]^2)/2
			}
			if(!is.null(max.dist)) {
				v <- v[u <= max.dist,  ]
				u <- u[u <= max.dist]
			}
		}
		else {
			v <- as.vector(dist(data))
			if(estimator.type == "robust")
				v <- v^(0.5)
			else v <- (v^2)/2
			if(is.numeric(max.dist)) {
				v <- v[u <= max.dist]
				u <- u[u <= max.dist]
			}
		}
		if(option == "cloud") {
			result <- list(u = u, v = v)
		}
		if(option == "bin") {
			if(!is.null(max.dist))
				umax <- max(u[u < max.dist])
			else umax <- max(u)
			result <- rfm.bin(cloud = list(u = u, v = v), 
				estimator.type = estimator.type, uvec = uvec,
				nugget.tolerance = nugget.tolerance, bin.cloud
				 = bin.cloud, max.dist = umax, ...)
			indp <- rep(TRUE, length(u))
			if(pairs.min > 0) {
				indp <- (result$n >= pairs.min)
				if(is.matrix(result$v)) {
					result$v <- result$v[indp,  ]
					result$sd <- result$sd[indp,  ]
				}
				else {
					result$v <- result$v[indp]
					result$sd <- result$sd[indp]
				}
				result$u <- result$u[indp]
				result$n <- result$n[indp]
			}
			result$ind.bin <- indp
		}
		if(option == "smooth") {
			if(is.matrix(v))
				stop("smooth not yet available for several variables"
					)
			temp <- ksmooth(u, v, ...)
			result <- list(u = temp[[1]], v = temp[[2]])
		}
	}
	result <- c(result, list(var.mark = data.var, output.type = option,
		estimator.type = estimator.type, n.data = n.data, lambda = 
		lambda, trend = trend, call = call))
	class(result) <- "variogram"
	return(result)
}

"variog.env" <- 
function(obj.variog, coords, model.pars, nsim = 99, messages.screen = T)
{
	# envelops for empirical variograms (produced by variog)
	# coords.data: data coordinates 
	# beta: mean parameter
	# cov.pars: covariance parameters c(sigmasq,phi)
	call <- match.call()
	##
	## reading input
	##
	if(!is.null(model.pars$beta)) beta <- model.pars$beta else beta <- 0
	if(!is.null(model.pars$cov.model))
		cov.model <- model.pars$cov.model
	else cov.model <- "exponential"
	if(!is.null(model.pars$kappa))
		kappa <- model.pars$kappa
	else kappa <- 0.5
	if(!is.null(model.pars$nugget))
		nugget <- model.pars$nugget
	else nugget <- 0
	cov.pars <- model.pars$cov.pars
	if(obj.variog$output.type != "bin")
		stop("envelops can be computed only for binned variogram")
	##
	## generating simulations from the model with parameters provided
	##
	if(messages.screen == T) cat(paste("variog.env: generating", nsim,
			"simulations(", obj.variog$n.data, 
			"points) using the function grf\n"))
	simula <- grf(obj.variog$n.data, grid = as.matrix(coords), cov.model = 
		cov.model, cov.pars = cov.pars, nugget = nugget, kappa = kappa,
		nsim = nsim, messages.screen = FALSE, lambda = obj.variog$
		lambda)
	if(messages.screen == T)
		cat("variog.env: adding the mean or trend\n")
	x.mat <- trend.spatial(trend = obj.variog$trend, coords = coords)
	simula$data <- as.vector(x.mat %*% beta) + simula$data
	##
	## computing empirical variograms for the simulations
	##
	if(messages.screen == T) cat(paste(
			"variog.env: computing the empirical variogram for the",
			nsim, "simulations\n"))
	nbins <- length(obj.variog$bins.lim) - 1
	temp.list <- list(n.data = obj.variog$n.data, coords = coords, nbins = 
		nbins, bins.lim = obj.variog$bins.lim, estimator.type = 
		obj.variog$estimator.type, max.u = max(obj.variog$u))
	bin.f <- function(sim, temp.list)
	{
		cbin <- vbin <- sdbin <- rep(0, temp.list$nbins)
		temp <- .C("binit",
			as.integer(temp.list$n.data),
			as.double(as.vector(temp.list$coords[, 1])),
			as.double(as.vector(temp.list$coords[, 2])),
			as.double(as.vector(sim)),
			as.integer(temp.list$nbins),
			as.double(as.vector(temp.list$bins.lim)),
			as.integer(temp.list$estimator.type == "robust"),
			as.double(temp.list$max.u),
			as.double(cbin),
			vbin = as.double(vbin),
			as.integer(FALSE),
			as.double(sdbin))$vbin
		return(temp)
	}
	simula.bins <- apply(simula$data, 2, bin.f, temp.list = temp.list)
	simula.bins <- simula.bins[obj.variog$ind.bin,  ]
	##
	## computing envelops
	##
	if(messages.screen) cat("variog.env: computing the envelops\n")
	limits <- apply(simula.bins, 1, range)
	res.env <- list(u = obj.variog$u, v.lower = limits[1,  ], v.upper = 
		limits[2,  ])
	return(res.env)
}

"wireframe.grf" <- 
function(obj, sim.number = 1, ...)
{
	obj$data <- as.matrix(obj$data)
	wireframe(obj$data[, sim.number] ~ obj$coords[, 1] * obj$coords[, 2],
		xlab = "Coord X", ylab = "Coord Y", zlab = "Simulated", ...)
}

"wireframe.grfsim" <- 
function(obj, sim.number = 1, ...)
{
	obj$data <- as.matrix(obj$data)
	wireframe(obj$data[, sim.number] ~ obj$coords[, 1] * obj$coords[, 2],
		xlab = "Coord X", ylab = "Coord Y", zlab = "Simulated", ...)
}

"wireframe.ksline" <- 
function(obj.ksline, grid = obj.ksline$grid, values = c("est", "kvar", 
	"oktrend", "ktrend", "ktetrend"), ...)
{
	#
	# obj.ksline: tipically an object generated by the function ksline
	# grid      : regular grid where the interpolation was performed
	# values    : valeus to be plotted. The options are:
	##        - est: the kriging estimates (default option)
	##        - kvar : the kriging variances
	values <- match.arg(values)
	#  grid <- obj.ksline$grid[order(grid[, 2], grid[, 1]),  ]
	#  if(nx == ny)
	#    par(pty = "s")
	if(values == "est") m <- obj.ksline$pred
	if(values == "kvar")
		m <- obj.ksline$krige.var
	wireframe(m ~ obj.ksline$grid[, 1] * obj.ksline$grid[, 2], xlab = 
		"Coord X", ylab = "Coord Y", zlab = "Simulated", ...)
}

"wlsfit" <- 
function(vario, ini, cov.model = c("exponential", "matern", "gaussian", 
	"spherical", "cubic", "wave", "powered.exponential", "cauchy", 
	"gneiting", "gneiting.matern", "pure.nugget"), kappa = NULL, 
	simul.number = NULL, max.dist = "all", minimisation.function = "nlminb",
	lower = 0, fix.nugget = F, nugget = 0, weight = c("npairs", "cressie"),
	messages.screen = TRUE)
{
	#
	# fit variogram model by WLS applied to binned variogram 
	#
	# vario : list with u, v and n components
	# ini : inial values of matern parameters theta=(tausq, sigmasq, phi)
	#         (PJD's parameterisation).
	# If a matrix of initial bvalues is provided, and minimisation.function="nlminb" a search for the best initial value is performed
	# kappa : FIXED value of kappa for Matern and powered classes of covariance function
	# max.dist : maximum distance in u component
	# minimisation.function : S-PLUS minimization procedure. options 'nlminb'(default) or 'ms'
	# lower : minimum value (or vector of minimum values) for parameters. 
	# This allows constrained minimization when using 'nlminb'
	#
	# results is list with components:
	#   theta : estimated value of theta
	#   kappa : fixed value of kappa used in Matern and powered exponential correlation functions
	#   value : minimised sum of squares (to allow informal optimisation
	#           over different values of kappa)
	#   minimisation.function : minimization minimisation.function used
	#   call : function call
	# simul.number : select one empirical variogram if the object in vario has more than one
	# weight : weights for the WLS
	call <- match.call()
	weight <- match.arg(weight)
	if(minimisation.function != "nlminb" & minimisation.function != "ms")
		stop("minimization functions available are: nlminb or ms")
	if(is.matrix(vario$v) & is.null(simul.number))
		stop("object in vario$v is a matrix. This function works for only 1 empirical variogram"
			)
	if(!is.null(simul.number))
		vario$v <- vario$v[, simul.number]
	if(is.matrix(ini)) {
		inilength <- dim(ini)[2]
		if(fix.nugget == F & inilength != 3)
			stop("wrong number of columns for ini (must be 3)")
		if(fix.nugget == T & inilength != 2)
			stop("wrong number of columns for ini (must be 2)")
	}
	else {
		inilength <- length(ini)
		if(fix.nugget == F & inilength != 3)
			stop("wrong length for ini (must be 3)")
		if(fix.nugget == T & inilength != 2)
			stop("wrong length for ini (must be 2)")
	}
	cov.model <- match.arg(cov.model)
	if(max.dist == "all")
		data <- list(u = vario$u, v = vario$v, n = vario$n)
	else data <- list(u = vario$u[vario$u <= max.dist], v = vario$v[vario$
			u <= max.dist], n = vario$n[vario$u <= max.dist])
	holdit <- data$v
	holdit[is.na(holdit)] <- 0
	data$v <- holdit
	data$kappa <- kappa
	data$cov.model <- cov.model
	if(minimisation.function == "ms") {
		if(fix.nugget == F) {
			if(weight == "npairs") {
				result <- ms( ~ n * ((v - ((sigmasq + tausq) -
					cov.spatial(u, cov.model = cov.model,
					kappa = kappa, cov.pars = c(sigmasq,
					phi))))^2), start = list(tausq = ini[
					1], sigmasq = ini[2], phi = ini[3]),
					data = data)
			}
			if(weight == "cressie") {
				cov.vec <- cov.spatial(u, cov.model = cov.model,
					kappa = kappa, cov.pars = c(sigmasq,
					phi))
				result <- ms( ~ (n/(cov.vec^2)) * ((v - ((
					sigmasq + tausq) - cov.vec))^2), start
					 = list(tausq = ini[1], sigmasq = ini[
					2], phi = ini[3]), data = data)
			}
			nugget <- result$parameter[1]
			cov.pars <- as.vector(result$parameter[2:3])
		}
		else {
			data$nugget <- nugget
			if(weight == "npairs") {
				result <- ms( ~ n * ((v - ((sigmasq + nugget) -
					cov.spatial(u, cov.model = cov.model,
					kappa = kappa, cov.pars = c(sigmasq,
					phi))))^2), start = list(sigmasq = ini[
					1], phi = ini[2]), data = data)
			}
			if(weight == "cressie") {
				cov.vec <- cov.spatial(u, cov.model = cov.model,
					kappa = kappa, cov.pars = c(sigmasq,
					phi))
				result <- ms( ~ (n/(cov.vec^2)) * ((v - ((
					sigmasq + tausq) - cov.vec))^2), start
					 = list(tausq = ini[1], sigmasq = ini[
					2], phi = ini[3]), data = data)
			}
			cov.pars <- as.vector(result$parameter[1:2])
		}
		value <- result$value
		value <- sum(value)
		message <- paste("ms does not return convergence message")
	}
	if(minimisation.function == "nlminb") {
		if(is.matrix(ini) | is.data.frame(ini)) {
			ini <- as.matrix(ini)
			scat("wlsfit: earching for the best initial value\n")
			ini.search <- ini
			dimnames(ini.search) <- list(NULL, NULL)
			wls.ini <- round(100000000. * apply(ini.search, 1,
				loss.wlsvario, data = data, fix.nugget = 
				fix.nugget, nugget = nugget, cov.model = 
				cov.model, weight = weight))
			ini <- as.vector(ini.search[wls.ini == min(wls.ini),
				])
			cat("wlsfit: best initial value:\n")
			if(fix.nugget == T) {
				names(ini) <- c("sill", "range")
				print(ini)
			}
			else {
				names(ini) <- c("nugget", "sill", "range")
				print(ini)
			}
		}
		result <- nlminb(ini, loss.wlsvario, lower = lower, data = data,
			fix.nugget = fix.nugget, nugget = nugget, kappa = kappa,
			cov.model = cov.model, weight = weight)
		value <- result$objective
		message <- result$message
		if(fix.nugget == F) {
			nugget <- result$parameter[1]
			cov.pars <- as.vector(result$parameter[2:3])
		}
		else {
			cov.pars <- as.vector(result$parameter[1:2])
		}
	}
	estimation <- list(nugget = nugget, cov.pars = cov.pars, cov.model = 
		cov.model, kappa = kappa, value = value, max.dist = max(data$
		u), minimisation.function = minimisation.function, message = 
		message)
	estimation$method <- "WLS"
	estimation$call <- call
	class(estimation) <- "variomodel"
	if(messages.screen) {
		cat("wlsfit: estimated model parameters are:\n")
		cat(paste("covariance model:", cov.model))
		if(cov.model == "matern" | cov.model == "powered.exponential" |
			cov.model == "cauchy" | cov.model == "gneiting.matern"
			)
			cat(paste(" with kappa =", kappa))
		if(!is.null(kappa))
			if(cov.model == "matern" & kappa == 0.5)
				cat(" (exponential)")
		cat("\n")
		print(c(nugget = estimation$nugget, sill = estimation$cov.pars[
			1], range = estimation$cov.pars[2]))
	}
	return(estimation)
}

"xyzplot" <- 
function(geodata, coords = geodata$coords, data = geodata$data, eyes, planepts
	 = F, ...)
{
	xr <- range(coords[, 1])
	yr <- range(coords[, 2])
	zr <- range(data)
	if(missing(eyes))
		eyes <- c(8, 8, 3)
	eyex <- min(xr) - eyes[1] * (xr[2] - xr[1])
	eyey <- min(yr) - eyes[2] * (yr[2] - yr[1])
	eyez <- min(zr) + 2 * eyes[3] * (zr[2] - zr[1])
	eye <- c(eyex, eyey, eyez)
	#eye_c(-113.6,29,20000)
	pp <- persp(xr, yr, matrix(zr[1], ncol = 2, nrow = 2), zlim = zr, eye
		 = eye, box = F, xlab = "Coord X", ylab = "Coord Y", zlab = 
		"data")
	ptsin3d <- perspp(coords[, 1], coords[, 2], data, pp)
	ptsonplane <- perspp(coords[, 1], coords[, 2], rep(zr[1], length(data)),
		pp)
	points(ptsin3d, pch = 16, ...)
	if(planepts) {
		points(ptsonplane)
	}
	segments(ptsin3d$x, ptsin3d$y, ptsonplane$x, ptsonplane$y)
}

